/*
 * Copyright (c) 2007 Christian W. Guenther (christian@deckfour.org)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * License to link and use is also granted to open source programs which are not
 * licensed under the terms of the GPL, given that they satisfy one or more of
 * the following conditions: 1) Explicit license is granted to the ProM and
 * ProMimport programs for usage, linking, and derivative work. 2) Carte blance
 * license is granted to all programs developed at Eindhoven Technical
 * University, The Netherlands, or under the umbrella of STW Technology
 * Foundation, The Netherlands. For further exemptions not covered by the above
 * conditions, please contact the author of this code.
 */
package org.processmining.ocbc.conformance.visualization.views;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JPanel;

import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.log.model.XOCObject;
import org.processmining.ocbc.log.model.impl.xConstants;
import org.processmining.ocbc.models.AOCEdge;
import org.processmining.ocbc.models.ActivityEdge;


/**
 * @author Guangming Li (sirliguangming@gmail.com)
 * show the conformance checking result in event view, i.e., get all deviating information of one event;
 */
public class ResultEventView extends JPanel {

	private static final long serialVersionUID = -6450104278248630541L;
	private ConformanceCheckingResultView conformanceCheckingResultView;

	/**
	 * record all deviating events, i.e., events who have inexistent activities, refer to inexistent object, 
	 * have incorrect number of objects, or have incorrect number of target events;
	 */
	private List<XOCEvent> deviatingEventList = new ArrayList<XOCEvent>();

	/**
	 * record all deviating events with inexistent activities;
	 */
	private Map<XOCEvent, String> deviatingEventVSInexistentActivity = new HashMap<XOCEvent, String>();
	
	/**
	 * record all deviating events referring to inexistent objects;
	 */
	private Map<XOCEvent, List<XOCObject>> deviatingEventVSInexistentObjectReferenceList = new HashMap<XOCEvent, List<XOCObject>>();
	
	/**
	 * get all deviating events referring to incorrect classes with objects;
	 * @return
	 */
	public Map<XOCEvent, Map<String, List<XOCObject>>> deviatingEventVSIncorrectClassVSObjectLis;
	
	/**
	 * record all deviating events with violating AOC relations and incorrect object references;
	 */	
	private Map<XOCEvent, List<AOCEdge>> eventVSAOCRelationList = new HashMap<XOCEvent, List<AOCEdge>>();
	
	/**
	 * record all deviating events with violating AOC relations and incorrect object references;
	 */	
	private Map<XOCEvent, Map<AOCEdge, List<XOCObject>>> eventVSAOCConstraintVSObjectList = new HashMap<XOCEvent, Map<AOCEdge, List<XOCObject>>>();
	
	
	/**
	 * record all deviating events have incorrect number of target events;
	 */	
	private Map<XOCEvent, Map<ActivityEdge, List<XOCEvent>>> eventVSBehavioralConstraintVSPrecedenceEventList = new HashMap<XOCEvent, Map<ActivityEdge, List<XOCEvent>>>();
	private Map<XOCEvent, Map<ActivityEdge, List<XOCEvent>>> eventVSBehavioralConstraintVSResponseEventList = new HashMap<XOCEvent, Map<ActivityEdge, List<XOCEvent>>>();
	
	public ResultEventView(ConformanceCheckingResultView conformanceCheckingResultView) {
		this.conformanceCheckingResultView = conformanceCheckingResultView;
	}

	public ConformanceCheckingResultView getConformanceCheckingResultView(){
		return conformanceCheckingResultView;
	}

	/**
	 * get all deviating events, i.e., events who have inexistent activities, refer to inexistent object, 
	 * have incorrect number of objects, or have incorrect number of target events;
	 * @return
	 */
	public List<XOCEvent> getDeviatingEventList(){
		return deviatingEventList;
	}
	
	/**
	 * get all deviating events with inexistent activities;
	 * @return
	 */
	public Map<XOCEvent, String> getDeviatingEventVSInexistentActivityMap(){
		return deviatingEventVSInexistentActivity;
	}
	
	/**
	 * get all deviating events referring to inexistent objects;
	 * @return
	 */
	public Map<XOCEvent, List<XOCObject>> getDeviatingEventVSInexistentObjectReferenceListMap(){
		return deviatingEventVSInexistentObjectReferenceList;
	}
	
	/**
	 * get all deviating events referring to incorrect classes with objects;
	 * @return
	 */
	public Map<XOCEvent, Map<String, List<XOCObject>>> getDeviatingEventVSReferredIncorrectClassVSObjectListMap(){
		return deviatingEventVSIncorrectClassVSObjectLis;
	}
	
	/**
	 * get all deviating events with violated AOC constraints and incorrect object reference list;
	 * @return
	 */
	public Map<XOCEvent, List<AOCEdge>> getDeviatingEventVSAOCRelationListMap(){
		return eventVSAOCRelationList ;
	}
	
//	/**
//	 * get all deviating events with violated AOC constraints and incorrect object reference list;
//	 * @return
//	 */
//	public Map<XOCEvent, Map<AOCEdge, List<XOCObject>>> getDeviatingEventVSAOCConstraintVSObjectListMap(){
//		return eventVSAOCConstraintVSObjectList ;
//	}
	
	/**
	 * get all deviating events with violated behavioral constraints and incorrect precedence events;
	 * @return
	 */
	public Map<XOCEvent, Map<ActivityEdge, List<XOCEvent>>> getDeviatingEventVSBehavioralConstraintVSPrecedenceEventListMap(){
		return eventVSBehavioralConstraintVSPrecedenceEventList;
	}
	
	/**
	 * get all deviating events with violated behavioral constraints and incorrect response events;
	 * @return
	 */
	public Map<XOCEvent, Map<ActivityEdge, List<XOCEvent>>> getDeviatingEventVSBehavioralConstraintVSResponseEventListMap(){
		return eventVSBehavioralConstraintVSResponseEventList;
	}
	
	/**
	 * get all information for one deviating event, i.e., inexistent activities, inexistent object references, 
	 * incorrect number of objects, or incorrect number of target events;
	 * @param event
	 */
	public void collectDeviatingInformationForEvent(XOCEvent event){
		//add deviating inexistent activities
		List<XOCEvent> deviatingEventWithInexistentList = conformanceCheckingResultView.getActivityExistenceChecker().
				getDeviatingEventList();
		if(deviatingEventWithInexistentList.contains(event)){
			deviatingEventVSInexistentActivity.put(event, event.getAttributes().get(xConstants.ACTIVITY).toString());
		}	
		
		//add deviating events with inexistent object reference list
		deviatingEventVSInexistentObjectReferenceList = conformanceCheckingResultView.getObjectExistenceChecker().getEventVSInexistentObjectListMap();
		
		//add deviating events referring to incorrect classes
		deviatingEventVSIncorrectClassVSObjectLis = conformanceCheckingResultView.getProperClassesChecker().getEventVSUnrelatedClassVSObjectListMap();
		
		//add deviating events with incorrect number of object references
		eventVSAOCRelationList = conformanceCheckingResultView.getObjectNumberPerEventChecker().getDeviatingEventVSAOCRelationListMap();
		
		//add deviating reference events (which do not have the required number of targets)
		Map<ActivityEdge, Map<XOCEvent, List<XOCEvent>>> behavioralConstraintVSEventVSPrecedenceEventList = conformanceCheckingResultView.getBehavioralConstraintsChecker()
				.getBehavioralConstraintVSEventVSPrecedenceEventListMap();
		for(ActivityEdge behavioralConstraint : behavioralConstraintVSEventVSPrecedenceEventList.keySet()){
			Map<XOCEvent, List<XOCEvent>> eventVSPrecedenceEventList = behavioralConstraintVSEventVSPrecedenceEventList.get(behavioralConstraint);
			if(eventVSPrecedenceEventList.containsKey(event)){
				if(!eventVSBehavioralConstraintVSPrecedenceEventList.containsKey(event)){
					eventVSBehavioralConstraintVSPrecedenceEventList.put(event, new HashMap<ActivityEdge, List<XOCEvent>>());
				}
				eventVSBehavioralConstraintVSPrecedenceEventList.get(event).put(behavioralConstraint, eventVSPrecedenceEventList.get(event));
			}
		}
		Map<ActivityEdge, Map<XOCEvent, List<XOCEvent>>> behavioralConstraintVSEventVSResponseEventList = conformanceCheckingResultView.getBehavioralConstraintsChecker()
				.getBehavioralConstraintVSEventVSResponseEventListMap();
		for(ActivityEdge behavioralConstraint : behavioralConstraintVSEventVSResponseEventList.keySet()){
			Map<XOCEvent, List<XOCEvent>> eventVSResponseEventList = behavioralConstraintVSEventVSResponseEventList.get(behavioralConstraint);
			if(eventVSResponseEventList.containsKey(event)){
				if(!eventVSBehavioralConstraintVSResponseEventList.containsKey(event)){
					eventVSBehavioralConstraintVSResponseEventList.put(event, new HashMap<ActivityEdge, List<XOCEvent>>());
				}
				eventVSBehavioralConstraintVSResponseEventList.get(event).put(behavioralConstraint, eventVSResponseEventList.get(event));
			}
		}
		
		if(deviatingEventVSInexistentActivity.containsKey(event) ||
				deviatingEventVSInexistentObjectReferenceList.containsKey(event) ||
				deviatingEventVSIncorrectClassVSObjectLis.containsKey(event) ||
				eventVSAOCRelationList.containsKey(event) ||
				eventVSBehavioralConstraintVSPrecedenceEventList.containsKey(event)){
			if(!deviatingEventList.contains(event)){
				deviatingEventList.add(event);
			}			
		}
	}
}
