/*
 * Copyright (c) 2007 Christian W. Guenther (christian@deckfour.org)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * License to link and use is also granted to open source programs which are not
 * licensed under the terms of the GPL, given that they satisfy one or more of
 * the following conditions: 1) Explicit license is granted to the ProM and
 * ProMimport programs for usage, linking, and derivative work. 2) Carte blance
 * license is granted to all programs developed at Eindhoven Technical
 * University, The Netherlands, or under the umbrella of STW Technology
 * Foundation, The Netherlands. For further exemptions not covered by the above
 * conditions, please contact the author of this code.
 */
package org.processmining.ocbc.conformance.visualization.views;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JPanel;

import org.processmining.common.structure.IntegerPair;
import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.log.model.XOCLog;
import org.processmining.ocbc.log.model.XOCObject;
import org.processmining.ocbc.models.AOCEdge;
import org.processmining.ocbc.models.ClassEdge;


/**
 * @author Guangming Li (sirliguangming@gmail.com)
 * show the conformance checking result in object version view, i.e., get all deviating information of one object version;
 */
public class ResultObjectVersionView extends JPanel {

	private static final long serialVersionUID = -6450104278248630541L;
	private ConformanceCheckingResultView conformanceCheckingResultView;

	/**
	 * indicate if the object version is deviating (violating always/eventually cardinality on class relation, 
	 * disappearing sometime or changing class, or having incorrect number of corresponding events)
	 */
	private boolean deviatingFlag = false;
	
	/**
	 * record all violated class relations (always) with corresponding object list;
	 */	
	private Map<ClassEdge, List<XOCObject>> violatedClassRelationAlwaysVSObjectList = new HashMap<ClassEdge, List<XOCObject>>();
	
	/**
	 * record all violated class relations (eventually) with corresponding object list;
	 */	
	private Map<ClassEdge, List<XOCObject>> violatedClassRelationEventuallyVSObjectList = new HashMap<ClassEdge, List<XOCObject>>();
	
	/**
	 * record all deviating AOC relations with corresponding event list;
	 */	
//	private Map<AOCEdge, List<XOCEvent>> violatedAOCRelationVSEventList = new HashMap<AOCEdge, List<XOCEvent>>();

	/**
	 * record all always violated AOC relation list;
	 */	
	private List<AOCEdge> alwaysViolatedAOCRelationList = new ArrayList<AOCEdge>();
	
	/**
	 * record all eventually violated AOC relation list;
	 */	
	private List<AOCEdge> eventuallyViolatedAOCRelationList = new ArrayList<AOCEdge>();
	
	public ResultObjectVersionView(ConformanceCheckingResultView conformanceCheckingResultView) {
		this.conformanceCheckingResultView = conformanceCheckingResultView;
	}

	public ConformanceCheckingResultView getConformanceCheckingResultView(){
		return conformanceCheckingResultView;
	}

	/**
	 * get the flag which indicates if object version is deviating (violating always/eventually cardinality on class relation, 
	 * disappearing sometime or changing class, or having incorrect number of corresponding events)
	 * @return
	 */
	public boolean getDeviatingFlag(){
		return deviatingFlag;
	} 
	
	/**
	 * get violatedClassRelationAlwaysVSObjectList which records all violated class relations (always) with corresponding object list;
	 * @return
	 */
	public Map<ClassEdge, List<XOCObject>> getViolatedClassRelationAlwaysVSObjectListMap(){
		return violatedClassRelationAlwaysVSObjectList;
	}
	
	/**
	 * get violatedClassRelationEventuallyVSObjectList which records all violated class relations (Eventually) with corresponding object list;
	 * @return
	 */
	public Map<ClassEdge, List<XOCObject>> getViolatedClassRelationEventuallyVSObjectListMap(){
		return violatedClassRelationEventuallyVSObjectList;
	}

	/**
	 * get violatedAOCRelationVSEventList which records all deviating AOC relations with corresponding event list;
	 * @return
	 */
//	public Map<AOCEdge, List<XOCEvent>> getViolatedAOCRelationVSEventListMap(){
//		return violatedAOCRelationVSEventList;
//	}
	
	/**
	 * get alwaysViolatedAOCRelationList which records all always violated AOC relation list;
	 * @return
	 */
	public List<AOCEdge> getAlwaysViolatedAOCRelationList(){
		return alwaysViolatedAOCRelationList;
	}
	
	/**
	 * get eventuallyViolatedAOCRelationList which records all eventually violated AOC relation list;
	 * @return
	 */
	public List<AOCEdge> getEventuallyViolatedAOCRelationList(){
		return eventuallyViolatedAOCRelationList;
	}
	
	/**
	 * get all deviation information for one object version (violated always/eventually cardinality on class relation, 
	 * disappearing sometime or changing class, or violated AOC relation cardinality);
	 * @param object
	 * @param event
	 * @param log
	 */
	public void collectDeviatingInformationForObjectVersion(XOCObject object, XOCEvent event, XOCLog log){
		int eventOrderNum = event.getOrderNum();
		//add violated class relation always
		Map<ClassEdge, List<XOCObject>> deviatingClassRelationVSSourceObjectList = conformanceCheckingResultView
				.getObjectModelValidityChecker().getEventVSDeviatingClassRelationVSSourceObjectListMap().get(event);
		if(deviatingClassRelationVSSourceObjectList != null){
			for(ClassEdge relation : deviatingClassRelationVSSourceObjectList.keySet()){
				if(deviatingClassRelationVSSourceObjectList.get(relation).contains(object)){
					violatedClassRelationAlwaysVSObjectList.put(relation, new ArrayList<XOCObject>());
				}
			}
		}
		Map<ClassEdge, List<XOCObject>> deviatingClassRelationVSTargetObjectList = conformanceCheckingResultView
				.getObjectModelValidityChecker().getEventVSDeviatingClassRelationVSTargetObjectListMap().get(event);
		if(deviatingClassRelationVSTargetObjectList != null){
			for(ClassEdge relation : deviatingClassRelationVSTargetObjectList.keySet()){
				if(deviatingClassRelationVSTargetObjectList.get(relation).contains(object)){
					violatedClassRelationAlwaysVSObjectList.put(relation, new ArrayList<XOCObject>());
				}
			}
		}

		
		//if the selected event is the last event, add violated class relation eventually
		if(log.size() == eventOrderNum){
			Map<ClassEdge, List<XOCObject>>  violatedSourceClassRelationVSObjectList = conformanceCheckingResultView
					.getFulfilmentChecker().getViolatedSourceClassRelationVSObjectListMap();
			for(ClassEdge relation : violatedSourceClassRelationVSObjectList.keySet()){
				if(violatedSourceClassRelationVSObjectList.get(relation).contains(object)){
					violatedClassRelationEventuallyVSObjectList.put(relation, new ArrayList<XOCObject>());
				}
			}
			Map<ClassEdge, List<XOCObject>> violatedTargetClassRelationVSObjectList = conformanceCheckingResultView
					.getFulfilmentChecker().getViolatedTargetClassRelationVSObjectListMap();
			for(ClassEdge relation : violatedTargetClassRelationVSObjectList.keySet()){
				if(violatedTargetClassRelationVSObjectList.get(relation).contains(object)){
					violatedClassRelationEventuallyVSObjectList.put(relation, new ArrayList<XOCObject>());
				}
			}
		}
		
		//add violated AOC relation
		Map<AOCEdge, Map<XOCObject, List<IntegerPair>>> violatedAOCRelationAlwaysVSObjectVSPeriodList = conformanceCheckingResultView.getEventNumberPerObjectChecker()
				.getViolatedAOCRelationAlwaysVSObjectVSPeriodListMap();
		for(AOCEdge relation : violatedAOCRelationAlwaysVSObjectVSPeriodList.keySet()){
			if(violatedAOCRelationAlwaysVSObjectVSPeriodList.get(relation).containsKey(object)){
				List<IntegerPair> periodList = violatedAOCRelationAlwaysVSObjectVSPeriodList.get(relation).get(object);
				for(IntegerPair period : periodList){
					int startValue = period.getFirstValue();
					int endValue = period.getSecondValue();
					if(startValue <= eventOrderNum && eventOrderNum <= endValue){
						alwaysViolatedAOCRelationList.add(relation);
						break;
					}
				}
			}			
		}
		Map<AOCEdge, Map<XOCObject, List<IntegerPair>>> violatedAOCRelationEventuallyVSObjectVSPeriodList = conformanceCheckingResultView.getEventNumberPerObjectChecker()
				.getViolatedAOCRelationEventuallyVSObjectVSPeriodListMap();
		for(AOCEdge relation : violatedAOCRelationEventuallyVSObjectVSPeriodList.keySet()){
			if(violatedAOCRelationEventuallyVSObjectVSPeriodList.get(relation).containsKey(object)){
				List<IntegerPair> periodList = violatedAOCRelationEventuallyVSObjectVSPeriodList.get(relation).get(object);
				for(IntegerPair period : periodList){
					int startValue = period.getFirstValue();
					int endValue = period.getSecondValue();
					if(startValue <= eventOrderNum && eventOrderNum <= endValue){
						eventuallyViolatedAOCRelationList.add(relation);
						break;
					}
				}
			}			
		}
		if(violatedClassRelationAlwaysVSObjectList.size() > 0 ||
				violatedClassRelationEventuallyVSObjectList.size() > 0 ||
				alwaysViolatedAOCRelationList.size() > 0 ||
				eventuallyViolatedAOCRelationList.size() > 0){
			deviatingFlag = true;
		}
	}
}
