package org.processmining.ocbc.discovery.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.processmining.framework.util.Pair;
import org.processmining.mxgraph.util.mxConstants;
import org.processmining.ocbc.discovery.metrics.MetricFormular;
import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.models.ClassEdge;


/**
 * A behavioral constraint pattern indicate the path to relate events.
 * It includes a reference activity, a target activity and a intermediary (a class or class relationship) and two AOC relations between activity and intermediary.
 * A pattern may have constraints of multiple constraint types, which has the same reference, target and intermediary as the pattern.
 * @author Guangming Li at August 30, 2017
 *
 */
public class BehavioralConstraintPattern {
	private String patternId; //referenceActivityNodeId + "-" + intermediaryId + "-" + targetActivityNodeId;	
	private String patternName; //referenceActivityName + "-" + intermediaryName + "-" + targetActivityName;	
	private String patternType; //indicate the type (mxConstants.CRELOCRELATION or mxConstants.CRELRTRELATION)
	private String referenceActivityName;
	private String referenceActivityNodeId;//records the  id of the node (in an OCBC model) corresponding the reference activity
	private String targetActivityName;
	private String targetActivityNodeId;
	
	/**
	 * source class is connected to reference activity while
	 * target class is connected to target activity;
	 */
	private String sourceClassName;
	private String sourceClassNodeId;
	private String targetClassName;
	private String targetClassNodeId;
	
	/**
	 * for the triangle (OC) pattern, the class relationship is null;
	 */
	private ClassEdge classRelationship;
	private String classRelationshipName;
	private String classRelationshipEdgeId;
	
	/**
	 * for the triangle (OC) pattern, the intermediary is common class;
	 * for the square (RT) pattern, the intermediary is class relationship;
	 */
	private String intermediaryName;
	private String intermediaryId;
	
	/**
	 * records the target events for each reference event corresponding to the pattern;
	 */
    private Map<XOCEvent, List<XOCEvent>> referenceEventVSPrecedenceTargetEventList = new HashMap<XOCEvent, List<XOCEvent>>();
    private Map<XOCEvent, List<XOCEvent>> referenceEventVSResponseTargetEventList = new HashMap<XOCEvent, List<XOCEvent>>();
    
	/**
	 * records the pairs each of which consists of two values;
	 * the first/second value indicates the number of target event before/after the reference event;
	 * the number of pairs equals to the number of reference events, i.e., each reference event corresponds one pair;
	 */
    private List<Pair<Integer, Integer>> precedenceAndResponsePairList = new ArrayList<Pair<Integer, Integer>>();
    
	/**
	 * records the metrics of each solution corresponding to this pattern;
	 */
    private Map<Long, Double> solutionVSFitness = new HashMap<Long, Double>();
    private Map<Long, Double> solutionVSPrecision = new HashMap<Long, Double>();
    private Map<Long, Double> solutionVSGeneralization = new HashMap<Long, Double>();
    
    /**
	 * records the types of constraints corresponding to the pattern;
	 */
    private List<String> constraintTypeList = new ArrayList<String>();
    
    /**
     * initialize parameters
     * @param patternType
     * @param referenceActivityName
     * @param referenceActivityNodeId
     * @param targetActivityName
     * @param targetActivityNodeId
     * @param sourceClassName
     * @param sourceClassNodeId
     * @param targetClassName
     * @param targetClassNodeId
     * @param classRelationshipName
     * @param classRelationshipEdgeId
     */
    public BehavioralConstraintPattern(String patternType, String referenceActivityName, String referenceActivityNodeId,
    		String targetActivityName, String targetActivityNodeId, String sourceClassName, String sourceClassNodeId,
    		String targetClassName, String targetClassNodeId, ClassEdge classRelationship){
    	this.patternType = patternType;
    	this.referenceActivityName = referenceActivityName;
    	this.referenceActivityNodeId = referenceActivityNodeId;
    	this.targetActivityName = targetActivityName;
       	this.targetActivityNodeId = targetActivityNodeId;
    	this.sourceClassName = sourceClassName;
    	this.sourceClassNodeId = sourceClassNodeId;
    	this.targetClassName = targetClassName;
    	this.targetClassNodeId = targetClassNodeId;
    	this.classRelationship = classRelationship;


    	if(patternType.equals(mxConstants.CRELOCRELATION)){
    		intermediaryName = sourceClassName;
			intermediaryId = sourceClassNodeId;
		}
		else if(patternType.equals(mxConstants.CRELRTRELATION)){
			intermediaryName = classRelationship.getEdgeName();
			intermediaryId = classRelationship.getID().toString();
		}
    	this.patternId = referenceActivityNodeId + "-" + intermediaryId + "-" + targetActivityNodeId;	
    	this.patternName = referenceActivityName + "-" + intermediaryName + "-" + targetActivityName;	
    }
    
	/**
	 * Transform a solution (parse the corresponding integer) into a model variant array
	 * A solution indicate the possible variants,
	 * <(0,0), (1,0), (2+,0), (0,1), (1,1), (2+,1), (0,2+), (1,2+), (2+,2+)> is the sequence of variants
	 * its state (allowed/not allowed) is denoted by a binary number with 9 bits: e.g., 0b101010000 
	 * the first variant (0,0) corresponds to the highest bit i.e., 1 while
	 * the last variant (2+,2+) corresponds to the lowest bit i.e., 0
	 * @param solution An integer to indicate the solution e.g., 0b101010000 
	 * @param modelVariantArray An array which contains the model variant flag, i.e., boolean values
	 * (the first element corresponds to the first variant (0,0)), e.g., model = {true,false,true,false,true,false,false,false,false}
	 */
	public void transformSolutionIntoModelVariantArray(Long solution, boolean[] modelVariantArray){
		long getBit = 1;
		for(int i=0; i<9; i++){
			if((solution & getBit) != 0){
				modelVariantArray[8-i] = true;
			}
			else{
				modelVariantArray[8-i] = false;
			}
			getBit = getBit << 1;
		}		
	}
	
	/**
	 * compute the metrics for each model solution (2^9 in total)
	 * based on the input pair list of a constraint source (consisting 
	 * of a reference activity, a target activity and a intermediary, i.e., a class or relationship);
	 * @param constraintPatternId: reference activity id + "-" + intermediary id + "-" + target activity id
	 * @param activityCardinalityPairList
	 */
	public void computeMetricsOfAllSolutionsForPattern(String fitnessAlgorithm, String precisionAlgorithm, String generalizationAlgorithm,
			double fitnessThreshold, double precisionThreshold, double generalizationThreshold,
			 boolean fitnessUpdateFlag, boolean precisionUpdateFlag, boolean generalizationUpdateFlag){
		//generate log variant array
		int beforeZeroAndAfterZero = 0;//corresponding to (0,0) which indicates there are no corresponding events after or before
		int beforeOneAndAfterZero = 0;
		int beforeMultipleAndAfterZero = 0;
		
		int beforeZeroAndAfterOne = 0;
		int beforeOneAndAfterOne = 0;
		int beforeMultipleAndAfterOne = 0;
		
		int beforeZeroAndAfterMultiple = 0;
		int beforeOneAndAfterMultiple = 0;
		int beforeMultipleAndAfterMultiple = 0;
		
		int acticvityCardinalityBefore = 0;
		int acticvityCardinalityAfter = 0;
		
		for(Pair<Integer, Integer> acticvityCardinalityPair : precedenceAndResponsePairList){
			acticvityCardinalityBefore = acticvityCardinalityPair.getFirst();
			acticvityCardinalityAfter = acticvityCardinalityPair.getSecond();
			
			if(acticvityCardinalityBefore == 0 && acticvityCardinalityAfter == 0){
				beforeZeroAndAfterZero ++;
			}			
			else if(acticvityCardinalityBefore == 1 && acticvityCardinalityAfter == 0){
				beforeOneAndAfterZero ++;
			}
			else if(acticvityCardinalityBefore >= 2 && acticvityCardinalityAfter == 0){
				beforeMultipleAndAfterZero ++;
			}
			
			else if(acticvityCardinalityBefore == 0 && acticvityCardinalityAfter == 1){
				beforeZeroAndAfterOne ++;
			}			
			else if(acticvityCardinalityBefore == 1 && acticvityCardinalityAfter == 1){
				beforeOneAndAfterOne ++;
			}
			else if(acticvityCardinalityBefore >= 2 && acticvityCardinalityAfter == 1){
				beforeMultipleAndAfterOne ++;
			}
			
			else if(acticvityCardinalityBefore == 0 && acticvityCardinalityAfter >= 2){
				beforeZeroAndAfterMultiple ++;
			}			
			else if(acticvityCardinalityBefore == 1 && acticvityCardinalityAfter >= 2){
				beforeOneAndAfterMultiple ++;
			}
			else if(acticvityCardinalityBefore >= 2 && acticvityCardinalityAfter >= 2){
				beforeMultipleAndAfterMultiple ++;
			}
		}		
		double[] logVariantArray = {beforeZeroAndAfterZero, beforeOneAndAfterZero, beforeMultipleAndAfterZero,
				beforeZeroAndAfterOne, beforeOneAndAfterOne, beforeMultipleAndAfterOne,
				beforeZeroAndAfterMultiple, beforeOneAndAfterMultiple, beforeMultipleAndAfterMultiple};
		
		//compute metrics between the log variant array and each model solution (2^9-1)
		boolean[] modelVariantArray = new boolean[9];			
		MetricFormular computingMetric = new MetricFormular();

		System.out.println("pattern: " + referenceActivityName + "-" + targetActivityName);
		for(long solution=1; solution<Math.pow(2, 9); solution++){
//			System.out.println("solution: " + solution);
			transformSolutionIntoModelVariantArray(solution, modelVariantArray);
//			for(int i=0; i<9; i++){S
//				System.out.println("modelVariantArray[i]: " + modelVariantArray[i]);
//			}
			if(fitnessUpdateFlag){
				double fitness = computingMetric.computeFitnessForModelAndLog(fitnessThreshold, 9, modelVariantArray, 
						logVariantArray, fitnessAlgorithm);
				solutionVSFitness.put(solution, fitness);
				System.out.println("add fitness for each solution");
				System.out.println("solution: " + Long.toBinaryString(solution));
				System.out.println("fitness: " + fitness);
			}

			if(precisionUpdateFlag){
				double precision = computingMetric.computePrecisionForModelAndLog(precisionThreshold, 9, modelVariantArray, 
						logVariantArray, precisionAlgorithm);
				solutionVSPrecision.put(solution, precision);
			}

			if(generalizationUpdateFlag){
				double generalization = computingMetric.computeGeneralizationForModelAndLog(0.95, 
						9, modelVariantArray, logVariantArray, generalizationAlgorithm);
				solutionVSGeneralization.put(solution, generalization);
			}		
		}
	}	
		
	/**
	 * find the best solutions based on weights of fitness, precision and generalization;
	 * ranked the solutions based on the mean
	 * Score = fitnessWeight*fitness(solution) + fitnessWeight*fitness(solution) + fitnessWeight*fitness(solution);
	 * @param constraintPatternId
	 * @param fitnessWeight: double in [0,1]
	 * @param precisionWeight: double in [0,1]
	 * @param generalizationWeight: double in [0,1]
	 * @return The best solution
	 */
	public Long searchBestSolutionBasedOnWeights(double fitnessWeight, double precisionWeight, double generalizationWeight){
		Long bestSolution = 511L;// corresponding a binary number: 1,1111,1111; i.e., every variant is positive
		double bestCombinedMetric = 0;
		
		double bestFitness = 0;
		double bestPrecision = 0;	
		double bestGeneralization = 0;
		
		for(Long solutionCandidate : solutionVSFitness.keySet()){
			double fitness = solutionVSFitness.get(solutionCandidate);
			double precision = solutionVSPrecision.get(solutionCandidate);
			double generalization = solutionVSGeneralization.get(solutionCandidate);
			
//			System.out.println("fitness in search best: " + fitness);
//			System.out.println("precision in search best: " + precision);
//			System.out.println("generalization in search best: " + generalization);
			
			double combinedMetricCandidate = (fitness*fitnessWeight + precision*precisionWeight + generalization*generalizationWeight)/
					(fitnessWeight + precisionWeight + generalizationWeight);
			//if it is a better solution (whose combined metric is larger)
			if(bestCombinedMetric < combinedMetricCandidate){
				bestSolution = solutionCandidate;
				bestCombinedMetric = combinedMetricCandidate;
				
				bestFitness = fitness;
				bestPrecision = precision;
				bestGeneralization = generalization;
			}			
		}
		System.out.println("bestFitness in search best: " + bestFitness);
		System.out.println("bestPrecision in search best: " + bestPrecision);
		System.out.println("bestGeneralization in search best: " + bestGeneralization);
		System.out.println("bestCombinedMetric in search best: " + bestCombinedMetric);
		System.out.println("bestSolution in search best: " + Long.toBinaryString(bestSolution));
		return bestSolution;		
	}
	
	/**
	 * Transform a solution (parse the corresponding integer) into a list contains the possible variants
	 * @param solution An integer to indicate the solution e.g., 0b101010000 
	 * @param modelVariantArray An array which contains the model variant flag, i.e., boolean values
	 * (the first element corresponds to the first variant (0,0)), e.g., model = {true,false,true,false,true,false,false,false,false}
	 */
	public void transformSolutionIntoPossibleVariantList(Long solution, List<String> modelVariantList){
		long getBit = 1;
//		System.out.println("solution: " + Long.toBinaryString(solution));
		for(int i=0; i<9; i++){
			if((solution & getBit) != 0){
				String allowedVariant = mxConstants.constraintVariantList.get(8-i);
				modelVariantList.add(allowedVariant);
//				System.out.println("allowed Variant in possible variant list: " + allowedVariant);
			}
//			System.out.println("getBit: " + Long.toBinaryString(getBit));
			getBit = getBit << 1;
		}		
	}
	
	/**
	 * Transform a solution (parse the corresponding integer) into a list of constraint types
	 * A solution indicate the possible variants,
	 * <(0,0), (1,0), (2+,0), (0,1), (1,1), (2+,1), (0,2+), (1,2+), (2+,2+)> is the sequence of variants
	 * its state (allowed/not allowed) is denoted by a binary number with 9 bits: e.g., 0b101010000 
	 * the first variant (0,0) corresponds to the highest bit i.e., 1 while
	 * the last variant (2+,2+) corresponds to the lowest bit i.e., 0
	 * @param solution An integer to indicate the solution
	 * @param minimalConstraintTypeList The minimal constraint type list
	 */
	public List<String> transformSolutionIntoConstraintTypeList(Long solution){		
		//transform solution into allowed variant list
		List<String> minimalConstraintTypeList = new ArrayList<String>();
		List<String> allowedVariantListBySolution = new ArrayList<String>();
		transformSolutionIntoPossibleVariantList(solution, allowedVariantListBySolution);
		
//		System.out.println("solution in transformSolutionIntoConstraintTypeList: " + solution);
		
		//check each constraint type
		for(String constraintType : mxConstants.constraintTypeVSVariantList.keySet()){
//			System.out.println("constraintType in transformSolutionIntoConstraintTypeList: " + constraintType);
			List<String> allowedVariantListByConstraintType = mxConstants.constraintTypeVSVariantList.get(constraintType);
			//if the constraint type allow the solution
			if(allowedVariantListByConstraintType.containsAll(allowedVariantListBySolution)){	
//				System.out.println("allowedVariantListByConstraintType.containsAll(allowedVariantListBySolution)");
				List<String> removedConstraintTypeList = new ArrayList<String>();
				boolean addedFlag = true;
				for(String addedConstraintType : minimalConstraintTypeList){
//					System.out.println("addedConstraintType: " + addedConstraintType);
					List<String> allowedVariantListByAddedConstraintType = mxConstants.constraintTypeVSVariantList.get(addedConstraintType);
					//if constraint type is looser than one type in the minimal list
					if(allowedVariantListByConstraintType.containsAll(allowedVariantListByAddedConstraintType)){
						addedFlag = false;
						break;
					}
					//if constraint type is stricter than one type in the minimal list
					else if(allowedVariantListByAddedConstraintType.containsAll(allowedVariantListByConstraintType)){
						removedConstraintTypeList.add(addedConstraintType);
						addedFlag = true;
					}
					//if they never contain each other
					else{
						addedFlag = true;
					}
				}
				if(addedFlag){
					minimalConstraintTypeList.add(constraintType);
				}
				minimalConstraintTypeList.removeAll(removedConstraintTypeList);				
			}
		}
		
		constraintTypeList = minimalConstraintTypeList;
		return minimalConstraintTypeList;
//		for(String constraintType : minimalConstraintTypeList){
//			System.out.println("constraintType in miminal set: " + constraintType);
//		}
	}
	
    /**
     * gets the pattern id (referenceActivityNodeId + "-" + intermediaryId + "-" + targetActivityNodeId)
     * @return
     */
    public String getPatternId(){
    	return patternId;
    }
    
    /**
     * gets the pattern name (referenceActivityName + "-" + intermediaryName + "-" + targetActivityName)
     * @return
     */
    public String getPatternName(){
    	return patternName;
    }
    
    /**
     * gets the pattern type (mxConstants.CRELOCRELATION or mxConstants.CRELRTRELATION)
     * @return
     */
    public String getPatternType(){
    	return patternType;
    }
 
    /**
     * gets the referenceActivityName
     * @return
     */
    public String getReferenceActivityName(){
    	return referenceActivityName;
    }
    
    /**
     * gets the  id of the node (in an OCBC model) corresponding the reference activity
     * @return
     */
    public String getReferenceActivityNodeId(){
    	return referenceActivityNodeId;
    }
    
    /**
     * gets the targetActivityName
     * @return
     */
    public String getTargetActivityName(){
    	return targetActivityName;
    }
    
    /**
     * gets the  id of the node (in an OCBC model) corresponding the target activity
     * @return
     */
    public String getTargetActivityNodeId(){
    	return targetActivityNodeId;
    }
    
    /**
     * gets the sourceClassName (source class is connected to reference activity)
     * @return
     */
    public String getSourceClassName(){
    	return sourceClassName;
    }
    
    
    /**
     * gets the  id of the node (in an OCBC model) corresponding the source class
     * @return
     */
    public String getSourceClassNodeId(){
    	return sourceClassNodeId;
    }
    
    /**
     * gets the targetClassName (target class is connected to target activity)
     * @return
     */
    public String getTargetClassName(){
    	return targetClassName;
    }
    
    
    /**
     * gets the  id of the node (in an OCBC model) corresponding the target class
     * @return
     */
    public String getTargetClassNodeId(){
    	return targetClassNodeId;
    }
    
    /**
     * gets the classRelationshipName;
     * it is null for the triangle (OC) pattern;
     * @return
     */
    public String getClassRelationshipName(){
    	return classRelationshipName;
    }
    
    /**
     * gets the classRelationshipEdgeId
     * @return
     */
    public String getClassRelationshipEdgeId(){
    	return classRelationshipEdgeId;
    }
    
    /**
     * gets the intermediaryName
     * @return
     */
    public String getIntermediaryName(){
    	return intermediaryName;
    }
    
    /**
     * gets the intermediaryId
     * @return
     */
    public String getIntermediaryId(){
    	return intermediaryId;
    }
    
    /**
     * gets the referenceEventVSPrecedenceTargetEventList
     * @return
     */
    public Map<XOCEvent, List<XOCEvent>> getReferenceEventVSPrecedenceTargetEventListMap(){
    	return referenceEventVSPrecedenceTargetEventList;
    }
    
    /**
     * gets the referenceEventVSResponseTargetEventList
     * @return
     */
    public Map<XOCEvent, List<XOCEvent>> getReferenceEventVSResposeTargetEventListMap(){
    	return referenceEventVSResponseTargetEventList;
    }
    
    /**
     * gets the constraintTypeList
     * @return
     */
    public List<Pair<Integer, Integer>> getPrecedenceAndResponsePairList(){
    	return precedenceAndResponsePairList;
    }
    
    /**
     * gets the solutionVSFitness
     * @return
     */
    public Map<Long, Double> getSolutionVSFitnessMap(){
    	return solutionVSFitness;
    }
    
    /**
     * gets the solutionVSPrecison
     * @return
     */
    public Map<Long, Double> getSolutionVSPrecisionMap(){
    	return solutionVSPrecision;
    }
    
    /**
     * gets the solutionVSGeneralization
     * @return
     */
    public Map<Long, Double> getSolutionVSGeneralizationMap(){
    	return solutionVSGeneralization;
    }
    
    /**
     * gets the constraintTypeList
     * @return
     */
    public List<String> getConstraintTypeList(){
    	return constraintTypeList;
    }
    
    public static void main(String[] args) throws Exception {

    }
}