package org.processmining.ocbc.discovery.metrics;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.math3.distribution.BetaDistribution;
import org.apache.commons.math3.distribution.BinomialDistribution;
import org.processmining.mxgraph.util.mxConstants;

public class MetricFormular {
	
	/**
	 * compute the fitness (one) based on formula: L_{a,b}(M_{a,b}) / L_{a,b}(V);
	 * M_{a,b} is the set of allowed constraint instance variants by the input model;
	 * L_{a,b}(v) indicate the number of reference events in activity a corresponding to variant v;
	 * L_{a,b}(V) indicate the sum of each L_{a,b}(v) where v in V;
	 * input constraintVariantNum: the number of constraint instance variants;
	 * input model: a set of (9) values in which one is true or false (e.g., 
	 * the third value is true, it means the third variant (2+,0) is allowed by the model);
	 * input log: a set of (9) values in which one is the number of corresponding instances;
	 */
	public double computeFitnessBasic(int constraintVariantNum, boolean[] model, double[] log){
		double fitness = 0;
		double logSizeTotal = 0;//corresponding to L_{a,b}(V)
		double logSizeFittingModel = 0;//corresponding to L_{a,b}(M_{a,b})		
		for(int i=0; i<constraintVariantNum; i++){
			logSizeTotal = logSizeTotal + log[i];
		}		
		for(int i=0; i<constraintVariantNum; i++){
			if(model[i]){
				logSizeFittingModel = logSizeFittingModel + log[i];
			}			
		}	
		if(logSizeTotal == 0){
			fitness = 1;
		}
		else{
			fitness = logSizeFittingModel/logSizeTotal;
		}		
		return fitness;
	}
	
	/**
	 * compute the fitness (two) based on formula: |{v in M_{a,b} | L_{a,b}(v) >= n}|  /  |{v in V | L_{a,b}(v) >= n}|;
	 * M_{a,b} is the set of allowed constraint instance variants by the input model;
	 * L_{a,b}(v) indicate the number of reference events in activity a corresponding to variant v;
	 * V is the set of constraint instance variants;
	 * input constraintVariantNum: the number of constraint instance variants;
	 * input model: a set of (9) values in which one is true or false (i.g., 
	 * the third value is true, it means the third variant is allowed by the model);
	 * input log: a set of (9) values in which one is the number of corresponding instances;
	 * input threshold: a positive value to filter the observed variant
	 */
	public double computeFitnessWithNumThreshold(int constraintVariantNum, boolean[] model, double[] log, int threshold){
		double fitness = 0;
		double satisfiedVariantNumInModel = 0; //correspond to |{v in M_{a,b} | L_{a,b}(v) >= n}|
		double satisfiedVariantNum = 0;//correspond to |{v in V | L_{a,b}(v) >= n}|
		for(int i=0; i<constraintVariantNum; i++){
			if(log[i] >= threshold){
				satisfiedVariantNum ++;
				if(model[i]){
					satisfiedVariantNumInModel ++;					
				}
			}			
		}	
		if(satisfiedVariantNum == 0){
			fitness = 1;
		}
		else{
			fitness = satisfiedVariantNumInModel/satisfiedVariantNum;
		}		
		return fitness;
	}
	
	/**
	 * compute the fitness (three) based on formula: |{v in M_{a,b} | L_{a,b}(v) >= %n}|  /  |{v in V | L_{a,b}(v) >= %n}|;
	 * M_{a,b} is the set of allowed constraint instance variants by the input model;
	 * L_{a,b}(v) indicate the number of reference events in activity a corresponding to variant v;
	 * V is the set of constraint instance variants;
	 * input constraintVariantNum: the number of constraint instance variants;
	 * input model: a set of (9) values in which one is true or false (i.g., 
	 * the third value is true, it means the third variant is allowed by the model);
	 * input log: a set of (9) values in which one is the number of corresponding instances;
	 * input threshold: a real value (between [0,1]) to filter the observed variant;
	 */
	public double computeFitnessWithRatioThreshold(int constraintVariantNum, boolean[] model, double[] log, double threshold){
		double fitness = 0;
		double satisfiedVariantNumInModel = 0; //correspond to |{v in M_{a,b} | L_{a,b}(v) >= n}|
		double satisfiedVariantNum = 0;//correspond to |{v in V | L_{a,b}(v) >= n}|
		double logSize = 0;
		for(int i=0; i<constraintVariantNum; i++){
			logSize = logSize + log[i];
		}
		for(int i=0; i<constraintVariantNum; i++){
			if((log[i]/logSize) >= threshold){
				satisfiedVariantNum ++;
				if(model[i]){
					satisfiedVariantNumInModel ++;					
				}
			}
			
		}	
		if(satisfiedVariantNum == 0){
			fitness = 1;
		}
		else{
			fitness = satisfiedVariantNumInModel/satisfiedVariantNum;
		}		
		return fitness;
	}
	
	/**
	 * compute the fitness (three) based on formula: |M_{a,b}|  /  |M_{a,b} + {v in V | L_{a,b}(v) >= n}|;
	 * M_{a,b} is the set of allowed constraint instance variants by the input model;
	 * L_{a,b}(v) indicate the number of reference events in activity a corresponding to variant v;
	 * V is the set of constraint instance variants;
	 * input constraintVariantNum: the number of constraint instance variants;
	 * input model: a set of (9) values in which one is true or false (i.g., 
	 * the third value is true, it means the third variant is allowed by the model);
	 * input log: a set of (9) values in which one is the number of corresponding instances;
	 * input threshold: a positive value to filter the observed variant
	 */
	public double computeFitnessHalfWithNumThreshold(int constraintVariantNum, boolean[] model, double[] log, int threshold){
		double fitness = 0;
		double variantNumInModel = 0; //correspond to |M_{a,b}|
		double satisfiedVariantNumInLog = 0;//correspond to |M_{a,b} + {v in V | L_{a,b}(v) >= n}|

		for(int i=0; i<constraintVariantNum; i++){
			if(model[i]){
				variantNumInModel ++;					
			}			
			if(model[i] || log[i] >= threshold){
				satisfiedVariantNumInLog ++;
			}			
		}
		if(satisfiedVariantNumInLog == 0){
			fitness = 1;
		}
		else{
			fitness = variantNumInModel/satisfiedVariantNumInLog;
		}			
		return fitness;
	}
	
	/**
	 * compute the precision (one) based on formula: |{v in M_{a,b} | L_{a,b}(v) >= n}|  /  |M_{a,b}|;
	 * M_{a,b} is the set of allowed constraint instance variants by the input model;
	 * L_{a,b}(v) indicate the number of reference events in activity a corresponding to variant v;
	 * input constraintVariantNum: the number of constraint instance variants;
	 * input model: a set of (9) values in which one is true or false (i.g., 
	 * the third value is true, it means the third variant is allowed by the model);
	 * input log: a set of (9) values in which one is the number of corresponding instances;
	 * input threshold: a positive value to filter the observed variant
	 */
	public double computePrecisionWithNumThreshold(int constraintVariantNum, boolean[] model, double[] log, int threshold){
		double precision = 0;
		double allowedVariantNumByModel = 0;
		double satisfiedVariantNumInLog = 0;
		for(int i=0; i<constraintVariantNum; i++){
			if(model[i]){
				allowedVariantNumByModel ++;
				if(log[i] >= threshold){
					satisfiedVariantNumInLog ++;
				}
			}
		}
		if(allowedVariantNumByModel == 0){
			precision = 1;
		}
		else{
			precision = satisfiedVariantNumInLog/allowedVariantNumByModel;
		}			
		return precision;
	}
	
	/**
	 * compute the precision (two) based on formula: |{v in M_{a,b} | L_{a,b}(v) >= n%}|  /  |M_{a,b}|;
	 * M_{a,b} is the set of allowed constraint instance variants by the input model;
	 * L_{a,b}(v) indicate the number of reference events in activity a corresponding to variant v;
	 * input constraintVariantNum: the number of constraint instance variants;
	 * input model: a set of (9) values in which one is true or false (i.g., 
	 * the third value is true, it means the third variant is allowed by the model);
	 * input log: a set of (9) values in which one is the number of corresponding instances;
	 * input threshold: a real value (between [0,1]) to filter the observed variant
	 */
	public double computePrecisionWithRatioThreshold(int constraintVariantNum, boolean[] model, double[] log, double threshold){
		double precision = 0;
		double allowedVariantNumByModel = 0;
		double satisfiedVariantNumInLog = 0;
		double logSize = 0;
		for(int i=0; i<constraintVariantNum; i++){
			logSize = logSize + log[i];
		}
		for(int i=0; i<constraintVariantNum; i++){
			if(model[i]){
				allowedVariantNumByModel ++;
				if((log[i]/logSize) >= threshold){
					satisfiedVariantNumInLog ++;
				}
			}
		}
		if(allowedVariantNumByModel == 0){
			precision = 1;
		}
		else{
			precision = satisfiedVariantNumInLog/allowedVariantNumByModel;
		}	
		return precision;
	}
	
	/**
	 * compute the precision (three) based on formula: (all v in M_{a,b}, accumulate -p*log_{2}(p)) /  log_{2}(|M_{a,b}|);
	 * M_{a,b} is the set of allowed constraint instance variants by the input model;
	 * L_{a,b}(v) indicate the number of reference events in activity a corresponding to variant v;
	 * input constraintVariantNum: the number of constraint instance variants;
	 * input model: a set of (9) values in which one is true or false (i.g., 
	 * the third value is true, it means the third variant is allowed by the model);
	 * input log: a set of (9) values in which one is the number of corresponding instances;
	 */
	public double computePrecisionWithEntropy(int constraintVariantNum, boolean[] model, double[] log){
		double precision = 0;
		double entropy = 0;
		double allowedVariantNumByModel = 0;
		double logSizeFittingModel = 0;//corresponding to L_{a,b}(M_{a,b})
		for(int i=0; i<constraintVariantNum; i++){
			if(model[i]){
				allowedVariantNumByModel ++;
				logSizeFittingModel = logSizeFittingModel + log[i];
			}			
		}
		for(int i=0; i<constraintVariantNum; i++){
			if(model[i]){
				double ratio = log[i]/logSizeFittingModel;
				if(ratio != 0){
					entropy = entropy + (-ratio*(Math.log(ratio)/Math.log(2)));
				}								
			}
		}
		if(Math.log(allowedVariantNumByModel) == 0){
			precision = 1;
		}
		else{
			precision = entropy/(Math.log(allowedVariantNumByModel)/Math.log(2));
		}	
		
		return precision;
	}
	
	/**
	 * compute fitness taking model variants and log variants as input;
	 * @param precisionThreshold: the threshold to filter log variant;
	 * @param constraintVariantNum: the number of constraint variants, i.e., current 9;
	 * @param modelVariantArray: a set of (9) values in which one is true or false (i.g., the third value is true, it means the third variant is allowed by the model);
	 * @param logVariantArray: a set of (9) values in which one is the number of corresponding instances;
	 * @param algorithm: indicate which algorithm is used to compute the precision
	 * @return
	 */
	public double computeFitnessForModelAndLog(double fitnessThreshold, int constraintVariantNum, 
			boolean[] modelVariantArray, double[] logVariantArray, String algorithm){
		double fitness = 0;
		if(algorithm.equals(mxConstants.fitnessAlgorithmWithoutThreshold)){
			fitness = computeFitnessBasic(constraintVariantNum, modelVariantArray, logVariantArray);
		}
		else if(algorithm.equals(mxConstants.fitnessAlgorithmWithNumThreshold)){
			fitness = computeFitnessWithNumThreshold(constraintVariantNum, modelVariantArray, logVariantArray, (int) fitnessThreshold);
		}
		else if(algorithm.equals(mxConstants.fitnessAlgorithmWithRatioThreshold)){
			fitness = computeFitnessWithRatioThreshold(constraintVariantNum, modelVariantArray, logVariantArray,  fitnessThreshold);
		}
		else if(algorithm.equals(mxConstants.fitnessAlgorithmHalfWithThreshold)){
			fitness = computeFitnessHalfWithNumThreshold(constraintVariantNum, modelVariantArray, logVariantArray,  (int) fitnessThreshold);
		}
		return fitness;
	}
	
	/**
	 * compute precision taking model variants and log variants as input;
	 * @param precisionThreshold: the threshold to filter log variant;
	 * @param constraintVariantNum: the number of constraint variants, i.e., current 9;
	 * @param modelVariantArray: a set of (9) values in which one is true or false (i.g., the third value is true, it means the third variant is allowed by the model);
	 * @param logVariantArray: a set of (9) values in which one is the number of corresponding instances;
	 * @param algorithm: indicate which algorithm is used to compute the precision
	 * @return
	 */
	public double computePrecisionForModelAndLog(double precisionThreshold, int constraintVariantNum, 
			boolean[] modelVariantArray, double[] logVariantArray, String algorithm){
		double precision = 0;
		if(algorithm.equals(mxConstants.precisionAlgorithmWithNumThreshold)){
			precision = computePrecisionWithNumThreshold(constraintVariantNum, modelVariantArray, logVariantArray, (int) precisionThreshold);
		}
		else if(algorithm.equals(mxConstants.precisionAlgorithmWithRatioThreshold)){
			precision = computePrecisionWithRatioThreshold(constraintVariantNum, modelVariantArray, logVariantArray, precisionThreshold);
		}
		else if(algorithm.equals(mxConstants.precisionAlgorithmByEntropy)){
			precision = computePrecisionWithEntropy(constraintVariantNum, modelVariantArray, logVariantArray);
		}
		return precision;
	}

	
	/**
	 * compute generalization (one) based on formula: F(p,n,f)=(all k in [0,n-f], accumulate C_n^k * p^k * q^(n-k));
	 * p: the probability fitting the model (fitModelProbability);
	 * n: the number of total instances (logSize);
	 * f: the number of violating instances (violatingCasetNum);
	 */
//	public double computeGeneralizationF(double fitModelProbability, int logSize, int violatingCasetNum){
//		ProbabilityFunction computingProbability = new ProbabilityFunction();
//		double generalization = 0;		
//		for(int k=0; k<=logSize-violatingCasetNum; k++){
////			System.out.printf("  circulation: " + k);
////			System.out.printf("  logSize: " + logSize);
//			double combinationValue = computingProbability.computeCombination(logSize, k);
//			generalization = generalization + combinationValue*Math.pow(fitModelProbability, k)*Math.pow(1-fitModelProbability, logSize-k);
////			System.out.printf("  combinationValue: " + combinationValue);
////			System.out.printf("  generalization: " + generalization);
//		}
//		BinomialDistribution binomialDistribution = new BinomialDistribution(logSize, fitModelProbability);
//		binomialDistribution.cumulativeProbability(logSize-violatingCasetNum);
//		return generalization;
//	}
	
	/**
	 * compute generalization taking model and log as input;
	 * fitModelProbability: the probability fitting the model (fitModelProbability);
	 * input constraintVariantNum: the number of constraint instance variants;
	 * input model: a set of (9) values in which one is true or false (i.g., 
	 * the third value is true, it means the third variant is allowed by the model);
	 * input log: a set of (9) values in which one is the number of corresponding instances;
	 * @param fitModelProbability
	 * @param constraintVariantNum
	 * @param model
	 * @param log
	 * @param algrithm : binomial distribution or beta distribution
	 * @return
	 */
	public double computeGeneralizationForModelAndLog(double fitModelProbability, int constraintVariantNum, boolean[] model, double[] log, String algorithm){
		double generalization = 0;	
		int logSize = 0;
		int logSizeViolatingModel = 0;//corresponding to L_{a,b}(V\M_{a,b})
		for(int i=0; i<constraintVariantNum; i++){
			logSize = (int) (logSize + log[i]);
		}
		
		for(int i=0; i<constraintVariantNum; i++){
			if(!model[i]){
				logSizeViolatingModel = (int) (logSizeViolatingModel + log[i]);
			}			
		}
		if(algorithm.equals(mxConstants.generalizationAlgorithmBinomialDistribution)){
			generalization = computeGeneralizationByBinomialDistribution(fitModelProbability, logSize, logSizeViolatingModel);
		}
		else if(algorithm.equals(mxConstants.generalizationAlgorithmBetaDistribution)){
			generalization = computeGeneralizationByBetaDistribution(fitModelProbability, logSize, logSizeViolatingModel, 1, 1);
		}
		return generalization;
	}
	
	/**
	 * compute generalization (the first method) based on formula: F(p,n,f)=(all k in [0,n-f], accumulate C_n^k * p^k * q^(n-k));
	 * p: the probability fitting the model (fitModelProbability);
	 * n: the number of total instances (logSize);
	 * f: the number of violating instances (violatingCasetNum);
	 */
	public double computeGeneralizationByBinomialDistribution(double fitModelProbability, int logSize, int violatingCasetNum){
		BinomialDistribution binomialDistribution = new BinomialDistribution(logSize, fitModelProbability);
		double generalization = binomialDistribution.cumulativeProbability(logSize-violatingCasetNum);
		return generalization;
	}
	
	/**
	 * compute generalization (the second method) based on formula: X=Beta(1+n-f,1+f),P(X > p(e.g., 0.95));
	 * p: the probability fitting the model (fitModelProbability), e.g., 0.95;
	 * n: the number of total instances (logSize);
	 * f: the number of violating instances (violatingCasetNum);
	 */
	public double computeGeneralizationByBetaDistribution(double fitModelProbability, int logSize, int violatingCasetNum, double alhpa, double beta){
		BetaDistribution betaDistribution = new BetaDistribution(alhpa+logSize-violatingCasetNum, beta+violatingCasetNum);		
		double generalization = 1 - betaDistribution.cumulativeProbability(fitModelProbability);			
		return generalization;
	}
	
	
	
	/**
	 * align a string to have a fixed length with space
	 */
	public String alignStringLength(int length, String inputString){
		String alignedString = inputString;
		int originalLength = inputString.length();
		String addedString = "";
		for(int i=0; i<length-originalLength; i++){
			addedString = addedString + " ";
		}
		return alignedString + addedString;
	}
	
	public static void main(String[] args) throws Exception{  	
		//the first value, such as true and 50, corresponds to the first variant (0,0)
		boolean[] modelA = {true,true,true,true,true,false,true,false,true};
		double[] logA = {50,20,2,20,5,0,2,0,1};
		
		boolean[] modelB = {true,true,false,true,true,false,false,false,false};
		double[] logB = {50,20,2,20,5,0,2,0,1};

		boolean[] modelC = {true,false,false,false,false,false,false,false,false};
		double[] logC = {50,20,2,20,5,0,2,0,1};

		boolean[] modelD = {false,false,false,false,true,true,false,true,true};
		double[] logD = {50,20,2,20,5,0,2,0,1};
		
		boolean[] modelE = {true,true,false,true,true,false,false,false,false};
		double[] logE = {5,2,0,2,0,0,0,0,1};
		
		boolean[] modelF = {true,true,false,true,true,false,false,false,false};
		double[] logF = {50,20,0,20,9,0,0,0,1};
		
		List<double[]> logList = new ArrayList<double[]>();
		logList.add(logA);
		logList.add(logB);
		logList.add(logC);
		logList.add(logD);
		logList.add(logE);
		logList.add(logF);
		
		List<boolean[]> modelList = new ArrayList<boolean[]>();
		modelList.add(modelA);
		modelList.add(modelB);
		modelList.add(modelC);
		modelList.add(modelD);
		modelList.add(modelE);
		modelList.add(modelF);
		
		MetricFormular computingMetrics = new MetricFormular();
		
		double precision = 0;
		double fitness = 0;
		double generalization = 0;
		int outputStringLength = 7;

		for(int i=0; i<6; i++){
			boolean[] model = modelList.get(i);
			double[] log = logList.get(i);					
			System.out.println("**********example-" + (i+1) + "***************");
			System.out.printf("index: ");
			String alignedLength; 
			for(int j=1; j<10; j++){				
				if(j == 9){
					alignedLength = j + ";";
					System.out.println(alignedLength);
				}
				else{
					alignedLength = j + ",";
					System.out.printf(computingMetrics.alignStringLength(outputStringLength, alignedLength));
				}				
			}
			System.out.printf("model: ");
			for(int j=0; j<model.length; j++){
				if(j == model.length-1){
					alignedLength = model[j] + ";";
					System.out.println(alignedLength);
				}
				else{
					alignedLength = model[j] + ",";
					System.out.printf(computingMetrics.alignStringLength(outputStringLength, alignedLength));
				}			
				
			}
			System.out.printf("log:   ");
			for(int j=0; j<log.length; j++){
				if(j == log.length-1){
					alignedLength = (int)log[j] + ";";
					System.out.println(alignedLength);
				}
				else{
					alignedLength = (int)log[j] + ",";
					System.out.printf(computingMetrics.alignStringLength(outputStringLength, alignedLength));
				}	
			}
			
			for(int numThreshold=1; numThreshold<14; numThreshold=numThreshold+4){
				for(double ratioThreshold=0.01; ratioThreshold<0.1; ratioThreshold=ratioThreshold+0.04){
					System.out.println("-----------parameter:" + numThreshold + "-" + ratioThreshold + "-------------");
					fitness = computingMetrics.computeFitnessBasic(9, model, log);
					System.out.println("fitness1: " + fitness);
					fitness = computingMetrics.computeFitnessWithNumThreshold(9, model, log, numThreshold);
					System.out.println("fitness2: " + fitness);
					fitness = computingMetrics.computeFitnessWithRatioThreshold(9, model, log, ratioThreshold);
					System.out.println("fitness3: " + fitness);
					fitness = computingMetrics.computeFitnessHalfWithNumThreshold(9, model, log, numThreshold);
					System.out.println("fitness4: " + fitness);
					
					precision = computingMetrics.computePrecisionWithNumThreshold(9, model, log, numThreshold);
					System.out.println("precision1: " + precision);
					precision = computingMetrics.computePrecisionWithRatioThreshold(9, model, log, ratioThreshold);
					System.out.println("precision2: " + precision);
					precision = computingMetrics.computePrecisionWithEntropy(9, model, log);
					System.out.println("precision3: " + precision);
					
//					generalization = computingMetrics.computeGeneralizationForModelAndLog(0.95, 9, model, log);
//					System.out.println("generalization1: " + generalization);
				}
			}			
		}
		
		System.out.println("***********computeGeneralizationByBinomialDistribution*********************");
		generalization = computingMetrics.computeGeneralizationByBinomialDistribution(0.95, 1, 0);
		System.out.println("generalization1: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBinomialDistribution(0.95, 10, 0);
		System.out.println("generalization2: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBinomialDistribution(0.95, 100, 0);
		System.out.println("generalization3: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBinomialDistribution(0.95, 1, 1);
		System.out.println("generalization4: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBinomialDistribution(0.95, 10, 1);
		System.out.println("generalization5: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBinomialDistribution(0.95, 100, 1);
		System.out.println("generalization6: " + generalization);
		
		System.out.println("***********computeGeneralizationByBetaDistribution*********************");
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 0, 0, 1, 1);
		System.out.println("generalization1: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 1, 0, 1, 1);
		System.out.println("generalization2: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 1, 1, 1, 1);
		System.out.println("generalization3: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 2, 1, 1, 1);
		System.out.println("generalization4: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 10, 0, 1, 1);
		System.out.println("generalization5: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 10, 10, 1, 1);
		System.out.println("generalization6: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 20, 10, 1, 1);
		System.out.println("generalization7: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 100, 0, 1, 1);
		System.out.println("generalization8: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 100, 100, 1, 1);
		System.out.println("generalization9: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 200, 100, 1, 1);
		System.out.println("generalization10: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 100, 1, 1, 1);
		System.out.println("generalization11: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 100, 5, 1, 1);
		System.out.println("generalization12: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 100, 10, 1, 1);
		System.out.println("generalization13: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 1000, 1, 1, 1);
		System.out.println("generalization14: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 1000, 5, 1, 1);
		System.out.println("generalization15: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 1000, 10, 1, 1);
		System.out.println("generalization16: " + generalization);
		generalization = computingMetrics.computeGeneralizationByBetaDistribution(0.95, 1000, 50, 1, 1);
		System.out.println("generalization17: " + generalization);
    }		
}
