/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.io;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.InputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import javax.swing.JOptionPane;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.deckfour.xes.util.XTimer;
import org.deckfour.xes.util.XsDateTimeConversion;
import org.deckfour.xes.util.XsDateTimeConversionJava7;
import org.processmining.common.runningtime.RunningTime;
import org.processmining.common.structure.IntegerPair;
import org.processmining.framework.util.Pair;
import org.processmining.ocbc.log.extension.XExtension;
import org.processmining.ocbc.log.factory.XFactory;
import org.processmining.ocbc.log.factory.XFactoryRegistry;
import org.processmining.ocbc.log.model.XAttribute;
import org.processmining.ocbc.log.model.XOCElement;
import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.log.model.XOCLog;
import org.processmining.ocbc.log.model.XOCORelation;
import org.processmining.ocbc.log.model.XOCObject;
import org.processmining.ocbc.log.model.XOCObjectModel;
import org.processmining.ocbc.log.model.impl.XOCEventImpl;
import org.processmining.ocbc.log.model.impl.XOCLogImpl;
import org.processmining.ocbc.log.model.impl.XOCORelationImpl;
import org.processmining.ocbc.log.model.impl.XOCObjectImpl;
import org.processmining.ocbc.log.model.impl.xConstants;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Parser for the XES XML serialization.
 * the object models for different events are different instances.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * 
 */
public class XocXmlParser extends XParser {

	protected XsDateTimeConversion xsDateTimeConversion = new XsDateTimeConversionJava7();
	private  XOCLog log;
//	private Map<String,XOCObject> idObjectMap = new HashMap<String,XOCObject>();
	public static XOCEvent event;
	public static XOCObjectModel objectModel;	
	public static Pair<String, String> objectAttribute;	
	public static XAttribute attribute = null;	
	public static List<XOCObject> objectList;	
	public static List<XOCORelation> objectRelationList;	
	public static XOCObject object;	
	private int eventNum;	
	public static XOCORelation objectRelation;
	
	//record the ids and corresponding objects (which are existent in object models)
	private Map<String,XOCObject> idVSObjectMapForLog = new HashMap<String,XOCObject>();
	//record the ids and corresponding objects (which are referred by some event and inexistent in the object model of this event)
//	private Map<String,XOCObject> idVSInexistentObjectMapForLog = new HashMap<String,XOCObject>();
	private Map<String,XOCORelation> idVSObjectRelationMapForLog = new HashMap<String,XOCORelation>();

	
	private boolean inEventFlag = false;
	private boolean inReferencesFlag = false;
	private boolean inModelFlag = false;
	private boolean inObjectListFlag = false;
	private boolean inObjectFlag = false;
	private boolean inRelationListFlag = false;
	private boolean inRelationFlag = false;
	private boolean inStringFlag = false;
	
	private String logFormat;
	
	private boolean showMessageFlag = false;
	
	/**
	 * Unique URI for the format definition.
	 */
	protected static final URI XES_URI = URI
			.create("http://www.xes-standard.org/");

	/**
	 * XES model factory used to build model.
	 */
	protected XFactory factory;

	/**
	 * Creates a new parser instance.
	 * 
	 * @param factory
	 *            The XES model factory instance used to build the model from
	 *            the serialization.
	 */
	public XocXmlParser(XFactory factory) {
		this.factory = factory;
	}

	/**
	 * Creates a new parser instance, using the currently-set standard factory
	 * for building the model.
	 */
	public XocXmlParser() {
		this(XFactoryRegistry.instance().currentDefault());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.in.XParser#author()
	 */
	@Override
	public String author() {
		return "Christian W. Günther";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.in.XParser#canParse(java.io.File)
	 */
	@Override
	public boolean canParse(File file) {
		String filename = file.getName();
		return endsWithIgnoreCase(filename, ".xes");
		// String suffix = filename.substring(filename.length() - 3);
		// return suffix.equalsIgnoreCase("xes");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.in.XParser#description()
	 */
	@Override
	public String description() {
		return "Reads XES models from plain XML serializations";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.in.XParser#name()
	 */
	@Override
	public String name() {
		return "XES XML";
	}
	
	public XOCLog getLog() {		
		long uint = 1;	
		long startTimeTotal = System.currentTimeMillis(); 
		long startTime = System.currentTimeMillis(); 
		long durationTime = System.currentTimeMillis();
        
        
        durationTime = (System.currentTimeMillis() - startTime)/uint;
		startTime = System.currentTimeMillis();
		RunningTime.writeTimeInformationIntoFile("process log start: " + System.currentTimeMillis()/uint, false, "D:\\timeImportXOCLog.txt");
	
		//set id vs object (relation) map in log
		log.setIdVSObjectMap(idVSObjectMapForLog);
		log.setIdVSObjectRelationMap(idVSObjectRelationMapForLog);
		
		//set object (relation) list in log
//		log.getObjectList().addAll(idVSObjectMapForLog.values());//in this way, objects are not sorted by birth time
		log.getObjectRelationList().addAll(idVSObjectRelationMapForLog.values());
		    
		//create the last object model and set its id vs object map
		XOCEvent lastEvent = log.getLastEvent();
		String lastEventId = lastEvent.getAttributes().get(xConstants.ID).toString();
	    XOCObjectModel lastObjectModel = lastEvent.getObjectModel();
	    lastObjectModel.getObjectSet().addAll(idVSObjectMapForLog.values());
	    lastObjectModel.getObjectRelationSet().addAll(idVSObjectRelationMapForLog.values());
	    lastObjectModel.setIdVSObjectMap(idVSObjectMapForLog);
	    
	    //create the event id list and the index for each event
	    List<String> eventIdList =  log.getEventIdList();
	    for(int i=0; i<log.getAllEvents().size(); i++){
	    	XOCEvent event = log.getAllEvents().get(i);
	    	event.setOrderNum(i+1);
	    	String eventId = event.getAttributes().get(xConstants.ID).toString();
	    	eventIdList.add(eventId);
	    }
	    
//	    //set the default death event as the last event for each object relation
//	    for(XOCORelation objectRelation : lastObjectModel.getObjectRelationSet()){
//	    	objectRelation.setDeathEventId(lastEventId);    	    	
//	    }
	    
	    
        
        durationTime = (System.currentTimeMillis() - startTime)/uint;
		startTime = System.currentTimeMillis();
		RunningTime.writeTimeInformationIntoFile("process log end: " + System.currentTimeMillis()/uint, false, "D:\\timeImportXOCLog.txt");
		RunningTime.writeTimeInformationIntoFile("~~~~~~~~~~~~~~process log duration time: " + durationTime, false, "D:\\timeImportXOCLog.txt");		
        
		System.out.println("finish parsing!");
		System.out.println("log.getAllEvents().size(): " + log.getAllEvents().size());
		
		return log;
	}

	/**
	 * Parses a log from the given input stream, which is supposed to deliver an
	 * XOC log in XML representation.
	 * 
	 * @param is
	 *            Input stream, which is supposed to deliver an XES log in XML
	 *            representation.
	 * @return The parsed log.
	 */
	public List<XOCLog> parse(InputStream is) throws Exception {	
		
		long uint = 1;	
		long startTime = System.currentTimeMillis(); 
		long durationTime = System.currentTimeMillis();
		
		RunningTime.writeTimeInformationIntoFile("parse prepare start time: " + startTime/uint, false, "D:\\timeImportXOCLog.txt");
		
		BufferedInputStream bis = new BufferedInputStream(is);
		// set up a specialized SAX2 handler to fill the container
		XocXmlHandler handler = new XocXmlHandler();
		// set up SAX parser and parse provided log file into the container
		SAXParserFactory parserFactory = SAXParserFactory.newInstance();
		parserFactory.setNamespaceAware(false);
		SAXParser parser = parserFactory.newSAXParser();			
		
		durationTime = (System.currentTimeMillis() - startTime)/uint;
		startTime = System.currentTimeMillis();
		RunningTime.writeTimeInformationIntoFile("parse prepare  end time: " + System.currentTimeMillis()/uint, false, "D:\\timeImportXOCLog.txt");
		RunningTime.writeTimeInformationIntoFile("~~~~~~~~~~~~~~parse prepare duration time: " + durationTime, false, "D:\\timeImportXOCLog.txt");
		RunningTime.writeTimeInformationIntoFile("parse sentence start time: " + System.currentTimeMillis()/uint, false, "D:\\timeImportXOCLog.txt");
		
		
		parser.parse(bis, handler);		
		
		durationTime = (System.currentTimeMillis() - startTime)/uint;
		startTime = System.currentTimeMillis();
		RunningTime.writeTimeInformationIntoFile("parse sentence  end time: " + System.currentTimeMillis()/uint, false, "D:\\timeImportXOCLog.txt");
		RunningTime.writeTimeInformationIntoFile("~~~~~~~~~~~~~~parse sentence duration time: " + durationTime, false, "D:\\timeImportXOCLog.txt");

		bis.close();
		ArrayList<XOCLog> wrapper = new ArrayList<XOCLog>();
		
		return wrapper;
	}
	

	/**
	 * SAX handler class for XOC in XML representation.
	 * @author Christian W. Guenther (christian@deckfour.org)
	 * 
	 *(1)within each event, the object referred by the relation
	 *use the same address as itself. For different events, the 
	 *objects with the same id correspond to different addresses, which guarantee 
	 *the operations on one event will not influence another one
	 *(2)the same objects (i.e., with the same id) in the related objects and all 
	 *objects correspond to different addresses
	 * 
	 */
	protected class XocXmlHandler extends DefaultHandler {	
		
		/**
		 * Buffer for extensions.
		 */
		protected HashSet<XExtension> extensions;
		
		private String key;
		private String value;
		private List<XOCElement> elementList = new ArrayList<XOCElement>();
		/**
		 * Creates a new handler instance.
		 */
	
		public XocXmlHandler() {
						
			eventNum = 1;
			log = null;
			event = null;
		    objectModel = null;
		    objectList = null;
		    objectRelationList = null;		

		}

		@Override
		/**
		 * build the empty structure of attribute, object, object relation, object list, object relation list, event and log
		 */
		public void startElement(String uri, String localName, String qName,
				Attributes attributes) throws SAXException {
			String tagName = qName;

			// parse content
			if (tagName.equals("string")) {
				// attribute tag.
				key = attributes.getValue("key");
				value = attributes.getValue("value");					
//				System.out.println("attribute");
				attribute = factory.createAttributeLiteral(key, value, null);
			} 
			else if (tagName.equals("object")) {
				object = new XOCObjectImpl("null");
//				elementList.add(object);
				inObjectFlag = true;
			}
			else if (tagName.equals("relation")) {
				objectRelation = new XOCORelationImpl("null");
//				elementList.add(objectRelation);
				inRelationFlag = true;
			}
			else if (tagName.equals("references")) {
				objectList = new ArrayList<XOCObject>();
//				elementList.add(objectList);
				inReferencesFlag = true;
			}
			else if (tagName.equals("objects")) {
				objectList = new ArrayList<XOCObject>();
				inObjectListFlag = true;
			}
			else if (tagName.equals("relations")) {
				objectRelationList = new ArrayList<XOCORelation>();
				inRelationListFlag = true;
			}
			else if (tagName.equals("event")) {
				event = new XOCEventImpl();
				inEventFlag = true;
			}
			else if (tagName.equals("log")) {
				log = new XOCLogImpl();
			}		
		}

		@Override
		/**
		 * fill in the empty structure of 
		 * attribute: key and attribute 
		 * object: attributes (include id, class and other attributes)
		 * object relation: attributes (include id, relation, sourceobjectid and targetobjectid), source object, target object and relation type
		 * object list: a list of objects
		 * object relation list: a list of object relations 
		 * event:  reference object list, object list and object relation list
		 * and log: a list of events
		 */
		public void endElement(String uri, String localName, String qName)
				throws SAXException {						
			
			String tagName = qName;

			if (tagName.equals("string")) {
//			if (tagName.equals("string") && attribute.getKey().equals(xConstants.ID)) {
				
				if(object != null){
					object.getAttributes().put(attribute.getKey(), attribute);
//					object.addXOCAttribute(attribute.getKey(), attribute.get, eventOrderNum, logFormat);
				}
				else if(objectRelation != null){
					objectRelation.getAttributes().put(attribute.getKey(), attribute);
				}
				else if(event != null){
					event.getAttributes().put(attribute.getKey(), attribute);
				}	
				else if(log != null){
					System.out.println("add log attribute");
					log.getAttributes().put(attribute.getKey(), attribute);
					if(attribute.getKey().equals(xConstants.FORMAT)){
						logFormat = attribute.toString();
//						System.out.println("logFormat: " + logFormat);
					}
				}
			}
			else if (tagName.equals("object")) {
				
				String objectId = object.getAttributes().get(xConstants.ID).toString();						
				//if the object is in the object list (i.e., object model)
				if(inObjectListFlag){
					int eventOrderNum = log.size() + 1;	
					if(logFormat.equals(xConstants.TOTAL)){
//						System.out.println("logFormat.equals(xConstants.TOTAL)");
						parseObjectInTotalVariant(object, eventOrderNum);
					}
					else if(logFormat.equals(xConstants.UPDATE)){
//						System.out.println("logFormat.equals(xConstants.UPDATE)");
						parseObjectInUpdateVariant(object, eventOrderNum);
					}
			
				}
				
				//add the related event for each object (the reference must be behind the object model)
				if(inReferencesFlag){		
					//when the object first appear in the reference (inexistent in OM), 
					//add it into the complete object set and set birth event as null
					if(!idVSObjectMapForLog.keySet().contains(objectId)){						
//						object.setBirthEventId(null);
//						object.setDeathEventId(null);
						object.setReferenceEvent(event);
//						object.setBirthEvent(null);
//						object.setDeathEvent(null);
						idVSObjectMapForLog.put(objectId,object);
						log.getObjectList().add(object);
						
						object.getRelatedEventList().add(event);
						event.getSimplifiedRelatedObjects().add(object);
						event.getRelatedObjects().add(object);
					}
					else{
						XOCObject existentObject = idVSObjectMapForLog.get(objectId);
						if(existentObject.getReferenceEvent() == null){
							existentObject.setReferenceEvent(event);
						}
						existentObject.getRelatedEventList().add(event);
						event.getSimplifiedRelatedObjects().add(existentObject);
						event.getRelatedObjects().add(existentObject);
					}					
				}
				
				object = null;
				inObjectFlag = false;
				
			} 
			else if (tagName.equals("relation")) {
				int eventOrderNum = log.size() + 1;	
				if(logFormat.equals(xConstants.TOTAL)){
//					System.out.println("logFormat.equals(xConstants.TOTAL)");
					parseObjectRelationInTotalVariant(objectRelation, eventOrderNum);
				}
				else if(logFormat.equals(xConstants.UPDATE)){
//					System.out.println("logFormat.equals(xConstants.UPDATE)");
					parseObjectRelationInUpdateVariant(objectRelation, eventOrderNum);
				}
				objectRelation = null;
				inRelationFlag = false;								
			}

			else if (tagName.equals("references")) {
				objectList = null;	
				inReferencesFlag = false;
			} 
			else if (tagName.equals("objects")) {			
				if(log.getAttributes().keySet().contains(xConstants.FORMAT)){				
			    	String logFormat = log.getAttributes().get(xConstants.FORMAT).toString();			    	
					if(logFormat.equals(xConstants.TOTAL)){
						event.getObjectModel().setObjectSet(objectList);
					}
					else if(logFormat.equals(xConstants.UPDATE)){
						event.getObjectModelUpdate().setObjectSet(objectList);
					}
				}
				else{
					JOptionPane.showMessageDialog(null, "Wrong! The log should contain the format attribute");
				}				
				objectList = null;		
				inObjectListFlag = false;
			}
			else if (tagName.equals("relations")) {	
				if(log.getAttributes().keySet().contains(xConstants.FORMAT)){				
			    	String logFormat = log.getAttributes().get(xConstants.FORMAT).toString();			    	
					if(logFormat.equals(xConstants.TOTAL)){
						event.getObjectModel().setObjectRelationSet(objectRelationList);
						event.getObjectModel().updateObjectRelationSet();
					}
					else if(logFormat.equals(xConstants.UPDATE)){
						event.getObjectModelUpdate().setObjectRelationSet(objectRelationList);
						event.getObjectModelUpdate().updateObjectRelationSet();
					}
				}
				else{
					JOptionPane.showMessageDialog(null, "Wrong! The log should contain the format attribute");
				}	
				objectRelationList = null;	
				inRelationListFlag = false;
			}
			else if (tagName.equals("event")) {
				event.setOrderNum(eventNum);
				eventNum ++;	
				log.add(event);	
				event = null;	
				inEventFlag = false;
				System.out.println("eventOrderNum in endElement: " + log.size());
				/*
				 *within each event, the object referred by the relation
				 *use the same address as itself. For different events, the 
				 *objects with the same id correspond to different addresses, which guarantee 
				 *the operations on one event will not influence another one
				 */			
			}
						
		}				
	}

	/**
	 * parse (i.e., parse operation) an object in the object model in the total log variant;
	 * (1)create or get the object
	 * (2)create or update the lifecycle period
	 * (3)create new attribute versions
	 * note that the end event is not null
	 * @param object: indicates the object which is parsing
	 * @param eventOrderNum: indicates the event order number of the object model where the object belongs to
	 */
	public void parseObjectInTotalVariant(XOCObject object, int eventOrderNum){
		String objectId = object.getAttributes().get(xConstants.ID).toString();		
		//if the object is born (i.e., first appear in the object model)
		if(!idVSObjectMapForLog.keySet().contains(objectId)){
			
			//record the start event for the object and set the end event as the same event by default
			//the end event will be updated (e.g., +1) if the same object is observed in the next object model
			IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, eventOrderNum);
			object.getLifecyclePeriodList().add(lifecyclePair);
			
			//add a new object into the map and list of log
			idVSObjectMapForLog.put(objectId,object);
			log.getObjectList().add(object);
			objectList.add(object);	
			
			//add attribute versions	
			for(String attributeKey : object.getAttributes().keySet()){
				String attributeValue = object.getAttributes().get(attributeKey).toString();
				object.addXOCAttribute(attributeKey, attributeValue, eventOrderNum, logFormat);
			}
		}
		//if the object already exists in the map of log, it is born or referred before;
		else {
			XOCObject existentObject = idVSObjectMapForLog.get(objectId);
//			if(existentObject.getBirthEvent() == null){
//				existentObject.setBirthEvent(event);
//			}
			
			//if the object is just referred before, it has no lifecycle period. 
			//In other words, it is born in current object model, and add a lifecycle period.
			if(existentObject.getLifecyclePeriodList().size() == 0){
				//record the start event for the object
				IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, eventOrderNum);
				existentObject.getLifecyclePeriodList().add(lifecyclePair);
			}
			//if it is born before, it has existing lifecycle periods, and update the end event of the last lifecycle period
			else{
				List<IntegerPair> periodList = existentObject.getLifecyclePeriodList();
				int existingSecondValue = periodList.get(periodList.size()-1).getSecondValue();
				//if the event order number is next to the current end event order number, extend the period
				if(eventOrderNum == existingSecondValue + 1){
					periodList.get(periodList.size()-1).setSecondValue(eventOrderNum);
				}
				//if the event order number is not next to the current end event order number, build a new period
				else{
					IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, eventOrderNum);
					existentObject.getLifecyclePeriodList().add(lifecyclePair);
				}
			}
			
			//add attribute versions
			for(String attributeKey : object.getAttributes().keySet()){
				String attributeValue = object.getAttributes().get(attributeKey).toString();
				existentObject.addXOCAttribute(attributeKey, attributeValue, eventOrderNum, logFormat);
			}
		}	
	}
	
	/**
	 * parse (i.e., parse operation) an object in the object model in the update log variant;
	 * (1)create or get the object
	 * (2)create or update the lifecycle period
	 * (3)create new attribute versions
	 * note that the end event could be null
	 * @param object: indicates the object which is parsing
	 * @param eventOrderNum: indicates the event order number of the object model where the object belongs to
	 */
	public void parseObjectInUpdateVariant(XOCObject object, int eventOrderNum){
		String objectId = object.getAttributes().get(xConstants.ID).toString();
		//if the object is born (i.e., first appear in the object model), indicating the lifecycle attribute is "add"
		if(!idVSObjectMapForLog.keySet().contains(objectId)){
			
			//add a new object into the map and list of log
			idVSObjectMapForLog.put(objectId,object);
			log.getObjectList().add(object);
			objectList.add(object);								

			//record the start event for the object and set the end event as null
			IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, null);
			object.getLifecyclePeriodList().add(lifecyclePair);
			
			//record the life cycle attribute
			object.getEventOrderNumVSObjectStateMap().put(eventOrderNum, xConstants.add);
			
			//add attribute versions
			for(String attributeKey : object.getAttributes().keySet()){
				String attributeValue = object.getAttributes().get(attributeKey).toString();
				object.addXOCAttribute(attributeKey, attributeValue, eventOrderNum, logFormat);
			}
		}
		else {
			XOCObject existentObject = idVSObjectMapForLog.get(objectId);
			objectList.add(existentObject);	
			
			//the old logs of update variant do not have the life cycle attribute; show a message box only once
			if(!object.getAttributes().containsKey(xConstants.lifecycle)){
				if(!showMessageFlag){
					JOptionPane.showMessageDialog(null, "Warn! The imported log has no life cycle attributes!");
					showMessageFlag = true;
				}				
			}
			else{
				String lifecycle = object.getAttributes().get(xConstants.lifecycle).toString();
				//if the object is (re)born at the current event	
				if(lifecycle.equals(xConstants.add)){
//					existentObject.setBirthEvent(event);
					
					//record the start event for the object and set the end event as null
					IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, null);
					existentObject.getLifecyclePeriodList().add(lifecyclePair);
					
					//add attribute versions
					for(String attributeKey : object.getAttributes().keySet()){
						String attributeValue = object.getAttributes().get(attributeKey).toString();
						existentObject.addXOCAttribute(attributeKey, attributeValue, eventOrderNum, logFormat);
					}
					
					//record the life cycle attribute
					existentObject.getEventOrderNumVSObjectStateMap().put(eventOrderNum, xConstants.add);
				}
				//if the object died at the current event							
				else if(lifecycle.equals(xConstants.delete)){
//					existentObject.setDeathEvent(event);
//					existentObject.setDeathEventOrderNum(eventOrderNum);
					
					//record the delete event for the object through setting the last pair	
					List<IntegerPair> periodList = existentObject.getLifecyclePeriodList();
					periodList.get(periodList.size()-1).setSecondValue(eventOrderNum-1);
					
					//add the ending event order number for the last version of each attribute
					for(String attributeKey : existentObject.getXOCAttributes().keySet()){
						existentObject.endXOCAttribute(attributeKey, eventOrderNum, logFormat);
					}
					
					//record the life cycle attribute
					existentObject.getEventOrderNumVSObjectStateMap().put(eventOrderNum, xConstants.delete);
				}
				//if the object is updated at the current event	
				else if (lifecycle.equals(xConstants.update)){
					//add attribute versions
					for(String attributeKey : object.getAttributes().keySet()){
						String attributeValue = object.getAttributes().get(attributeKey).toString();
//						System.out.println("attributeKey in lifecycle.equals(xConstants.update): " + attributeKey);
//						System.out.println("attributeValue in lifecycle.equals(xConstants.update): " + attributeValue);
						existentObject.addXOCAttribute(attributeKey, attributeValue, eventOrderNum, logFormat);
					}
					
					//record the life cycle attribute
					existentObject.getEventOrderNumVSObjectStateMap().put(eventOrderNum, xConstants.update);
				}
			}
		}	
	}
	
	/**
	 * parse (i.e., parse operation) an object relation in the object model in the total log variant;
	 * (1)create or get the object relation
	 * (2)create or update the lifecycle period
	 * (3)create new attribute versions or do nothing
	 * note that the end event is not null
	 * @param relationId
	 * @param eventOrderNum
	 */
	public void parseObjectRelationInTotalVariant(XOCORelation objectRelation, int eventOrderNum){
		String relationId = objectRelation.getAttributes().get(xConstants.ID).toString();	
		//if the relation is born (i.e., first appear)
		if(!idVSObjectRelationMapForLog.containsKey(relationId)){			
			//record the relation
			objectRelationList.add(objectRelation);
			idVSObjectRelationMapForLog.put(relationId, objectRelation);
			
			//record the start event for the object and set the end event as the same event by default
			//the end event will be updated (e.g., +1) if the same object is observed in the next object model
			IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, eventOrderNum);
			objectRelation.getLifecyclePeriodList().add(lifecyclePair);
			
			//add additional information
			addAdditionalInformationForObjectRelation(objectRelation);
		}
		//if the relation is not born (is not first observed)
		else{
			//retrieve the object relation
			XOCORelation existentObjectRelation = idVSObjectRelationMapForLog.get(relationId);
			
			//update the end event of the last lifecycle period		
			List<IntegerPair> periodList = existentObjectRelation.getLifecyclePeriodList();
			int existingSecondValue = periodList.get(periodList.size()-1).getSecondValue();
			//if the event order number is next to the current end event order number, extend the period
			if(eventOrderNum == existingSecondValue + 1){
				periodList.get(periodList.size()-1).setSecondValue(eventOrderNum);
			}
			//if the event order number is not next to the current end event order number, build a new period
			else{
				IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, eventOrderNum);
				existentObjectRelation.getLifecyclePeriodList().add(lifecyclePair);
			}
		}
	}
	
	/**
	 * parse (i.e., parse operation) an object in the object model in the update log variant;
	 * (1)create or get the object
	 * (2)create or update the lifecycle period
	 * (3)create new attribute versions
	 * note that the end event could be null
	 * @param objectId
	 * @param eventOrderNum
	 */
	public void parseObjectRelationInUpdateVariant(XOCORelation objectRelation, int eventOrderNum){
		String relationId = objectRelation.getAttributes().get(xConstants.ID).toString();	
		//if the object relation is born (i.e., first appear in the object model), indicating the lifecycle attribute is "add"
		if(!idVSObjectRelationMapForLog.keySet().contains(relationId)){
			//add a new object relation into the map 
			objectRelationList.add(objectRelation);
			idVSObjectRelationMapForLog.put(relationId, objectRelation);							

			//record the start event for the object and set the end event as null
			IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, null);
			objectRelation.getLifecyclePeriodList().add(lifecyclePair);
			
//			System.out.println("relationId in parseObjectRelationInUpdateVariant: " + relationId);
			
//			System.out.println("eventOrderNum in parseObjectRelationInUpdateVariant: " + eventOrderNum);
			//record the life cycle attribute
			objectRelation.getEventOrderNumVSObjectRelationStateMap().put(eventOrderNum, xConstants.add);
			
			//add additional information
			addAdditionalInformationForObjectRelation(objectRelation);
		}
		else {
//			System.out.println("relationId in contains(relationId)){: " + relationId);
			
			XOCORelation existentObjectRelation = idVSObjectRelationMapForLog.get(relationId);
			objectRelationList.add(existentObjectRelation);	
			
			//the old logs of update variant do not have the life cycle attribute; show a message box only once
			if(!objectRelation.getAttributes().containsKey(xConstants.lifecycle)){
				if(!showMessageFlag){
					JOptionPane.showMessageDialog(null, "Warn! The imported log has no life cycle attributes!");
					showMessageFlag = true;
				}				
			}
			else{
				String lifecycle = objectRelation.getAttributes().get(xConstants.lifecycle).toString();
				//if the object is (re)born at the current event	
				if(lifecycle.equals(xConstants.add)){
					//record the start event for the object and set the end event as null
					IntegerPair lifecyclePair = new IntegerPair(eventOrderNum, null);
					existentObjectRelation.getLifecyclePeriodList().add(lifecyclePair);		
					
					//record the life cycle attribute
					existentObjectRelation.getEventOrderNumVSObjectRelationStateMap().put(eventOrderNum, xConstants.add);
					
					System.out.println("relationId in lifecycle.equals(xConstants.add): " + relationId);
					System.out.println("eventOrderNum in lifecycle.equals(xConstants.add): " + eventOrderNum);
				}
				//if the object died at the current event							
				else if(lifecycle.equals(xConstants.delete)){
					//record the delete event for the object through setting the last pair	
					List<IntegerPair> periodList = existentObjectRelation.getLifecyclePeriodList();
					periodList.get(periodList.size()-1).setSecondValue(eventOrderNum-1);
					
					//record the life cycle attribute
					existentObjectRelation.getEventOrderNumVSObjectRelationStateMap().put(eventOrderNum, xConstants.delete);
					
					System.out.println("relationId in lifecycle.equals(xConstants.delete): " + relationId);
					System.out.println("eventOrderNum in lifecycle.equals(xConstants.delete): " + eventOrderNum);
				}
				//if the object is updated at the current event	
				else if (lifecycle.equals(xConstants.update)){
					//current, the object relation cannot be updated; reserve this part for future

					
				}
			}		
		}
	}
	
	/**
	 * add the additional information (attributes) for one object relation and its related two objects
	 * @param objectRelation
	 */
	public void addAdditionalInformationForObjectRelation(XOCORelation objectRelation){
		String sourceObjectId = objectRelation.getAttributes().get("sourceobjectid").toString();
		String targetObjectId = objectRelation.getAttributes().get("targetobjectid").toString();
		String relationType = objectRelation.getAttributes().get("relation").toString();
		
		XOCObject sourceObject = idVSObjectMapForLog.get(sourceObjectId);
		XOCObject targetObject = idVSObjectMapForLog.get(targetObjectId);
		
		//in future, remove the condition if it takes time
		if(!sourceObject.getRelatedObjectIdList().contains(targetObjectId)){
			sourceObject.getRelatedObjectIdList().add(targetObjectId);
		}
		if(!targetObject.getRelatedObjectIdList().contains(sourceObjectId)){
			targetObject.getRelatedObjectIdList().add(sourceObjectId);
		}
		
		//add the related object relations to the object
		if(!sourceObject.getRelatedObjectRelationList().contains(objectRelation)){
			sourceObject.getRelatedObjectRelationList().add(objectRelation);
		}
		if(!targetObject.getRelatedObjectRelationList().contains(objectRelation)){
			targetObject.getRelatedObjectRelationList().add(objectRelation);
		}		
		
		//in future, remove the following sentences if it takes time
		objectRelation.setSourceObject(sourceObject);
		objectRelation.setTargetObject(targetObject);
		objectRelation.setRelationType(relationType);
	}
	
	public static void main(String[] str) throws Exception{	
		XTimer timer = new XTimer();
		XocXmlParser parser = new XocXmlParser();
		
		long uint = 1;	
		long startTimeTotal = System.currentTimeMillis(); 
		long startTime = System.currentTimeMillis(); 
		long durationTime = System.currentTimeMillis();
		
		RunningTime.writeTimeInformationIntoFile("************start parsing************", false, "D:\\timeImportXOCLog.txt");
//		RunningTime.writeTimeInformationIntoFile("model discovery start time: " + startTime/uint, false, "D:\\timeImportXOCLog.txt");
		
		parser.parse(new File("d:\\log.xoc"));
		
		
		durationTime = (System.currentTimeMillis() - startTime)/uint;
		startTime = System.currentTimeMillis();
		RunningTime.writeTimeInformationIntoFile("parse log  end time: " + System.currentTimeMillis()/uint, false, "D:\\timeImportXOCLog.txt");
		RunningTime.writeTimeInformationIntoFile("~~~~~~~~~~~~~~parse log duration time: " + durationTime, false, "D:\\timeImportXOCLog.txt");
		RunningTime.writeTimeInformationIntoFile("process log start time: " + System.currentTimeMillis()/uint, false, "D:\\timeImportXOCLog.txt");
		
		
		XOCLog log = parser.getLog();
        
        durationTime = (System.currentTimeMillis() - startTime)/uint;
		startTime = System.currentTimeMillis();
		RunningTime.writeTimeInformationIntoFile("process log  end time: " + System.currentTimeMillis()/uint, false, "D:\\timeImportXOCLog.txt");
		RunningTime.writeTimeInformationIntoFile("~~~~~~~~~~~~~~process log duration time: " + durationTime, false, "D:\\timeImportXOCLog.txt");
		
		durationTime = (System.currentTimeMillis() - startTimeTotal)/uint;		
		RunningTime.writeTimeInformationIntoFile("--------------log size: " + log.getAllEvents().size(), false, "D:\\timeImportXOCLog.txt");
		RunningTime.writeTimeInformationIntoFile("--------------total time: " + durationTime, false, "D:\\timeImportXOCLog.txt");
		
		
		
//		System.out.print("\n log size" + log.getAllEvents().size());
		
//		for(XOCEvent event : log.getAllEvents()){
//
//			System.out.print("\n *********");
//			System.out.print("\n event id: "+ event.getAttributes().get("id"));
//			System.out.print("\n activity: "+ event.getAttributes().get("activity"));
//			
//			System.out.print("\n relatedobjects size"+ event.getRelatedObjects().size());
//			for(XOCObject object : event.getRelatedObjects()){	
//				System.out.print("\n related object name: "+ object.getAttributes().get("id"));
//
//			}
//			
//			
//		}
//		
//		System.out.print("\n *****object****");
//		
//		XOCEvent event = log.getLastEvent();
//		
//		System.out.print("\n objects size"+ event.getObjectModel().getObjectSet().size());
//		for(XOCObject object : event.getObjectModel().getObjectSet()){
//			System.out.print("\n -----------------");
//			System.out.print("\n object id: "+ object.getAttributes().get("id"));
//			System.out.print("\n object class: "+ object.getAttributes().get("class"));
//			System.out.print("\n birth event: "+ object.getBirthEventId());
//			for(XOCEvent relatedEvent : object.getRelatedEventList()){
//				System.out.print("\n related event: "+ relatedEvent.getAttributes().get("id"));
//			}
//			for(String relatedObjectId : object.getRelatedObjectIdList()){
//				System.out.print("\n related object: "+ relatedObjectId);
//			}
//		}
//		
//		System.out.print("\n relations size"+ event.getObjectModel().getObjectRelationSet().size());
//		for(XOCORelation relation : event.getObjectModel().getObjectRelationSet()){
//			System.out.print("\n sourceobjectid: "+ relation.getAttributes().get("sourceobjectid"));
//			System.out.print("\n targetobjectid: "+ relation.getAttributes().get("targetobjectid"));
//			System.out.print("\n id: "+ relation.getAttributes().get("id"));
//			System.out.print("\n relation: "+ relation.getAttributes().get("relation"));
//		}
	}
}
