/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model;

import java.util.List;

import org.processmining.common.structure.IntegerPair;
import org.processmining.framework.util.Pair;

/**
 * This interface defines attributes used for describing
 * meta-information about event log hierarchy elements.
 * 
 * Attributes have a name (i.e., a key), which is string-based. 
 * The value of an attribute is strongly typed, and can be
 * accessed and modified via sub-interface methods specified
 * by type.
 * 
 * Attributes may further be defined by an extension,
 * which makes it possible to assign semantic meaning to
 * them within a specific domain.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 */
public interface XOCAttribute {

	/**
	 * Retrieves the key, i.e. unique identifier, of this attribute.
	 * Currently, we use the key to indicate the attribute name
	 * @return The key of this attribute, as a string.
	 */
	public String getKey();

	/**
	 * get the value list
	 * @return
	 */
	public List<String> getValueList();
	
	/**
	 * get the current value (the last value in the value list)
	 * @return
	 */
	public String getCurrentValue();
	
	/**
	 * return the value version corresponding to the input event order number;
	 * return null if there is no value at the event order number;
	 * @param eventOrderNum
	 * @return
	 */
	public String getValueVersion(int eventOrderNum);
	
	/**
	 * add a new value version into the corresponding list
	 * @param value
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void insertNewValueVersion(String value, int eventOrderNum, String logFormat);
	
	/**
	 * add a value version into the corresponding list
	 * if the value equals the current value, update current value version
	 * otherwise, insert a new value version
	 * @param value
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void addValueVersion(String value, int eventOrderNum, String logFormat);
	
	/**
	 * add an ending event order number (eventOrderNum-1) for the last value version
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void endLastValueVersion(int eventOrderNum, String logFormat);
	
	/**
	 * update the current value version
	 */
	public void updateCurrentValueVersion();
	
	/**
	 * generate the pair list for one input value, i.e., all survival periods of one value 
	 * @return
	 */
	public List<Pair<Integer, Integer>> getSurvivalPeriodListForValue(String value);
	
	/**
	 * generate the pair (IntegerPair) list for one input value, i.e., all survival periods of one value ;
	 * @return
	 */
	public List<IntegerPair> getSurvivalPeriodListOfValue(String value);
	
	/**
	 * get the period list
	 * @return
	 */
	public List<Pair<Integer, Integer>> getPeriodList();
	
	/**
	 * Attributes must be cloneable.
	 * @return A clone of this attribute.
	 */
//	public Object clone();
	
}