/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model;

import java.util.List;
import java.util.Map;

/**
 * A log is an element of an XES event log structure. Logs are contained in
 * archives. Any log is a list of traces.
 * 
 * Logs represent a collection of traces, which are all representing executions
 * of the same kind of process.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * 
 */
public interface XOCLog extends XOCElement, List<XOCEvent> {

	/**
	 * This method returns a list of attributes which are global for all traces,
	 * i.e. every trace in the log is guaranteed to have these attributes.
	 * 
	 * @return List of ubiquitous trace attributes.
	 */
	public List<XAttribute> getGlobalEventAttributes();

	/**
	 * This method returns a list of attributes which are global for all events,
	 * i.e. every event in the log is guaranteed to have these attributes.
	 * 
	 * @return List of ubiquitous event attributes.
	 */
	public List<XAttribute> getGlobalObjectModelAttributes();
	
	public List<XAttribute> getGlobalObjectRelationAttributes();
	
	public List<XAttribute> getGlobalObjectAttributes();
	
	public List<XOCEvent> getAllEvents();
	
	public List<String> getEventIdList();
	
	/*
	 * get all events of one activity
	 */
	public List<XOCEvent> getEventsOfActivity(String activityName);
	
	public Map<String, XOCEvent> getIdVSEventMap();
	
	public XOCEvent getLastEvent();
	
	public void setAllEvents(List<XOCEvent> allEvents);
	
	//with sorting activities as frequency
	public List<String> createActivityNames();
	
	public List<String> generateActivityNameList();
	
	public Map<String, Integer> createActivityVSNumMap();
	
	public List<XOCEvent> createStartEvents();
	
	public List<String> createStartActivityNames();
	
	public Map<String, Integer> createStartActivityVSNumMap();
	
	public List<XOCEvent> createEndEvents();
	
	public List<String> createEndActivityNames();
	
	public Map<String, Integer> createEndActivityVSNumMap();
	
	public List<XOCEvent> createEventsForOneActivity(String activity);
	
	public List<XOCObject> getAllObjectVersions();

	/**
	 * generate and get the list of class names;
	 * it is time-consuming when it is called multiple times
	 */
	public List<String> getClassNames();
	
	/**
	 * generate the list of class names
	 */
	public List<String> generateClassNameList();
	
	/**
	 * get the list of class names after generating
	 * (should be used after executing the generateClassNameList() function,
	 * otherwise, the list is of size 0)
	 */
	public List<String> getClassNameList();
	
	public List<String> getResources();
	
	public List<XOCRRelation> getReferenceRelations();
	
	//accumulate (overlap) objects in each object model
	public List<XOCObject> getRepeatableObjects();
	
	/*
	 * get all related event of one event by a class node or relation
	 */
	public List<XOCEvent> getRelatedEventsForOneEvent(XOCEvent event);
	
	/*
	 * get all related event by a class node of one event
	 */
	public List<XOCEvent> getRelatedEventsForOneEventA(XOCEvent event, String classNodeName);
	
	/*
	 * get all related event by two class nodes of a class relation of one event
	 */
	public List<XOCEvent> getRelatedEventsForOneEventB(XOCEvent event, String classNodeNameX, String classNodeNameY);
	
	//get all objects connected to related object of an event
	public List<XOCObject> getIndirectRelatedObjectsForOneEvent(XOCEvent event);
	
	/**
	 * get all objects connected to related objects of an event + related objects of an event
	 * @param event
	 * @return
	 */
	public List<XOCObject> getAllRelatedObjectsForOneEvent(XOCEvent event);
	
	//get the source object of an object relation
	public XOCObject getSourceObjectOfObjectRelation(XOCObjectModel objectModel, XOCORelation objectRelation);
	
	//get the target object of an object relation
	public XOCObject getTargetObjectOfObjectRelation(XOCObjectModel objectModel, XOCORelation objectRelation);	
	
	public Object clone();
	
	/**
	 *  create the complete object model for one input event and
	 *  fill it into the object model parameter
	 * @param event
	 * @return complete object model
	 */
	public XOCObjectModel createCompleteObjectModel(XOCEvent event);
	
	/**
	 *  create the updated parts of the object model for one input event 
	 * @param event
	 * @return complete object model
	 */
	public XOCObjectModel createObjectModelUpdate(XOCEvent event);
	
	//set and get id vs object (relation) mapping
	public Map<String,XOCObject> getIdVSObjectMap();
	
	public Map<String,XOCORelation> getIdVSObjectRelationMap();
	
	public void setIdVSObjectMap(Map<String,XOCObject> idVSObject);
	
	public void setIdVSObjectRelationMap(Map<String,XOCORelation> idVSObjectRelation);
	
	/**
	 * get all objects (maybe only containing object id)
	 * @return
	 */
	public List<XOCObject> getObjectList();
	
	/**
	 * get all  object relations
	 * @return
	 */
	public List<XOCORelation> getObjectRelationList();
	
	/**
	 * get target events correlated to an input event by a path (an  array of class names)
	 * @param event
	 * @param path
	 * @return
	 */
	public List<XOCEvent> getRelatedEventsForOneEventByPath(XOCEvent event, String[] path, String targetActivity);
	
	/**
	 * get target events correlated to an input object by a path (an array of class names);
	 * the first class in the path is the class of the input object;
	 * @param object
	 * @param path
	 * @param targetActivity
	 * @return
	 */
	public List<XOCEvent> getRelatedEventsForOneObjectByPath(XOCObject object, String[] path, String targetActivity);
	
	/**
	 * sort an input list of events based on their orders in the event log
	 * @param eventList
	 * @return
	 */
	public List<XOCEvent> sortEventList(List<XOCEvent> eventList);
}
