/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model;

import java.util.List;
import java.util.Map;

import org.processmining.common.structure.IntegerPair;
import org.processmining.ocbc.models.MyOriginalElement;



/**
 * An event is an element of an XES event log structure. Events are sequentially
 * contained in traces.
 * 
 * Events refer to something that has happened during the execution of a
 * process, e.g. the execution of an activity.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * 
 */
public interface XOCObject extends MyOriginalElement, XOCElement {

	/**
	 * Returns the id of the event
	 * 
	 * @return the ID of this event
	 */
//	XID getID();

//	public XOCClass getClassType();
		
	public String getName();
	
//	public String getBirthEventId();
//	
//	public String getDeathEventId();
	
	/**
	 * get the event when the object is first referred to
	 * @return
	 */
	public XOCEvent getReferenceEvent();
	
//	public XOCEvent getBirthEvent();
//	
//	public XOCEvent getDeathEvent();
//	
//	/**
//	 * Get the death event order number i (i starts from 1, i.e., corresponding event is log.get(i-1))
//	 */
//	public int getDeathEventOrderNum();
	
	public List<XOCEvent> getRelatedEventList();
	
	public List<String> getRelatedObjectIdList();
	
	public List<XOCORelation> getRelatedObjectRelationList();
	
//	public void setClassType(XOCClass Class);
	
	public void setName(String name);
	
//	public void setBirthEventId(String birthEventId);
//	
//	public void setDeathEventId(String deathEventId);
	
	public void setReferenceEvent(XOCEvent observationEvent);
	
//	public void setBirthEvent(XOCEvent birthEvent);
//	
//	public void setDeathEvent(XOCEvent deathEvent);
//	
//	/**
//	 * Set the death event order number i (i starts from 1, i.e., corresponding event is log.get(i-1))
//	 * @param deathEventOrderNum
//	 */
//	public void setDeathEventOrderNum(int deathEventOrderNum);
	
//	public void setID(XID ID);
	
	public Object clone();
	
	public boolean isAttributesEqual(XOCObject object);
	
	/*
	 * minus the overlap part (attribute name and value are the same) between the
	 * current object and the input object from the current object
	 */
	public XOCObject minusAttributeOverlap(XOCObject object);	
	
	public XOCObject plusAttributes(XOCObject object);
	
	public Map<String, XOCAttribute> getXOCAttributes();
	
	public void setXOCAttributes(Map<String, XOCAttribute> attributeVersionMap);
	
	/**
	 * add an XOC attribute (key, value, event order number, i.e., moment) into the corresponding map
	 * @param key
	 * @param value
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void addXOCAttribute(String key, String value, int eventOrderNum, String logFormat);
	
	/**
	 * end an XOC attribute, i.e., adding the ending event order number in the last value version corresponding to the input key
	 * @param key
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void endXOCAttribute(String key, int eventOrderNum, String logFormat);
	
	/**
	 * get the lifecycle period list of one object;
	 * each period consists of two values: start event order number and end event order number;
	 * the object still (maybe last) observed in the object model of the end event;
	 * @return
	 */
	public List<IntegerPair> getLifecyclePeriodList();
	
	/**
	 * check if the object is alive at the moment of input event based on the life cycle period list
	 * @param event
	 * @return
	 */
	public boolean isAliveAtEvent(XOCEvent event);
	
	/**
	 * check if the object is alive at the moment of input event order number based on the life cycle period list
	 * @param event
	 * @return
	 */
	public boolean isAliveAtEventOrderNum(int inputEventOrderNum);
	/**
	 * check if the object is updated at the moment of input event based on the life cycle period list
	 * if one object is alive and has different attributes at previous and current events, the object is updated
	 * @param event
	 * @return
	 */
	public boolean isUpdatedAtEvent(XOCEvent event);
	
	/**
	 * check if the object is born at the moment of input event based on the life cycle period list
	 * if one object is alive at current event and dead at previous event, the object is added at current event
	 * @param event
	 * @return
	 */
	public boolean isAddedAtEvent(XOCEvent event);
	
	/**
	 * get the map between event order number and object state;
	 * if one event number is not in the map, it means (1) the object is not alive (after delete && before add if any), or 
	 * (2) remain the same (state value: remain) after add or update && before delete (if any)
	 * @return
	 */
	public Map<Integer, String> getEventOrderNumVSObjectStateMap();
	
	/**
	 * create the updated attribute map through comparing values of the input event with values of its previous event;
	 * @param event
	 * @return
	 */
	public XAttributeMap createUpdatedAttributeMap(XOCEvent event);
}