/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model.impl;

import java.util.ArrayList;
import java.util.List;

import org.processmining.common.structure.IntegerPair;
import org.processmining.framework.util.Pair;
import org.processmining.ocbc.log.model.XOCAttribute;

/**
 * This class implements the abstract base class for strongly-typed attributes.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * 
 */
public class XOCAttributeImpl implements XOCAttribute {
	/**
	 * Key, i.e. unique name, of this attribute. 
	 */
	private final String key;

	/**
	 * List of attribute values, i.e., value versions of the attribute.
	 */
	private List<String> attributeValueList = new ArrayList<String>();
	
	/**
	 * List of time periods, i.e., a period corresponds to one value version.
	 * different period no overlapping. i.e., a value keep from the start event (event order num) until the end event (include the moment of end event)
	 */
	private List<Pair<Integer, Integer>> eventPeriodList = new ArrayList<Pair<Integer, Integer>>();

	/**
	 * record the last value in the value list
	 */
	private String currentValue;
	
	/**
	 * Creates a new, empty attribute.
	 * 
	 * @param key
	 *            The key, i.e. unique name identifier, of this attribute.
	 */
	public XOCAttributeImpl(String key) {
		this.key = key;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.impl.XAttribute#getKey()
	 */
	public String getKey() {
		return key;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.impl.XAttribute#clone()
	 */
//	public Object clone() {
//		XOCAttributeImp clone = null;
//		try {
//			clone = (XOCAttributeImp) super.clone();
//		} catch (CloneNotSupportedException e) {
//			e.printStackTrace();
//			return null;
//		}
//		return clone;
//	}

	/**
	 * get the current value (the last value in the value list)
	 * @return
	 */
	public String getCurrentValue(){
		return currentValue;
	}
	
	/**
	 * get the value list
	 * @return
	 */
	public List<String> getValueList(){
		return attributeValueList;
	}
	
	/**
	 * get the period list
	 * @return
	 */
	public List<Pair<Integer, Integer>> getPeriodList(){
		return eventPeriodList;
		 
	}
	
	/**
	 * generate the pair (Pair<Integer, Integer>) list for one input value, i.e., all survival periods of one value;
	 * @return
	 */
	public List<Pair<Integer, Integer>> getSurvivalPeriodListForValue(String value){
		List<Pair<Integer, Integer>> survivalPeriodList = new ArrayList<Pair<Integer, Integer>>();
		for (int i = 0; i < attributeValueList.size(); i++){
			if(value.equals(attributeValueList.get(i))){
				survivalPeriodList.add(eventPeriodList.get(i));
			}
		}          
		return survivalPeriodList;
	}
	
	/**
	 * generate the pair (IntegerPair) list for one input value, i.e., all survival periods of one value ;
	 * @return
	 */
	public List<IntegerPair> getSurvivalPeriodListOfValue(String value){
		List<IntegerPair> survivalPeriodList = new ArrayList<IntegerPair>();
		for (int i = 0; i < attributeValueList.size(); i++){
			if(value.equals(attributeValueList.get(i))){
				survivalPeriodList.add(new IntegerPair(eventPeriodList.get(i).getFirst(), eventPeriodList.get(i).getSecond()));
			}
		}          
		return survivalPeriodList;
	}
	
	static ArrayList<Integer> indexOfAll(Object obj, ArrayList list){
	    ArrayList<Integer> indexList = new ArrayList<Integer>();
	    for (int i = 0; i < list.size(); i++)
	        if(obj.equals(list.get(i)))
	            indexList.add(i);
	    return indexList;
	}
	
	/**
	 * return the value version corresponding to the input event order number;
	 * return null if there is no value at the event order number;
	 * @param eventIndex
	 * @return
	 */
	public String getValueVersion(int eventOrderNum) {
		for(Pair<Integer, Integer> eventPair : eventPeriodList){
			int pairIndex = eventPeriodList.indexOf(eventPair);
			int startEventIndex = eventPair.getFirst();
			int endEventIndex;
			if(eventPair.getSecond() == null){
				endEventIndex = eventOrderNum;
			}
			else{
				endEventIndex = eventPair.getSecond();
			}		
			
//			System.out.println("in getValueVersion startEventIndex: " + startEventIndex);
//			System.out.println("in getValueVersion endEventIndex: " + endEventIndex);
			
			if(eventOrderNum >= startEventIndex  && eventOrderNum  <= endEventIndex){
//				System.out.println("in getValueVersion attributeValueList.get(pairIndex): " + attributeValueList.get(pairIndex));
				return attributeValueList.get(pairIndex);
			}
		}
		return null;
	}
	
	/**
	 * add a new value version into the corresponding list
	 * @param value
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void insertNewValueVersion(String value, int eventOrderNum, String logFormat){
		if(value != null){
			currentValue = value;
			attributeValueList.add(value);
			if(logFormat.equals(xConstants.TOTAL)){
				eventPeriodList.add(new Pair<Integer, Integer>(eventOrderNum, eventOrderNum));

			}
			//in the update variant, the default ending event for an attribute value is null (or a number larger that the log size)
			else if(logFormat.equals(xConstants.UPDATE)){
				int valueCount = eventPeriodList.size();
				//if there is existing versions and the ending event is null, add the ending event for the last version			
				if(valueCount>0 && eventPeriodList.get(valueCount-1).getSecond() == null){
					int startEventOrderNumOfLastVersion = eventPeriodList.get(valueCount-1).getFirst();	
					eventPeriodList.remove(valueCount-1);
					eventPeriodList.add(new Pair<Integer, Integer>(startEventOrderNumOfLastVersion, eventOrderNum-1));
				}
				eventPeriodList.add(new Pair<Integer, Integer>(eventOrderNum, null));
			}
		}
		//if the value is null, current used in the delete life cycle, set the end event order number
		else{
			if(logFormat.equals(xConstants.TOTAL)){				

			}		
			else if(logFormat.equals(xConstants.UPDATE)){
				int valueCount = eventPeriodList.size();
				//if there is existing versions, add the ending event for the last version			
				if(valueCount>0){
					int startEventOrderNumOfLastVersion = eventPeriodList.get(valueCount-1).getFirst();	
					eventPeriodList.remove(valueCount-1);
					eventPeriodList.add(new Pair<Integer, Integer>(startEventOrderNumOfLastVersion, eventOrderNum-1));
				}
			}
		}
		
		
	}

	/**
	 * update the current value version
	 */
	public void updateCurrentValueVersion(){
		int valueCount = eventPeriodList.size();
		int startEventOrderNum = eventPeriodList.get(valueCount-1).getFirst();
		int endEventOrderNum = eventPeriodList.get(valueCount-1).getSecond();
		eventPeriodList.remove(valueCount-1);
		eventPeriodList.add(new Pair<Integer, Integer>(startEventOrderNum, endEventOrderNum+1));
	}
	
	/**
	 * add a value version into the corresponding list
	 * if the value equals the current value, update current value version
	 * otherwise, insert a new value version
	 * @param value
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void addValueVersion(String value, int eventOrderNum, String logFormat){
		//if the attribute value equals the current value 		
		if(currentValue.equals(value)){
			if(logFormat.equals(xConstants.TOTAL)){
				int valueCount = eventPeriodList.size();
				//the ending event order number in total variant always exists
				int endEventOrderNum = eventPeriodList.get(valueCount-1).getSecond();
				//if there is no interval, update the value version through extending the ending boundary with 1
				if(eventOrderNum == endEventOrderNum + 1){
					updateCurrentValueVersion();
				}
				else{
					insertNewValueVersion(value, eventOrderNum, logFormat);
				}		
			}	
			//in the update variant, if update attribute with the same old value. do nothing
			else{
				
			}
		}
		//if the attribute value does not equal the current value 
		else if(!currentValue.equals(value)){
			insertNewValueVersion(value, eventOrderNum, logFormat);
		}
	}
	
	/**
	 * add an ending event order number (eventOrderNum-1) for the last value version
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void endLastValueVersion(int eventOrderNum, String logFormat){
		if(logFormat.equals(xConstants.UPDATE)){
			int valueCount = eventPeriodList.size();
			int startEventOrderNumOfLastVersion = eventPeriodList.get(valueCount-1).getFirst();	
			eventPeriodList.remove(valueCount-1);
			eventPeriodList.add(new Pair<Integer, Integer>(startEventOrderNumOfLastVersion, eventOrderNum-1));
		}	
		//in the total variant, do nothing because the
		else{
			
		}
	}
}
