/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.ocbc.log.extension.XExtension;
import org.processmining.ocbc.log.id.XID;
import org.processmining.ocbc.log.model.XAttributeMap;
import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.log.model.XOCORelation;
import org.processmining.ocbc.log.model.XOCObject;
import org.processmining.ocbc.log.model.XOCObjectModel;
import org.processmining.ocbc.log.model.XOCRRelation;
import org.processmining.ocbc.models.visualization.cell.MyOriginalCell;




/**
 * Implementation for the XEvent interface.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * 
 */
public class XOCEventImpl implements XOCEvent {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * ID of this event.
	 */
//	private XID ID;

	/**
	 * Map of attributes for this event.
	 */
	private XAttributeMap attributes;
	
	private XOCObjectModel objectModel;//contain all objects and relations abstracted from a database state
	private List<XOCObject> relatedObjects;
	
	//just contain the updated objects and relations compared with last state;
	//used for XOC log file
	private XOCObjectModel objectModelUpdate = new XOCObjectModelImpl();
	
	//just the minimal Object Model (object model update + source and target objects of relations);
	//used for conformance checking
	private XOCObjectModel minimalObjectModel = new XOCObjectModelImpl();

	
	private List<XOCObject> relatedSimplifiedObjects  = new ArrayList<XOCObject>();
	private List<String> relatedObjectIdList = new ArrayList<String>();
	private int orderNum;//the orderNum of the first event is 1; N+
	private String activity;
	private MyOriginalCell correspondingCell;

	private List<XOCRRelation> referenceRelationList = new ArrayList<XOCRRelation>();
	
	public XOCEventImpl() {
		this("activity");
	}
	
	/**
	 * (1)input the activity name for the event
	 * (2)automatically create empty attributes, related object list and object model
	 */
	public XOCEventImpl(String activity) {
		this(activity, -1);
	}
	
	public XOCEventImpl(String activity, int orderNum) {
		this(new XID(), activity, orderNum);
	}
	
	public XOCEventImpl(XID ID, String activity, int orderNum) {
		this(ID, new XAttributeMapImpl(), activity, orderNum, new ArrayList<XOCObject>(), new XOCObjectModelImpl());
	}
	
	/**
	 * Creates a new event with the given id and attributed
	 * 
	 * @param id
	 *            the id for this event
	 * @param attributes
	 *            Map of attribute for the event.
	 */
	public XOCEventImpl(XID ID, XAttributeMap attributes) {
		this(ID, attributes, "activity", -1, new ArrayList<XOCObject>(), new XOCObjectModelImpl());
	}
	
	public XOCEventImpl(String activity, int orderNum, List<XOCObject> relatedObjects, XOCObjectModel objectModel) {
		this(new XID(), new XAttributeMapImpl(), activity, orderNum, relatedObjects, objectModel);	
	}
	
	public XOCEventImpl(XID ID, XAttributeMap attributes, String activity, int orderNum, List<XOCObject> relatedObjects, XOCObjectModel objectModel) {
//		this.ID = ID;
		this.attributes = attributes;
		this.activity = activity;
		this.orderNum = orderNum;
		this.relatedObjects = relatedObjects;
		this.objectModel = objectModel;		
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.XAttributable#getAttributes()
	 */
	public XAttributeMap getAttributes() {
		return attributes;
	}

	public XOCObjectModel getObjectModel() {
		return objectModel;
	}
	
	public XOCObjectModel getObjectModelUpdate() {
		return objectModelUpdate;
	}
	
	public XOCObjectModel getMinimalObjectModel(){
		return minimalObjectModel;
	}
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.XAttributable#setAttributes(java.util.Map)
	 */
	public void setAttributes(XAttributeMap attributes) {
		this.attributes = attributes;
	}

	public void setObjectModel(XOCObjectModel objectModel) {
		this.objectModel = objectModel;
	}
	
	public void setObjectModelUpdate(XOCObjectModel objectModelUpdate) {
		this.objectModelUpdate = objectModelUpdate;
	}
	
	public void setMinimalObjectModel(XOCObjectModel minimalObjectModel){
		this.minimalObjectModel = minimalObjectModel;
	}
	
	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XAttributable#hasAttributes()
	 */
	@Override
	public boolean hasAttributes() {
		return !attributes.isEmpty();
	}
	


	/**
	 * Clones this event, i.e. creates a deep copy, but with a new ID, so equals
	 * does not hold between this and the clone
	 */
	public Object clone() {
		XOCEventImpl clone;
		clone = new XOCEventImpl();
//		try {
//			clone = (XOCEventImpl) super.clone();
//		} catch (CloneNotSupportedException e) {
//			e.printStackTrace();
//			return null;
//		}
//		clone.ID = (XID) this.ID.clone();
		clone.attributes = (XAttributeMap) attributes.clone();

		List<XOCObject> objectsInObjectModel = new ArrayList<XOCObject>();
		for(XOCObject object : this.objectModel.getObjectSet()){
			XOCObject objectClone = (XOCObject) object.clone();
			objectsInObjectModel.add(objectClone);
			
		}
		clone.objectModel.getObjectSet().addAll(objectsInObjectModel);
		
		
		clone.objectModel.generateIdVSObjectMap();
		
		Map<String, XOCObject> idVSObjectMapOfCloneObjectModel = clone.objectModel.getIdVSObjectMap();
			
		
		for(XOCORelation objectRelation : this.objectModel.getObjectRelationSet()){
			XOCObject sourceObject = objectRelation.getSourceObject();
			String sourceObjectId = sourceObject.getAttributes().get("id").toString();
			XOCObject sourceObjectClone = idVSObjectMapOfCloneObjectModel.get(sourceObjectId);
			XOCObject targetObject = objectRelation.getTargetObject();
			String targetObjectId = targetObject.getAttributes().get("id").toString();
			XOCObject targetObjectClone = idVSObjectMapOfCloneObjectModel.get(targetObjectId);
			String type = objectRelation.getRelationType();
			
			XOCORelation objectRelationClone = new XOCORelationImpl(type, sourceObjectClone, targetObjectClone);
			objectRelationClone.setAttributes((XAttributeMap) objectRelation.getAttributes().clone());
			clone.objectModel.getObjectRelationSet().add(objectRelationClone);
			
		}
		
		//clone related objects	
		List<XOCObject> relatedObjectClones = new ArrayList<XOCObject>();//
    	for(XOCObject objectOriginal : this.getRelatedObjects()){
    		String objectId = objectOriginal.getAttributes().get("id").toString();
    		XOCObject objectInObjectModel = idVSObjectMapOfCloneObjectModel.get(objectId);
    		if(objectInObjectModel != null){
    			relatedObjectClones.add(objectOriginal);
    		}
    		else
    			relatedObjectClones.add((XOCObject) objectOriginal.clone());
    	}
    	clone.getRelatedObjects().addAll(relatedObjectClones);	

    	//clone related object id list
    	for(String objectIdOriginal : this.getRelatedObjectIdList()){
    		clone.getRelatedObjectIdList().add(objectIdOriginal);  		
    	}  	
		return clone;
	}

	/**
	 * Tests for equality of IDs
	 */
	public boolean equals(Object o) {
		if (o instanceof XOCEventImpl) {
			return ((XOCEventImpl) o).getAttributes().get(xConstants.ID).toString().
					equals(attributes.get(xConstants.ID).toString());
		} else {
			return false;
		}
	}

	public boolean relatedObjectsContain(XOCObject object){
		for(XOCObject relatedObject : relatedObjects){
			if(relatedObject.getAttributes().get("id").equals(object.getAttributes().get("id")))
				return true;
		}
		return false;
	}
	
	public boolean objectModelContain(XOCObject object){
		for(XOCObject objectOfModel : objectModel.getObjectSet()){
			if(objectOfModel.getAttributes().get("id").equals(object.getAttributes().get("id")))
				return true;
		}
		return false;
	}
	/**
	 * Returns the hashCode of the id
	 */
//	public int hashCode() {
//		return ID.hashCode();
//	}
//
//	/*
//	 * (non-Javadoc)
//	 * @see org.deckfour.xes.model.XEvent#getID()
//	 */
//	public XID getID() {
//		return ID;
//	}
//
//	/**
//	 * Sets the ID. Should only be used for deserialization purposes
//	 * 
//	 * @param id
//	 *            the new id.
//	 */
//	public void setID(XID ID) {
//		this.ID = ID;
//	}

	public Set<XExtension> getExtensions() {
		// TODO Auto-generated method stub
		return null;
	}



	public String getActivity() {
		return activity;
	}

	public List<XOCObject> getRelatedObjects() {
		return relatedObjects;
	}

	/**
	 * Get the list of reference relations of the event (e.g., used in the object model graph to refer to the corresponding cell of one reference relation)
	 * @return
	 */
	public List<XOCRRelation> getReferenceRelationList(){
		return referenceRelationList;
	}
	
	/**
	 * Get the reference relation connecting the event and the input object.
	 * Return null if there is no available reference relation.
	 * @param object
	 * @return
	 */
	public XOCRRelation getReferenceRelation(XOCObject object){
		for(XOCRRelation referenceRelation : referenceRelationList){
			if(referenceRelation.getTargetObject().equals(object)){
				return referenceRelation;
			}		
		}
		return null;
	}
	
	public List<XOCObject> getSimplifiedRelatedObjects(){		
		return relatedSimplifiedObjects;
	}
	
	public List<String> getRelatedObjectIdList() {
		return relatedObjectIdList;
	}
	
	/**
	 * return the index of one event
	 * note that the orderNum of the first event is 1; N+
	 */
	public int getOrderNum() {
		return orderNum;
	}

	public void setActivity(String activity) {
		this.activity = activity;		
	}

	public void setRelatedObjects(List<XOCObject> relatedObjects) {
		this.relatedObjects = relatedObjects;		
	}

	public void setSimplifiedRelatedObjects(List<XOCObject> relatedSimplifiedObjects){
		this.relatedSimplifiedObjects = relatedSimplifiedObjects;		
	}
	
	public void setRelatedObjectIdList(List<String> relatedObjectIdList) {
		this.relatedObjectIdList = relatedObjectIdList;		
	}
	
	public void setOrderNum(int orderNum) {
		this.orderNum = orderNum;		
	}

	public MyOriginalCell getCorrespondingCell(){
		return correspondingCell;
	}
	
	public void setCorrespondingCell(MyOriginalCell correspondingCell){
		this.correspondingCell = correspondingCell;
	}

	/*
	 * get all relevant objects for one object by checking object relations
	 */
	public List<XOCObject> getRelevantObjectsForOneObject(XOCObject object) {
		List<XOCObject>  relevantObjectList = new ArrayList<XOCObject>();
		String objectId = object.getAttributes().get("id").toString();
		for(XOCORelation relation : objectModel.getObjectRelationSet()){
			String sourceObjectId = relation.getSourceObject().getAttributes().get("id").toString();
			String targetObjectId = relation.getTargetObject().getAttributes().get("id").toString();
			if(objectId.equals(sourceObjectId)){
				if(!relevantObjectList.contains(relation.getTargetObject()))
					relevantObjectList.add(relation.getTargetObject());
			}
			else if(objectId.equals(targetObjectId)){
				if(!relevantObjectList.contains(relation.getSourceObject()))
					relevantObjectList.add(relation.getSourceObject());
			}
		}	
		return relevantObjectList;
	}

	public List<XOCORelation> getRelevantObjectRelationsForOneObject(XOCObject object) {
		List<XOCORelation>  relevantObjectRelationList = new ArrayList<XOCORelation>();
		String objectId = object.getAttributes().get("id").toString();
		for(XOCORelation relation : objectModel.getObjectRelationSet()){
			String sourceObjectId = relation.getSourceObject().getAttributes().get("id").toString();
			String targetObjectId = relation.getTargetObject().getAttributes().get("id").toString();
			if(objectId.equals(sourceObjectId) || objectId.equals(targetObjectId)){
				if(!relevantObjectRelationList.contains(relation))
					relevantObjectRelationList.add(relation);
			}
		}		
		return relevantObjectRelationList;
	}



}
