/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model.impl;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.ocbc.discovery.filter.FilterFunction;
import org.processmining.ocbc.log.extension.XExtension;
import org.processmining.ocbc.log.factory.XFactory;
import org.processmining.ocbc.log.factory.XFactoryRegistry;
import org.processmining.ocbc.log.io.XocXmlParser;
import org.processmining.ocbc.log.model.XAttribute;
import org.processmining.ocbc.log.model.XAttributeMap;
import org.processmining.ocbc.log.model.XOCAttribute;
import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.log.model.XOCLog;
import org.processmining.ocbc.log.model.XOCORelation;
import org.processmining.ocbc.log.model.XOCObject;
import org.processmining.ocbc.log.model.XOCObjectModel;
import org.processmining.ocbc.log.model.XOCRRelation;

/**
 * Implementation for the XLog interface.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 *
 */
public class XOCLogImpl extends ArrayList<XOCEvent> implements XOCLog {

	/**
	 * serial version UID.
	 */
	private static final long serialVersionUID = -9192919845877466525L;
	
	/**
	 * Map of attributes for this log.
	 */
	private XAttributeMap attributes;
//	/**
//	 * Extensions.
//	 */
//	private Set<XExtension> extensions;
//	/**
//	 * Classifiers.
//	 */
//	private List<XEventClassifier> classifiers;
//	/**
//	 * Global trace attributes.
//	 */
	private List<XAttribute> globalEventAttributes;
	/**
	 * Global event attributes.
	 */
	private List<XAttribute> globalObjectModelAttributes;
	
	private List<XAttribute> globalObjectRelationlAttributes;
	
	private List<XAttribute> globalObjectAttributes;
	
	private List<XOCEvent> allEvents;
	
	private List<String> eventIdList = new ArrayList<String>();;
	
	private List<XOCRRelation> allReferenceRelations = new ArrayList<XOCRRelation>();
	
	private List<String> classNameList = new ArrayList<String>();
	
	private Map<String,XOCObject> idVSObject = new HashMap<String,XOCObject>();

	private Map<String,XOCORelation> idVSObjectRelation = new HashMap<String,XOCORelation>();
	
	private List<XOCObject> objectList = new ArrayList<XOCObject>();

	private List<XOCORelation> objectRelationList = new ArrayList<XOCORelation>();
	
	private int XOCLogSize;
	protected XFactory factory = XFactoryRegistry.instance().currentDefault();
	
	protected FilterFunction filter = new FilterFunction();
	/**
	 * Single-item cache. Only the last info is cached. 
	 * Typically, only one classifier will be used for a log.
	 */

	
	/**
	 * Creates a new log.
	 * 
	 * @param attributeMap The attribute map used to store this
	 * 	log's attributes.
	 */
	public XOCLogImpl() {	
		this.allEvents = new ArrayList<XOCEvent>();
		this.attributes = new XAttributeMapImpl();
		this.globalEventAttributes = new ArrayList<XAttribute>();
		this.globalObjectModelAttributes = new ArrayList<XAttribute>();
		this.globalObjectRelationlAttributes = new ArrayList<XAttribute>();
		this.globalObjectAttributes = new ArrayList<XAttribute>();
		this.allReferenceRelations = new ArrayList<XOCRRelation>();
//		this.classNames = new ArrayList<String>();
//		this.activityNames = new ArrayList<String>();
//		this.resources = new ArrayList<String>();
//		this.repeatableObjects = new ArrayList<XOCObject>();
	}
	
	public XOCLogImpl(XAttributeMap attributeMap) {
		this.attributes = attributeMap;		
		this.allEvents = new ArrayList<XOCEvent>();
		this.globalEventAttributes = new ArrayList<XAttribute>();
		this.globalObjectModelAttributes = new ArrayList<XAttribute>();
		this.globalObjectRelationlAttributes = new ArrayList<XAttribute>();
		this.globalObjectAttributes = new ArrayList<XAttribute>();
		this.allReferenceRelations = new ArrayList<XOCRRelation>();
//		this.classNames = new ArrayList<String>();
//		this.activityNames = new ArrayList<String>();
//		this.resources = new ArrayList<String>();
//		this.repeatableObjects = new ArrayList<XOCObject>();
	}

	public Object clone() {
		XOCLogImpl clone;
//		clone = (XOCLogImpl) super.clone();
		clone = new XOCLogImpl();
		clone.attributes = (XAttributeMap) attributes.clone();
//		clone.objectModel = new XOCObjectModelImpl();
//		clone.relatedObjects = new ArrayList<XOCObject>();
//		
//		for(XOCObject object : this.objectModel.getObjectSet()){
//			XOCObject objectClone = (XOCObject) object.clone();
//			clone.objectModel.getObjectSet().add(objectClone);
//			if(this.relatedObjects.contains(object)){
//				clone.relatedObjects.add(objectClone);
//			}
//		}
//		Map(object, objectClone)
		List<XOCEvent> eventCloneList = new ArrayList<XOCEvent>();;
		for(XOCEvent event : this.allEvents){				
			eventCloneList.add((XOCEvent) event.clone());		
		}
		clone.allEvents.addAll(eventCloneList);
		
	
		
		//create reference relations based on the related objects for each event
        for(XOCEvent event : clone.getAllEvents()){
        	for(XOCObject object : event.getRelatedObjects()){
        		XOCRRelation referenceRelation = new XOCRRelationImpl(event.getAttributes().get("activity").toString() + 
        				object.getAttributes().get("class").toString(), event, object);      		       		
        		XAttribute attribute = factory.createAttributeLiteral("id", referenceRelation.getID().toString(), null);
        		referenceRelation.getAttributes().put(attribute.getKey(), attribute);        		
        		clone.getReferenceRelations().add(referenceRelation);
        	}    	
        }
        
		return clone;
	}
	
	public List<XAttribute> getGlobalEventAttributes() {
		return globalEventAttributes;
	}

	public List<XAttribute> getGlobalObjectModelAttributes() {
		return globalObjectModelAttributes;
	}

	public List<XAttribute> getGlobalObjectRelationAttributes() {
		return globalObjectRelationlAttributes;
	}
	
	public List<XAttribute> getGlobalObjectAttributes() {
		return globalObjectAttributes;
	}
	

	public List<XOCEvent> getAllEvents() {
		return allEvents;
	}
	
	public List<String> getEventIdList(){
		return eventIdList;
	}
	
	/**
	 * get all events of one activity. sorted by the their positions in the XOC log
	 */
	public List<XOCEvent> getEventsOfActivity(String activityNameInput){
		List<XOCEvent>  eventList = new ArrayList<XOCEvent>();
		for(XOCEvent event : allEvents){
			String activityName = event.getAttributes().get(xConstants.ACTIVITY).toString();
			if(activityName.equals(activityNameInput)){
				eventList.add(event);
			}
		}
		return eventList;
	}
	
	public Map<String, XOCEvent> getIdVSEventMap(){
		Map<String, XOCEvent> idVSEventMap = new HashMap<String, XOCEvent>();
		for(XOCEvent event : allEvents){
			idVSEventMap.put(event.getAttributes().get(xConstants.ID).toString(), event);
		}
		return idVSEventMap;		
	}
	
	public XOCEvent getLastEvent() {
		return allEvents.get(allEvents.size()-1);
	}
	
	public List<XOCEvent> createStartEvents() {
		List<XOCEvent>  startEvents = new ArrayList<XOCEvent>();		
		for(XOCEvent event : allEvents){
			boolean startFlag = true;
			for(XOCEvent relatedEvent :getRelatedEventsForOneEvent(event)){
				if(allEvents.indexOf(event) > allEvents.indexOf(relatedEvent)){
					startFlag = false;
					break;
				}
			}
			if(startFlag)
				startEvents.add(event);
		}
		return startEvents;
	}

	public  Map<String, Integer> createStartActivityVSNumMap() {
		Map<String, Integer> activityVSNumMap = new HashMap<String, Integer>();
		for(XOCEvent event : createStartEvents()){
			String activity = event.getAttributes().get("activity").toString();
			if(!activityVSNumMap.containsKey(activity)){
				activityVSNumMap.put(activity, 1);
			}
			else
				activityVSNumMap.put(activity, activityVSNumMap.get(activity)+1);
		}
		return activityVSNumMap;
	}
	
	public List<String> createStartActivityNames() {
		List<String>  activityNames = new ArrayList<String>();
		Map<String, Integer> activityVSNumMap = createStartActivityVSNumMap();		
		//sort activities by their numbers
		while(activityVSNumMap.keySet().size()>0){
			int activityMaximalNum = 0;
			String activityMaximal = null;
			for(String activity : activityVSNumMap.keySet()){
				if(activityMaximalNum < activityVSNumMap.get(activity)){
					activityMaximalNum = activityVSNumMap.get(activity);
					activityMaximal = activity;
				}
			}
			System.out.print("\n activityMaximal" + activityMaximal);
			activityVSNumMap.remove(activityMaximal);
			activityNames.add(activityMaximal);
		}
		return activityNames;
	}
	
	
	public List<XOCEvent> createEndEvents() {
		List<XOCEvent>  startEvents = new ArrayList<XOCEvent>();		
		for(XOCEvent event : allEvents){
			boolean startFlag = true;
			for(XOCEvent relatedEvent :getRelatedEventsForOneEvent(event)){
				if(allEvents.indexOf(event) < allEvents.indexOf(relatedEvent)){
					startFlag = false;
					break;
				}
			}
			if(startFlag)
				startEvents.add(event);
		}
		return startEvents;
	}
	
	public  Map<String, Integer> createEndActivityVSNumMap() {
		Map<String, Integer> activityVSNumMap = new HashMap<String, Integer>();
		for(XOCEvent event : createEndEvents()){
			String activity = event.getAttributes().get("activity").toString();
			if(!activityVSNumMap.containsKey(activity)){
				activityVSNumMap.put(activity, 1);
			}
			else
				activityVSNumMap.put(activity, activityVSNumMap.get(activity)+1);
		}
		return activityVSNumMap;
	}
	
	public List<String> createEndActivityNames() {
		List<String>  activityNames = new ArrayList<String>();
		Map<String, Integer> activityVSNumMap = createEndActivityVSNumMap();		
		//sort activities by their numbers
		while(activityVSNumMap.keySet().size()>0){
			int activityMaximalNum = 0;
			String activityMaximal = null;
			for(String activity : activityVSNumMap.keySet()){
				if(activityMaximalNum < activityVSNumMap.get(activity)){
					activityMaximalNum = activityVSNumMap.get(activity);
					activityMaximal = activity;
				}
			}
			System.out.print("\n activityMaximal" + activityMaximal);
			activityVSNumMap.remove(activityMaximal);
			activityNames.add(activityMaximal);
		}
		return activityNames;
	}
	
	public  Map<String, Integer> createActivityVSNumMap() {
		Map<String, Integer> activityVSNumMap = new HashMap<String, Integer>();
		for(XOCEvent event : allEvents){
			String activity = event.getAttributes().get("activity").toString();
			if(!activityVSNumMap.containsKey(activity)){
				activityVSNumMap.put(activity, 1);
			}
			else
				activityVSNumMap.put(activity, activityVSNumMap.get(activity)+1);
		}		
		return activityVSNumMap;
	}
	
	public List<String> createActivityNames() {
		List<String>  activityNames = new ArrayList<String>();
		Map<String, Integer> activityVSNumMap = createActivityVSNumMap();		
		//sort activities by their numbers
		while(activityVSNumMap.keySet().size()>0){
			int activityMaximalNum = 0;
			String activityMaximal = null;
			for(String activity : activityVSNumMap.keySet()){
				if(activityMaximalNum < activityVSNumMap.get(activity)){
					activityMaximalNum = activityVSNumMap.get(activity);
					activityMaximal = activity;
				}
			}
			System.out.print("\n activityMaximal" + activityMaximal);
			activityVSNumMap.remove(activityMaximal);
			activityNames.add(activityMaximal);
		}
		return activityNames;
	}
	
	public List<String> generateActivityNameList() {
		List<String>  activityNames = new ArrayList<String>();
		for(XOCEvent event : allEvents){
			String activity = event.getAttributes().get(xConstants.ACTIVITY).toString();
			if(!activityNames.contains(activity)){
				activityNames.add(activity);
			}		
		}		
		return activityNames;
	}
	
	public List<String> getClassNames() {
		List<String>  classNames = new ArrayList<String>();
		for(XOCObject object : getLastEvent().getObjectModel().getObjectSet()){
			if(object.getAttributes().get(xConstants.CLASS) != null){
				String classType = object.getAttributes().get(xConstants.CLASS).toString();
				if(!classNames.contains(classType)){
					classNames.add(classType);
				}
			}		
		}
		return classNames;
	}
	
	/**
	 * generate the list of class names
	 * based on the object model of the last event
	 */
	public List<String> generateClassNameList(){
		for(XOCObject object : getLastEvent().getObjectModel().getObjectSet()){
			if(object.getAttributes().get(xConstants.CLASS) != null){
				String classType = object.getAttributes().get(xConstants.CLASS).toString();
				if(!classNameList.contains(classType)){
					classNameList.add(classType);
				}
			}			
		}
		return classNameList;		
	}
	
	/**
	 * get the list of class names after generating
	 * (should be used after executing the generateClassNameList() function,
	 * otherwise, the list is of size 0)
	 */
	public List<String> getClassNameList(){
		return classNameList;
	}
	
	public List<String> getResources() {
		List<String>  resources = new ArrayList<String>();
		for(XOCEvent event : allEvents){
			if(event.getAttributes().containsKey("resource")){
				String resource = event.getAttributes().get("resource").toString();
				if(!resources.contains(resource)){
					resources.add(resource);
				}
			}
		}
		return resources;
	}
	
	public List<XOCObject> getRepeatableObjects() {
		List<XOCObject>  repeatableObjects = new ArrayList<XOCObject>();
		for(XOCEvent event : allEvents){
			for(XOCObject object : event.getObjectModel().getObjectSet()){			
				repeatableObjects.add(object);
			}
		}
		return repeatableObjects;
	}

	public List<XOCRRelation> getReferenceRelations(){
		return allReferenceRelations;
	}
	 
	//get the source object of an object relation
	public XOCObject getSourceObjectOfObjectRelation(XOCObjectModel objectModel, XOCORelation objectRelation) {
		String sourceObjectId = objectRelation.getAttributes().get("sourceobjectid").toString();
		XOCObject sourceObject = new XOCObjectImpl("None");
		for(XOCObject object : objectModel.getObjectSet()){
			String objectId = object.getAttributes().get("id").toString();
			if(sourceObjectId.equals(objectId))
				sourceObject = object;
		}
		return sourceObject;
	}
	
	//get the target object of an object relation
	public XOCObject getTargetObjectOfObjectRelation(XOCObjectModel objectModel, XOCORelation objectRelation) {
		String targetObjectId = objectRelation.getAttributes().get("targetobjectid").toString();
		XOCObject targetObject = new XOCObjectImpl("None");
		for(XOCObject object : objectModel.getObjectSet()){
			String objectId = object.getAttributes().get("id").toString();
			if(targetObjectId.equals(objectId))
				targetObject = object;
		}
		return targetObject;
	}
	

	
	public void setAllEvents(List<XOCEvent> allEvents) {
		this.allEvents = allEvents;
		
	}

	public XAttributeMap getAttributes() {
		// TODO Auto-generated method stub
		return attributes;
	}

	public void setAttributes(XAttributeMap attributes) {
		// TODO Auto-generated method stub
		this.attributes = attributes;
	}

	public boolean hasAttributes() {
		// TODO Auto-generated method stub
		return false;
	}

	public Set<XExtension> getExtensions() {
		// TODO Auto-generated method stub
		return null;
	}
	
	@Override
	public boolean add(XOCEvent event) {		
		allEvents.add(event);
		return super.add(event);
	}
	
	//get all objects connected to related objects of an event + related objects of an event
	public List<XOCObject> getAllObjectVersions() {
		List<XOCObject> allObjectVersionList = new ArrayList<XOCObject>();
		for(XOCEvent event : allEvents){
			for(XOCObject object : event.getObjectModel().getObjectSet()){
				allObjectVersionList.add(object);
			}
		}	
		return allObjectVersionList;
	}
	
	//get all objects connected to related object of an event
	public List<XOCObject> getIndirectRelatedObjectsForOneEvent(XOCEvent event) {
		XOCEvent lastEvent = allEvents.get(allEvents.size()-1);
		List<XOCObject> indirectRelatedObjects = new ArrayList<XOCObject>();
		for(XOCObject relatedObject : event.getRelatedObjects()){
			for(XOCObject indirectRelatedObject : lastEvent.getRelevantObjectsForOneObject(relatedObject)){
				if(!indirectRelatedObjects.contains(indirectRelatedObject))
					indirectRelatedObjects.add(indirectRelatedObject);
			}
		}
		return indirectRelatedObjects;		
	}
	
	/**
	 * get all objects connected to related objects of an event + related objects of an event
	 */
	public List<XOCObject> getAllRelatedObjectsForOneEvent(XOCEvent event) {
		XOCEvent lastEvent = allEvents.get(allEvents.size()-1);
		List<XOCObject> allRelatedObjects = getIndirectRelatedObjectsForOneEvent(event);
		for(XOCObject object : lastEvent.getObjectModel().getObjectSet()){
			if(event.relatedObjectsContain(object))
				allRelatedObjects.add(object);		
		}
		
		return allRelatedObjects;
	}
	
	/**
	 * get target events correlated to an input event by a path (an array of class names)
	 * @param event
	 * @param path
	 * @param targetActivity
	 * @return
	 */
	public List<XOCEvent> getRelatedEventsForOneEventByPath(XOCEvent event, String[] path, String targetActivity) {
		List<XOCEvent> relatedEventList = new ArrayList<XOCEvent>();
		if(path.length >= 1){
			List<XOCObject> intermediaryObjectList = new ArrayList<XOCObject>();
			String startClassName = path[0];
			
			List<XOCObject> startObjectList = filter.filterObjects(event.getRelatedObjects(), startClassName);	
			for(XOCObject startObject : startObjectList){
				System.out.println("startObjectId: " + startObject.getAttributes().get(xConstants.ID).toString());
				intermediaryObjectList.add(startObject);
			}
			for(int i=1; i < path.length; i++){
				List<XOCObject> intermediaryObjectListTemp = new ArrayList<XOCObject>();
				String intermediaryClassName = path[i];
				System.out.println("intermediaryClassName: " + intermediaryClassName);
				for(XOCObject intermediaryObject : intermediaryObjectList){
					for(String relatedObjectId : intermediaryObject.getRelatedObjectIdList()){
						System.out.println("relatedObjectId: " + relatedObjectId);
						XOCObject relatedObject = getLastEvent().getObjectModel().getObjectById(relatedObjectId);
						String objectClassName = relatedObject.getAttributes().get(xConstants.CLASS).toString();
						if(objectClassName.equals(intermediaryClassName)){
							System.out.println("relatedObjectId of pointed class: " + relatedObjectId);
							intermediaryObjectListTemp.add(relatedObject);
						}						
					}
				}
				intermediaryObjectList.clear();
				for(XOCObject objectTemp : intermediaryObjectListTemp){
					System.out.println("objectTempId: " + objectTemp.getAttributes().get(xConstants.ID).toString());
					intermediaryObjectList.add(objectTemp);
				}
			}
			
			for(XOCObject intermediaryObject : intermediaryObjectList){
				for(XOCEvent relatedEvent : intermediaryObject.getRelatedEventList()){
					String activity = relatedEvent.getAttributes().get(xConstants.ACTIVITY).toString();
					System.out.println("activityId: " + relatedEvent.getAttributes().get(xConstants.ID).toString());
					if(activity.equals(targetActivity)){					
						if(!relatedEventList.contains(relatedEvent)){
							System.out.println("activityId of pointed activity: " + relatedEvent.getAttributes().get(xConstants.ID).toString());
							relatedEventList.add(relatedEvent);
						}					
					}						
				}
			}
		}
		
		return relatedEventList;
	}
	
	/**
	 * get target events correlated to an input object by a path (an array of class names);
	 * the first class in the path is the class of the input object;
	 * @param object
	 * @param path
	 * @param targetActivity
	 * @return
	 */
	public List<XOCEvent> getRelatedEventsForOneObjectByPath(XOCObject object, String[] path, String targetActivity) {
		List<XOCEvent> relatedEventList = new ArrayList<XOCEvent>();
		if(path.length >= 1){
			List<XOCObject> intermediaryObjectList = new ArrayList<XOCObject>();
			intermediaryObjectList.add(object);
//			String startClassName = path[0];
			
//			List<XOCObject> startObjectList = filter.filterObjects(event.getRelatedObjects(), startClassName);	
//			for(XOCObject startObject : startObjectList){
//				System.out.println("startObjectId: " + startObject.getAttributes().get(xConstants.ID).toString());
//				intermediaryObjectList.add(startObject);
//			}
			for(int i=1; i < path.length; i++){
				List<XOCObject> intermediaryObjectListTemp = new ArrayList<XOCObject>();
				String intermediaryClassName = path[i];
				for(XOCObject intermediaryObject : intermediaryObjectList){
					for(String relatedObjectId : intermediaryObject.getRelatedObjectIdList()){
						System.out.println("relatedObjectId: " + relatedObjectId);
						XOCObject relatedObject = getLastEvent().getObjectModel().getObjectById(relatedObjectId);
						String objectClassName = relatedObject.getAttributes().get(xConstants.CLASS).toString();
						if(objectClassName.equals(intermediaryClassName)){
							System.out.println("relatedObjectId of pointed class: " + relatedObjectId);
							intermediaryObjectListTemp.add(relatedObject);
						}						
					}
				}
				intermediaryObjectList.clear();
				for(XOCObject objectTemp : intermediaryObjectListTemp){
					System.out.println("objectTempId: " + objectTemp.getAttributes().get(xConstants.ID).toString());
					intermediaryObjectList.add(objectTemp);
				}
			}
			
			for(XOCObject intermediaryObject : intermediaryObjectList){
				for(XOCEvent relatedEvent : intermediaryObject.getRelatedEventList()){
					String activity = relatedEvent.getAttributes().get(xConstants.ACTIVITY).toString();
					System.out.println("activityId: " + relatedEvent.getAttributes().get(xConstants.ID).toString());
					if(activity.equals(targetActivity)){					
						if(!relatedEventList.contains(relatedEvent)){
							System.out.println("activityId of pointed activity: " + relatedEvent.getAttributes().get(xConstants.ID).toString());
							relatedEventList.add(relatedEvent);
						}					
					}						
				}
			}
		}
		
		return relatedEventList;
	}
	
	public List<XOCEvent> getRelatedEventsForOneEvent(XOCEvent event) {
		List<XOCEvent> relatedEvents = new ArrayList<XOCEvent>();
		List<XOCObject> allRelatedObjects = getAllRelatedObjectsForOneEvent(event);
		for(XOCEvent eventCheck : allEvents){
			for(XOCObject object : allRelatedObjects)
				if(eventCheck.relatedObjectsContain(object) && !eventCheck.equals(event))
					relatedEvents.add(eventCheck);
		}
		return relatedEvents;
	}

	/*
	 * get all related event by a class node name of one event
	 */
	public List<XOCEvent> getRelatedEventsForOneEventA(XOCEvent event, String classNodeName){
		List<XOCEvent> relatedEvents = new ArrayList<XOCEvent>();		
		List<XOCObject> allRelatedObjects = filter.filterObjects(event.getRelatedObjects(), classNodeName);	
		for(XOCEvent eventCheck : allEvents){
			for(XOCObject object : allRelatedObjects)
				if(eventCheck.relatedObjectsContain(object) && !eventCheck.equals(event) && !relatedEvents.contains(eventCheck)){
					relatedEvents.add(eventCheck);
				}
					
		}
		return relatedEvents;
	}
	
	/*
	 * get all related event by two class nodes of a class relation of one event
	 */
	public List<XOCEvent> getRelatedEventsForOneEventB(XOCEvent event, String classNodeNameX, String classNodeNameY){
		XOCEvent lastEvent = allEvents.get(allEvents.size()-1);
		List<XOCEvent> relatedEvents = new ArrayList<XOCEvent>();		
		List<XOCObject> allRelatedObjectsX = filter.filterObjects(event.getRelatedObjects(), classNodeNameX);
		for(XOCObject objectX : allRelatedObjectsX){
			List<XOCObject> allIndirectRelatedObjectY = filter.filterObjects(lastEvent.getRelevantObjectsForOneObject(objectX), classNodeNameY);
			for(XOCObject objectY : allIndirectRelatedObjectY){
				for(XOCEvent eventCheck : allEvents){					
					if(eventCheck.relatedObjectsContain(objectY) && !eventCheck.equals(event) && !relatedEvents.contains(eventCheck)){
						relatedEvents.add(eventCheck);
					}					
				}
			}
		}		
		
		List<XOCObject> allRelatedObjectsY = filter.filterObjects(event.getRelatedObjects(), classNodeNameY);
		for(XOCObject objectY : allRelatedObjectsY){
			List<XOCObject> allIndirectRelatedObjectX = filter.filterObjects(lastEvent.getRelevantObjectsForOneObject(objectY), classNodeNameX);
			for(XOCObject objectX : allIndirectRelatedObjectX){
				for(XOCEvent eventCheck : allEvents){					
					if(eventCheck.relatedObjectsContain(objectX) && !eventCheck.equals(event) && !relatedEvents.contains(eventCheck)){
						relatedEvents.add(eventCheck);
					}					
				}
			}
		}
		return relatedEvents;
	}
	
	public List<XOCEvent> createEventsForOneActivity(String activity) {
		List<XOCEvent> eventsOfOneActivity = new ArrayList<XOCEvent>();
		for(XOCEvent event : allEvents){
			if(event.getAttributes().get("activity").toString().equals(activity))
				eventsOfOneActivity.add(event);
				
		}
		return eventsOfOneActivity;
	}	
	
	public static void main(String[] str) throws Exception{
		XocXmlParser XOCParser = new XocXmlParser();
		XOCParser.parse(new File("e:\\normal.xoc"));		
		
		XOCLog inputLog = XOCParser.getLog();
		XOCLog cloneLog = (XOCLog) inputLog.clone();
		for(int i=0; i< inputLog.getAllEvents().size(); i++){
			XOCEvent originalEvent = inputLog.getAllEvents().get(i);
			XOCEvent cloneEvent = cloneLog.getAllEvents().get(i);
//			System.out.print("\n original id" + originalEvent.getAttributes().get("id").toString());
//			System.out.print("\n original address"+  originalEvent.hashCode());
//			System.out.print("\n 888888888888");
//			System.out.print("\n original id" + cloneEvent.getAttributes().get("id").toString());
//			System.out.print("\n original address"+  cloneEvent.hashCode());
			if(originalEvent == cloneEvent){
				System.out.print("\n event equal");
			}
			else
				System.out.print("\n event not equal");
			
			if(originalEvent == cloneEvent){
				System.out.print("\n event equal");
			}
			else
				System.out.print("\n event not equal");
		}
		
		System.out.print("\n original address"+  inputLog.hashCode());
		System.out.print("\n clone address"+  cloneLog.hashCode());
		
//		if(inputLog == cloneLog){
		if(inputLog == cloneLog){
			System.out.print("\n log equal");
		}
		else
			System.out.print("\n log not equal");
	}
	
	/**
	 *  create the complete object model for one input event and
	 *  fill it into the object model parameter
	 * @param event
	 * @return complete object model
	 */
	public XOCObjectModel createCompleteObjectModel(XOCEvent event){
		int currentEventIndex = event.getOrderNum();
//		System.out.println("\n currentEventIndex： " + currentEventIndex);
		XOCObjectModel objectModelComplete = new XOCObjectModelImpl();
		
		for(String id : idVSObject.keySet()){
			XOCObject object = idVSObject.get(id);
			//if the object is alive at the moment of event
			if(object.isAliveAtEvent(event)){
				//create object version, i.e., create the attributes at the moment of input event
				object.getAttributes().clear();
				XAttribute attribute = null;		
				System.out.println("in create complete model object****************");
				for(String key : object.getXOCAttributes().keySet()){
					XOCAttribute attributeVersion = object.getXOCAttributes().get(key);						
					String value = attributeVersion.getValueVersion(currentEventIndex);
					
//					System.out.println("in create complete model xoc attribute key: " + key);
//					System.out.println("in create complete model attributeVersion.getValueList().size(): " + attributeVersion.getValueList().size());
//					System.out.println("in create complete model xoc attribute value: " + value);
					//currently, if value is null, we hide the corresponding attribute
					if(value != null){
						attribute = factory.createAttributeLiteral(key, value, null);
						object.getAttributes().put(attribute.getKey(), attribute);
					}									
				}
				objectModelComplete.getObjectSet().add(object);
			}
		}
		
		for(String id : idVSObjectRelation.keySet()){
			XOCORelation objectRelation = idVSObjectRelation.get(id);
			//if the object relation is alive at the moment of event
			if(objectRelation.isAliveAtEvent(event)){
				objectModelComplete.getObjectRelationSet().add(objectRelation);
			}
		}
		
		return objectModelComplete;
	}

	/**
	 *  create the updated parts (added objects and updated objects) of the object model for one input event 
	 * @param event
	 * @return complete object model
	 */
	public XOCObjectModel createObjectModelUpdate(XOCEvent event){
		int currentEventIndex = event.getOrderNum();
//		System.out.println("\n currentEventIndex： " + currentEventIndex);
		XOCObjectModel objectModelUpdate = new XOCObjectModelImpl();
		
		for(String id : idVSObject.keySet()){
			XOCObject object = idVSObject.get(id);
			//if the object is alive at the moment of event
			if(object.isAddedAtEvent(event) || object.isUpdatedAtEvent(event)){
				//create object version, i.e., create the attributes at the moment of input event
				object.getAttributes().clear();
				XAttribute attribute = null;		
				System.out.println("in create complete model object****************");
				for(String key : object.getXOCAttributes().keySet()){
					XOCAttribute attributeVersion = object.getXOCAttributes().get(key);						
					String value = attributeVersion.getValueVersion(currentEventIndex);				
					//currently, if value is null, we hide the corresponding attribute
					if(value != null){
						attribute = factory.createAttributeLiteral(key, value, null);
						object.getAttributes().put(attribute.getKey(), attribute);
					}									
				}				
				objectModelUpdate.getObjectSet().add(object);
			}
			
			//record the life cycle attribute
			XAttribute attributeLifecycle = null;
			if(object.isAddedAtEvent(event)){
				object.getEventOrderNumVSObjectStateMap().put(currentEventIndex, xConstants.add);
				attributeLifecycle = factory.createAttributeLiteral(xConstants.lifecycle, xConstants.add, null);
				object.getAttributes().put(attributeLifecycle.getKey(), attributeLifecycle);
			}
			else if(object.isUpdatedAtEvent(event)){
				object.getEventOrderNumVSObjectStateMap().put(currentEventIndex, xConstants.update);
				System.out.println("object is updated: " + id);
				attributeLifecycle = factory.createAttributeLiteral(xConstants.lifecycle, xConstants.update, null);
				object.getAttributes().put(attributeLifecycle.getKey(), attributeLifecycle);
			}
			else if(object.isAliveAtEventOrderNum(currentEventIndex-1) && !object.isAliveAtEvent(event)){
				object.getEventOrderNumVSObjectStateMap().put(currentEventIndex, xConstants.delete);
				attributeLifecycle = factory.createAttributeLiteral(xConstants.lifecycle, xConstants.delete, null);
				object.getAttributes().put(attributeLifecycle.getKey(), attributeLifecycle);
			}		
		}
		
		for(String id : idVSObjectRelation.keySet()){
			XOCORelation objectRelation = idVSObjectRelation.get(id);
			//if the object relation is alive at the moment of event
			if(objectRelation.isAddedAtEvent(event)){
				objectModelUpdate.getObjectRelationSet().add(objectRelation);
				
				//record the life cycle attribute
				XAttribute attributeLifecycle = null;
				attributeLifecycle = factory.createAttributeLiteral(xConstants.lifecycle, xConstants.add, null);
				objectRelation.getAttributes().put(attributeLifecycle.getKey(), attributeLifecycle);
			}
		}		
		return objectModelUpdate;
	}
	
	public Map<String,XOCObject> getIdVSObjectMap() {
		return idVSObject;
	}
	
	public Map<String,XOCORelation> getIdVSObjectRelationMap() {
		return idVSObjectRelation;
	}
	
	public void setIdVSObjectMap(Map<String,XOCObject> idVSObject) {
		this.idVSObject = idVSObject;
	}
	
	public void setIdVSObjectRelationMap(Map<String,XOCORelation> idVSObjectRelation) {
		this.idVSObjectRelation = idVSObjectRelation;
	}
	
	public List<XOCObject> getObjectList() {
		return objectList;
	}
	
	public List<XOCORelation> getObjectRelationList() {
		return objectRelationList;
	}
	
	/**
	 * sort an input list of events based on their orders in the event log
	 * @param eventList
	 * @return List<XOCEvent> sortedEventList
	 */
	public List<XOCEvent> sortEventList(List<XOCEvent> eventList){
		List<XOCEvent> sortedEventList = new ArrayList<XOCEvent>();
		for(XOCEvent event : allEvents){
			if(eventList.contains(event)){
				sortedEventList.add(event);
			}
		}
		
		return sortedEventList;
	}
}
