/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.common.structure.IntegerPair;
import org.processmining.ocbc.log.extension.XExtension;
import org.processmining.ocbc.log.id.XID;
import org.processmining.ocbc.log.model.XAttributeMap;
import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.log.model.XOCORelation;
import org.processmining.ocbc.log.model.XOCObject;
import org.processmining.ocbc.models.visualization.cell.MyOriginalCell;


/**
 * Implementation for the XEvent interface.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * 
 */
public class XOCORelationImpl implements XOCORelation {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * ID of thi event.
	 */
//	private XID ID;

	/**
	 * Map of attributes for this event.
	 */
	private XAttributeMap attributes;
	private String relationType;
	private XOCObject sourceObject;
	private XOCObject targetObject;
	private MyOriginalCell correspondingCell;
//	private String birthEventId;
//	private String deathEventId;
//	private XOCEvent birthEvent;
//	private XOCEvent deathEvent;
	
	/**
	 * record the lifecycle period list of one object relation;
	 * each period consists of two values: start event order number (1..*) and end event order number;
	 * the object still (maybe last) observed in the object model of the end event (no overlapping)
	 */
	private List<IntegerPair> lifecyclePeriodList = new ArrayList<IntegerPair>();	
	
	/**
	 * record the state (add, delete, update) of one object relation at some event (identified by the event order number);
	 * if one event number is not in the map, it means (1) the object relation is not alive (after delete && before add if any), or 
	 * (2) remain the same (state value: remain) after add or update && before delete (if any)
	 */
	private Map<Integer, String> eventOrderNumVSObjectRelationState = new HashMap<Integer, String>();
	
	public XOCORelationImpl(String relationType) {
		this(new XID(), new XAttributeMapImpl(), relationType, null, null);
	}
	
	public XOCORelationImpl(String relationType, XOCObject sourceObject, XOCObject targetObject) {
		this(new XID(), new XAttributeMapImpl(), relationType, sourceObject, targetObject);
	}
	
	public XOCORelationImpl(XID ID, XAttributeMap attributes, String relationType, XOCObject sourceObject, XOCObject targetObject) {
//		this.ID = ID;
		this.attributes = attributes;
		this.relationType = relationType;		
		this.sourceObject = sourceObject;
		this.targetObject = targetObject;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.XAttributable#getAttributes()
	 */
	public String getRelationType() {
		return relationType;
	}

	public XOCObject getSourceObject() {
		return sourceObject;
	}
	
	public XOCObject getTargetObject() {
		return targetObject;
	}
	
//	public String getBirthEventId(){
//		return birthEventId;
//	}
//	
//	public String getDeathEventId(){
//		return deathEventId;
//	}

	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XAttributable#hasAttributes()
	 */
	@Override
	public boolean hasAttributes() {
		return !attributes.isEmpty();
	}
	


	/**
	 * Clones this event, i.e. creates a deep copy, but with a new ID, so equals
	 * does not hold between this and the clone
	 */
	public Object clone() {
		XOCORelationImpl clone;
		try {
			clone = (XOCORelationImpl) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return null;
		}
//		clone.ID = XIDFactory.instance().createId();
		clone.attributes = (XAttributeMap) attributes.clone();
		return clone;
	}

	/**
	 * Tests for equality of IDs
	 */
	public boolean equals(Object o) {
		if (o instanceof XOCORelationImpl) {
			return ((XOCORelationImpl) o).getAttributes().get(xConstants.ID).toString().
					equals(attributes.get(xConstants.ID).toString());
		} else {
			return false;
		}
	}

	/**
	 * Returns the hashCode of the id
	 */
//	public int hashCode() {
//		return ID.hashCode();
//	}

//	public XID getID() {
//		return ID;
//	}


	public Set<XExtension> getExtensions() {
		// TODO Auto-generated method stub
		return null;
	}

	public XAttributeMap getAttributes() {
		return attributes;
	}

//	public XOCEvent getBirthEvent(){
//		return birthEvent;
//	}
//	
//	public XOCEvent getDeathEvent(){
//		return deathEvent;
//	}
//	
//	public void setBirthEvent(XOCEvent birthEvent){
//		this.birthEvent = birthEvent;
//	}
//	
//	public void setDeathEvent(XOCEvent deathEvent){
//		this.deathEvent = deathEvent;
//	}
//	
//	public void setBirthEventId(String birthEventId){
//		this.birthEventId = birthEventId;
//	}
//	
//	public void setDeathEventId(String deathEventId){
//		this.deathEventId = deathEventId;
//	}
	
	
	public void setAttributes(XAttributeMap attributes) {
		this.attributes = attributes;		
	}

	public void setRelationType(String relationType) {
		this.relationType = relationType;		
	}

	public void setSourceObject(XOCObject sourceObject) {
		this.sourceObject = sourceObject;		
	}

	public void setTargetObject(XOCObject targetObject) {
		this.targetObject = targetObject;		
	}

	public boolean compareObjectRelation(XOCORelation objectRelation) {
		if(this.getAttributes().get("relation").toString().equals(objectRelation.getRelationType())  		
			&& this.sourceObject.getAttributes().get("id").toString().equals(objectRelation.getSourceObject().getAttributes().get("id").toString())  
			&& this.targetObject.getAttributes().get("id").toString().equals(objectRelation.getTargetObject().getAttributes().get("id").toString())){
			return true;
		}
		else return false;
	}
	
	public boolean isAttributesEqual(XOCORelation objectRelation){
		boolean equalFlag = true;		
		XAttributeMap inputAttributes = objectRelation.getAttributes();
		if(attributes.keySet().size() == inputAttributes.keySet().size()){
			for(String key : attributes.keySet()){
				if(inputAttributes.containsKey(key)){
					String value = attributes.get(key).toString();
					if(!value.equals(inputAttributes.get(key).toString())){
						equalFlag = false;
						break;
					}
				}
				else{
					equalFlag = false;
					break;
				}						
			}
		}
		else{
			equalFlag = false;
		}		
		return equalFlag;
	}
	
	public MyOriginalCell getCorrespondingCell(){
		return correspondingCell;
	}
	
	public void setCorrespondingCell(MyOriginalCell correspondingCell){
		this.correspondingCell = correspondingCell;
	}
	
	/**
	 * get the lifecycle period list of one object;
	 * each period consists of two values: start event order number and end event order number;
	 * the object still (maybe last) observed in the object model of the end event;
	 * @return
	 */
	public List<IntegerPair> getLifecyclePeriodList(){
		return lifecyclePeriodList;
	}
	
	/**
	 * check if the object relation is alive at the moment of input event based on the life cycle period list
	 * @param event
	 * @return
	 */
	public boolean isAliveAtEvent(XOCEvent event){
		int inputEventOrderNum = event.getOrderNum();
		for(IntegerPair period : lifecyclePeriodList){
			int startValue = period.getFirstValue();
			int endValue = period.getSecondValue() != null ? period.getSecondValue() : Integer.MAX_VALUE;
			if(startValue <= inputEventOrderNum && endValue >= inputEventOrderNum){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * check if the object relation is alive at the moment of input event order number based on the life cycle period list
	 * @param event
	 * @return
	 */
	public boolean isAliveAtEventOrderNum(int eventOrderNum){
		for(IntegerPair period : lifecyclePeriodList){
			int startValue = period.getFirstValue();
			int endValue = period.getSecondValue() != null ? period.getSecondValue() : Integer.MAX_VALUE;
			if(startValue <= eventOrderNum && endValue >= eventOrderNum){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * check if the object relation is added at the moment of input event based on the life cycle period list
	 * if one object relation is alive at current event and dead at previous event, the object relation is added at current event
	 * @param event
	 * @return
	 */
	public boolean isAddedAtEvent(XOCEvent event){
		int inputEventOrderNum = event.getOrderNum();
		if(isAliveAtEventOrderNum(inputEventOrderNum) && !isAliveAtEventOrderNum(inputEventOrderNum-1)){
			return true;
		}
		return false;
	}
	
	/**
	 * get the map between event order number and object relation state;
	 * if one event number is not in the map, it means (1) the object relation is not alive (after delete && before add if any), or 
	 * (2) remain the same (state value: remain) after add or update && before delete (if any)
	 * @return
	 */
	public Map<Integer, String> getEventOrderNumVSObjectRelationStateMap(){
		return eventOrderNumVSObjectRelationState;
	}
}
