/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.common.structure.IntegerPair;
import org.processmining.ocbc.log.extension.XExtension;
import org.processmining.ocbc.log.factory.XFactoryRegistry;
import org.processmining.ocbc.log.id.XID;
import org.processmining.ocbc.log.model.XAttribute;
import org.processmining.ocbc.log.model.XAttributeMap;
import org.processmining.ocbc.log.model.XOCAttribute;
import org.processmining.ocbc.log.model.XOCClass;
import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.log.model.XOCORelation;
import org.processmining.ocbc.log.model.XOCObject;
import org.processmining.ocbc.models.visualization.cell.MyOriginalCell;


/**
 * Implementation for object interface.
 * 
 * we embedded class function (mapping objects to classes) into object
 * each object has a (XOCClass) class attribute, and the attribute can be reset
 */
public class XOCObjectImpl implements XOCObject {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	/**
	 * ID of this event.
	 */
//	private XID ID;
	private String name;
	
	private List<XOCEvent> relatedEventList = new ArrayList<XOCEvent>();//record the events referring to the object
	private List<String> relatedObjectIdList = new ArrayList<String>();//record the objects which are connected to the object by object relations
	private List<XOCORelation> relatedObjectRelationList = new ArrayList<XOCORelation>();//record the object relations which are connected to the object 
	/**
	 * record the lifecycle period list of one object;
	 * each period consists of two values: start event order number (1..*) and end event order number;
	 * the object still (maybe last) observed in the object model of the end event (no overlapping)
	 */
	private List<IntegerPair> lifecyclePeriodList = new ArrayList<IntegerPair>();
	
	/**
	 * record the state (add, delete, update) of one object at some event (identified by the event order number);
	 * if one event number is not in the map, it means (1) the object is not alive (after delete && before add if any), or 
	 * (2) remain the same (state value: remain) after add or update && before delete (if any)
	 */
	private Map<Integer, String> eventOrderNumVSObjectState = new HashMap<Integer, String>();
	
	private String birthEventId;
	private String deathEventId;
	
	private XOCEvent observationEvent;//record the event when the object first is referred to
	private XOCEvent birthEvent;//record the event when the object first appears in the object model
	private XOCEvent deathEvent;//record the event when the object first disappears in the object model
	
	
	private int deathEventOrderNum = 0;
	
	/**
	 * Map of attributes for this event.
	 */
	private XAttributeMap attributes;
	private MyOriginalCell correspondingCell;
	
	private Map<String, XOCAttribute> attributeVersionMap = new HashMap<String, XOCAttribute>();
	public XOCObjectImpl(String name) {
		this(new XID(), name);
	}

	public XOCObjectImpl(String name, XOCClass Class) {
		this(new XID(), name, new XAttributeMapImpl(), Class);
	}
	
	public XOCObjectImpl(XID ID, String name) {
		this(ID, name, new XAttributeMapImpl(), new XOCClassImpl("class"));
	}
	
	/**
	 * Creates a new event.
	 * 
	 * @param attributes
	 *            Map of attribute for the event.
	 */
	public XOCObjectImpl(XAttributeMap attributes) {
		this(new XID(), attributes);
	}

	/**
	 * Creates a new event with the given id and attributed
	 * 
	 * @param id
	 *            the id for this event
	 * @param attributes
	 *            Map of attribute for the event.
	 */
	public XOCObjectImpl(XID ID, XAttributeMap attributes) {
		this(ID, "object", attributes, new XOCClassImpl("class"));
	}
	
	public XOCObjectImpl(XID ID, String name, XAttributeMap attributes, XOCClass Class) {
//		this.ID = ID;
		this.name = name;
		this.attributes = attributes;
//		this.Class = Class;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.XAttributable#getAttributes()
	 */
	public XAttributeMap getAttributes() {
		return attributes;
	}

	/*
	 * (non-Javadoc)
	 */
	public Map<String, XOCAttribute> getXOCAttributes() {
		return attributeVersionMap;
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.XAttributable#setAttributes(java.util.Map)
	 */
	public void setAttributes(XAttributeMap attributes) {
		this.attributes = attributes;
	}

	/*
	 * (non-Javadoc)
	 */
	public void setXOCAttributes(Map<String, XOCAttribute> attributeVersionMap) {
		this.attributeVersionMap = attributeVersionMap;
	}
	
	/**
	 * add an XOC attribute (key, value, event order number, i.e., moment) into the corresponding map
	 * @param key
	 * @param value
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void addXOCAttribute(String key, String value, int eventOrderNum, String logFormat){
		//if the attribute map does not contain the attribute
		if(!attributeVersionMap.containsKey(key)){
			XOCAttribute attributeVersion = new XOCAttributeImpl(key);
			attributeVersion.insertNewValueVersion(value, eventOrderNum, logFormat);
			attributeVersionMap.put(key, attributeVersion);
		}
		//if the attribute map contains the attribute
		else{
			XOCAttribute attributeVersion = attributeVersionMap.get(key);
			attributeVersion.addValueVersion(value, eventOrderNum, logFormat);
		}
	}
	
	/**
	 * end an XOC attribute, i.e., adding the ending event order number in the last value version corresponding to the input key
	 * @param key
	 * @param eventOrderNum
	 * @param logFormat
	 */
	public void endXOCAttribute(String key, int eventOrderNum, String logFormat){
		//if the attribute map contains the attribute
		if(attributeVersionMap.containsKey(key)){
			XOCAttribute attributeVersion = attributeVersionMap.get(key);
			attributeVersion.endLastValueVersion(eventOrderNum, logFormat);
		}
	}
	
	/**
	 * get the lifecycle period list of one object;
	 * each period consists of two values: start event order number and end event order number;
	 * the object still (maybe last) observed in the object model of the end event;
	 * @return
	 */
	public List<IntegerPair> getLifecyclePeriodList(){
		return lifecyclePeriodList;
	}
	
	/**
	 * check if the object is alive at the moment of input event based on the life cycle period list
	 * @param event
	 * @return
	 */
	public boolean isAliveAtEvent(XOCEvent event){
		int inputEventOrderNum = event.getOrderNum();
		for(IntegerPair period : lifecyclePeriodList){
			int startValue = period.getFirstValue();
			int endValue = period.getSecondValue() != null ? period.getSecondValue() : Integer.MAX_VALUE;
			if(startValue <= inputEventOrderNum && endValue >= inputEventOrderNum){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * check if the object is alive at the moment of input event order number based on the life cycle period list
	 * @param event
	 * @return
	 */
	public boolean isAliveAtEventOrderNum(int inputEventOrderNum){
		for(IntegerPair period : lifecyclePeriodList){
			int startValue = period.getFirstValue();
			int endValue = period.getSecondValue() != null ? period.getSecondValue() : Integer.MAX_VALUE;
			if(startValue <= inputEventOrderNum && endValue >= inputEventOrderNum){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * check if the object is updated at the moment of input event based on the life cycle period list
	 * if one object is alive and has different attributes at previous and current events, the object is updated
	 * @param event
	 * @return
	 */
	public boolean isUpdatedAtEvent(XOCEvent event){
		int inputEventOrderNum = event.getOrderNum();
//		System.out.println("enter isUpdatedAtEvent");
		if(isAliveAtEventOrderNum(inputEventOrderNum) && isAliveAtEventOrderNum(inputEventOrderNum-1)){
			for(String key : attributeVersionMap.keySet()){
				if(!key.equals(xConstants.lifecycle)){
					String currentValue = attributeVersionMap.get(key).getValueVersion(inputEventOrderNum);
					String previousValue = attributeVersionMap.get(key).getValueVersion(inputEventOrderNum-1);
					if(previousValue == null){
						previousValue = "null";
					}
					if(currentValue == null){
						currentValue = "null";
					}
					if(currentValue != previousValue){
//						System.out.println("currentValue: " + currentValue);
//						System.out.println("previousValue: " + previousValue);
						return true;
					}
				}
			}	
		}
		return false;
	}
	
	/**
	 * check if the object is born at the moment of input event based on the life cycle period list
	 * if one object is alive at current event and dead at previous event, the object is added at current event
	 * @param event
	 * @return
	 */
	public boolean isAddedAtEvent(XOCEvent event){
		int inputEventOrderNum = event.getOrderNum();
		if(isAliveAtEventOrderNum(inputEventOrderNum) && !isAliveAtEventOrderNum(inputEventOrderNum-1)){
			return true;
		}
		return false;
	}
	
	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XAttributable#hasAttributes()
	 */
	@Override
	public boolean hasAttributes() {
		return !attributes.isEmpty();
	}	

	/**
	 * Clones this event, i.e. creates a deep copy, but with a new ID, so equals
	 * does not hold between this and the clone
	 */
	public Object clone() {
		XOCObjectImpl clone;

		try {
			clone = (XOCObjectImpl) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return null;
		}
//		clone.ID = (XID) this.ID.clone();
		clone.attributes = (XAttributeMap) this.attributes.clone();
		return clone;
	}


	/**
	 * Returns the hashCode of the id
	 */
//	public int hashCode() {
//		return ID.hashCode();
//	}

	/*
	 * (non-Javadoc)
	 * @see org.deckfour.xes.model.XEvent#getID()
	 */
//	public XID getID() {
//		return ID;
//	}

	/**
	 * Sets the ID. Should only be used for deserialization purposes
	 * 
	 * @param id
	 *            the new id.
	 */
//	public void setID(XID ID) {
//		this.ID = ID;
//	}

	public Set<XExtension> getExtensions() {
		// TODO Auto-generated method stub
		return null;
	}

//	public XOCClass getClassType(){
//		return Class;
//	}

//	public void setClassType(XOCClass Class) {
//		this.Class = Class;
//		
//	}

	public String getName() {
		return name;
	}

	
//	public String getBirthEventId(){
//		return birthEventId;
//	}
//	
//	public String getDeathEventId(){
//		return deathEventId;
//	}
		
	public XOCEvent getReferenceEvent(){
		return observationEvent;
	}
	
//	public XOCEvent getBirthEvent(){
//		return birthEvent;
//	}
//	
//	public XOCEvent getDeathEvent(){
//		return deathEvent;
//	}	
//	
//	/**
//	 * Get the death event order number i (i starts from 1, i.e., corresponding event is log.get(i-1))
//	 */
//	public int getDeathEventOrderNum(){
//		return deathEventOrderNum;
//	}
	
	public List<XOCEvent> getRelatedEventList(){
		return relatedEventList;
	}
	
	public List<String> getRelatedObjectIdList(){
		return relatedObjectIdList;
	}
	
	/**
	 * return the list of related object relations
	 * note: this can save memory than returning the list of ids
	 */
	public List<XOCORelation> getRelatedObjectRelationList(){
		return relatedObjectRelationList;
	}
	
	public void setName(String name) {
		this.name = name;		
	}
	
//	public void setBirthEventId(String birthEventId){
//		this.birthEventId = birthEventId;
//	}
//	
//	public void setDeathEventId(String deathEventId){
//		this.deathEventId = deathEventId;
//	}

	public void setReferenceEvent(XOCEvent observationEvent){		
		this.observationEvent = observationEvent;
	}
	
//	public void setBirthEvent(XOCEvent birthEvent){
//		this.birthEvent = birthEvent;
//	}
//	
//	public void setDeathEvent(XOCEvent deathEvent){
//		this.deathEvent = deathEvent;
//	}
//	
//	/**
//	 * Set the death event order number i (i starts from 1, i.e., corresponding event is log.get(i-1))
//	 * @param deathEventOrderNum
//	 */
//	public void setDeathEventOrderNum(int deathEventOrderNum){
//		this.deathEventOrderNum = deathEventOrderNum;
//	}
	
	public MyOriginalCell getCorrespondingCell(){
		return correspondingCell;
	}
	
	public void setCorrespondingCell(MyOriginalCell correspondingCell){
		this.correspondingCell = correspondingCell;
	}

	public boolean isAttributesEqual(XOCObject object){
		boolean equalFlag = true;		
		XAttributeMap inputAttributes = object.getAttributes();
		if(attributes.keySet().size() == inputAttributes.keySet().size()){
			for(String key : attributes.keySet()){
				if(inputAttributes.containsKey(key)){
					String value = attributes.get(key).toString();
					if(!value.equals(inputAttributes.get(key).toString())){
						equalFlag = false;
						break;
					}
				}
				else{
					equalFlag = false;
					break;
				}						
			}
		}
		else{
			equalFlag = false;
		}		
		return equalFlag;
	}
	
	/**
	 * minus the overlap part (attribute name and value are the same) between the
	 * current object and the input object from the current object;
	 * keep the current object id
	 */
	public XOCObject minusAttributeOverlap(XOCObject object){	
		XAttributeMap attributeDifference = new XAttributeMapImpl();
		XAttributeMap inputAttributes = object.getAttributes();
		for(String key : attributes.keySet()){
			if(inputAttributes.containsKey(key)){
				String value = attributes.get(key).toString();
				if(!value.equals(inputAttributes.get(key).toString())){
					attributeDifference.put(key, attributes.get(key));
				}
			}
			else{
				attributeDifference.put(key, attributes.get(key));
			}						
		}	
		attributeDifference.put(xConstants.ID, attributes.get(xConstants.ID));
		this.setAttributes(attributeDifference);
		return this;
	}
	
	/**
	 *plus the different attributes (attribute name are not the same) between the
	 * current object and the input object to the current object;
	 */
	public XOCObject plusAttributes(XOCObject object){	
		XAttributeMap attributePlus = new XAttributeMapImpl();
		XAttributeMap inputAttributes = object.getAttributes();
		for(String key : inputAttributes.keySet()){
			if(!attributes.containsKey(key)){
				attributePlus.put(key, inputAttributes.get(key));
			}
			else{
				attributePlus.put(key, attributes.get(key));
			}
		}	
		this.setAttributes(attributePlus);
		return this;
	}
	
	/**
	 * get the map between event order number and object state;
	 * if one event number is not in the map, it means (1) the object is not alive (after delete && before add if any), or 
	 * (2) remain the same (state value: remain) after add or update && before delete (if any)
	 * @return
	 */
	public Map<Integer, String> getEventOrderNumVSObjectStateMap(){
		return eventOrderNumVSObjectState;
	}
	
	/**
	 * create the updated attribute map through comparing values of the input event with values of its previous event;
	 * @param event
	 * @return
	 */
	public XAttributeMap createUpdatedAttributeMap(XOCEvent event){
		XAttributeMap updatedAttributeMap = new XAttributeMapImpl();
		for(String key : attributeVersionMap.keySet()){
			String currentValue = attributeVersionMap.get(key).getValueVersion(event.getOrderNum());
			if(event.getOrderNum() == 1){				
				XAttribute attribute = XFactoryRegistry.instance().currentDefault().createAttributeLiteral(key, currentValue, null);
				updatedAttributeMap.put(key, attribute);
			}
			else{
				String previousValue = attributeVersionMap.get(key).getValueVersion(event.getOrderNum()-1);
				if(currentValue == null){
					System.out.println("currentValue is null in createUpdatedAttributeMap!");
					currentValue = "null";
				}
				if(previousValue == null){
					System.out.println("previousValue is null in createUpdatedAttributeMap!");
					previousValue = "null";
				}
				if(!currentValue.equals(previousValue)){
					XAttribute attribute = XFactoryRegistry.instance().currentDefault().createAttributeLiteral(key, currentValue, null);
					updatedAttributeMap.put(key, attribute);
				}
			}			
		}
		return updatedAttributeMap;
	}
}
