/*
 * OpenXES
 * 
 * The reference implementation of the XES meta-model for event 
 * log data management.
 * 
 * Copyright (c) 2008 Christian W. Guenther (christian@deckfour.org)
 * 
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * The use of this software can also be conditionally licensed for
 * other programs, which do not satisfy the specified conditions. This
 * requires an exemption from the general license, which may be
 * granted on a per-case basis.
 * 
 * If you want to license the use of this software with a program
 * incompatible with the LGPL, please contact the author for an
 * exemption at the following email address: 
 * christian@deckfour.org
 * 
 */
package org.processmining.ocbc.log.model.impl;

import java.util.Set;

import org.processmining.ocbc.log.extension.XExtension;
import org.processmining.ocbc.log.id.XID;
import org.processmining.ocbc.log.id.XIDFactory;
import org.processmining.ocbc.log.model.XAttributeMap;
import org.processmining.ocbc.log.model.XOCEvent;
import org.processmining.ocbc.log.model.XOCObject;
import org.processmining.ocbc.log.model.XOCRRelation;
import org.processmining.ocbc.models.visualization.cell.MyOriginalCell;


/**
 * Implementation for the XEvent interface.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * 
 */
public class XOCRRelationImpl implements XOCRRelation {

	/**
	 * ID of thi event.
	 */
	private XID ID;

	/**
	 * Map of attributes for this event.
	 */
	private XAttributeMap attributes;
	private String relationType;
	private XOCEvent sourceEvent;
	private XOCObject targetObject;
	private MyOriginalCell correspondingCell;

	
	public XOCRRelationImpl(String relationType) {
		this(new XID(), new XAttributeMapImpl(), relationType, null, null);
	}
	
	public XOCRRelationImpl(String relationType, XOCEvent sourceEvent, XOCObject targetObject) {
		this(new XID(), new XAttributeMapImpl(), relationType, sourceEvent, targetObject);
	}
	
	public XOCRRelationImpl(XID ID, XAttributeMap attributes, String relationType, XOCEvent sourceEvent, XOCObject targetObject) {
		this.ID = ID;
		this.attributes = attributes;
		this.relationType = relationType;		
		this.sourceEvent = sourceEvent;
		this.targetObject = targetObject;		
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.XAttributable#getAttributes()
	 */
	public String getRelationType() {
		return relationType;
	}

	public XOCEvent getSourceEvent() {
		return sourceEvent;
	}
	
	public XOCObject getTargetObject() {
		return targetObject;
	}
	

	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XAttributable#hasAttributes()
	 */
	@Override
	public boolean hasAttributes() {
		return !attributes.isEmpty();
	}
	


	/**
	 * Clones this event, i.e. creates a deep copy, but with a new ID, so equals
	 * does not hold between this and the clone
	 */
	public Object clone() {
		XOCRRelationImpl clone;
		try {
			clone = (XOCRRelationImpl) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return null;
		}
		clone.ID = XIDFactory.instance().createId();
		clone.attributes = (XAttributeMap) attributes.clone();
		return clone;
	}

	/**
	 * Tests for equality of IDs
	 */
	public boolean equals(Object o) {
		if (o instanceof XOCRRelationImpl) {
			return ((XOCRRelationImpl) o).ID.equals(ID);
		} else {
			return false;
		}
	}

	/**
	 * Returns the hashCode of the id
	 */
	public int hashCode() {
		return ID.hashCode();
	}

	public XID getID() {
		return ID;
	}


	public Set<XExtension> getExtensions() {
		// TODO Auto-generated method stub
		return null;
	}

	public XAttributeMap getAttributes() {
		return attributes;
	}

	public void setAttributes(XAttributeMap attributes) {
		this.attributes = attributes;		
	}

	public void setRelationType(String relationType) {
		this.relationType = relationType;		
	}

	public void setSourceEvent(XOCEvent sourceEvent) {
		this.sourceEvent = sourceEvent;		
	}

	public void setTargetObject(XOCObject targetObject) {
		this.targetObject = targetObject;		
	}

	public boolean compareReferenceRelation(XOCRRelation referenceRelation) {
		if(this.getAttributes().get("relation").toString().equals(referenceRelation.getRelationType())  		
			&& this.sourceEvent.getAttributes().get("id").toString().equals(referenceRelation.getSourceEvent().getAttributes().get("id").toString())  
			&& this.targetObject.getAttributes().get("id").toString().equals(referenceRelation.getTargetObject().getAttributes().get("id").toString())){
			return true;
		}
		else return false;
	}
	
	public MyOriginalCell getCorrespondingCell(){
		return correspondingCell;
	}
	
	public void setCorrespondingCell(MyOriginalCell correspondingCell){
		this.correspondingCell = correspondingCell;
	}

}
