package org.processmining.ocbc.performance.logprojection;

import gnu.trove.list.TDoubleList;
import gnu.trove.list.TIntList;
import gnu.trove.list.TLongList;
import gnu.trove.list.array.TDoubleArrayList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.list.array.TLongArrayList;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TObjectIntHashMap;
import gnu.trove.set.TDoubleSet;
import gnu.trove.set.TIntSet;
import gnu.trove.set.TLongSet;
import gnu.trove.set.hash.TDoubleHashSet;
import gnu.trove.set.hash.TIntHashSet;
import gnu.trove.set.hash.TLongHashSet;

import java.io.Serializable;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeID;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.jfree.chart.util.RelativeDateFormat;
import org.processmining.ocbc.performance.logprojection.calculators.XLightLiteralAttribute;
import org.processmining.ocbc.performance.logprojection.calculators.XLightTimestampAttribute;

public abstract class DataManager<T extends XAttributable> implements Serializable {

	private static final long serialVersionUID = 7999782581473915686L;
	//	private static final long MASK1 = 0x00000000001FFFFFL;
	//	private static final long MASK2 = 0x000003FFFFE00000L;
	//	private static final long MASK3 = 0xFFFFFC0000000000L;

	//	private static final long XORMASK = 0x000002AAAAFFFFFFL;
	protected static final long MAXHASH = 0xFFFFFFFFFFFFFFFFL;
	private static final DateFormat DF_ABS = new SimpleDateFormat();
	private static final DateFormat DF_REL = new RelativeDateFormat();

	protected static final XFactory FACTORY = XFactoryRegistry.instance().currentDefault();

	protected StringHandler stringHandler;

	/**
	 * Utility class to add attributes to all objects. During computation,
	 * access to the header information is provided for each object.
	 * 
	 * Implementations may assume that objects are passed in the order in which
	 * they are stored in the manager.
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static interface AttributeValueCalculator<T extends XAttributable> {
		public XAttribute[] computeAttributes(T object, int objectIndex, int... header);
	};

	public static interface StringHandler extends Serializable {
		public static final char[] UNKNOWN = "UNKNOWN by LogView".toCharArray();

		public int addLiteral(String value, TObjectIntMap<String> lit2index);

		public String getLiteral(int index);

		public char[] getLiteralChars(int index);
	}

	// //////////////////////////////////////////////////////////////////////
	// Stores a mapping from each global key to an index in the object's size.
	//
	protected final TObjectIntMap<Attribute> key2index = new TObjectIntHashMap<Attribute>(10, 0.5f);
	protected Attribute[] keys;
	protected int[] type2count = new int[6];

	// the size is determined by the global attributes
	protected int usedBytesPerObject;
	protected int bytesReservedPerObject;
	protected int blockSize;
	protected int blocks;
	protected byte[][] store;
	protected int numberOfObjects;
	protected final int headerSize;
	protected final boolean handlesEventAttributes;

	protected int shift;
	protected int mask;

	// //////////////////////////////////////////////////////////////////////

	// protected DataManager(StringHandler stringHandler, int headerSize) {
	// this.stringHandler = stringHandler;
	// this.headerSize = headerSize;
	// }

	public DataManager(StringHandler stringHandler, Collection<XAttribute> attributes, TObjectIntMap<String> lit2index,
			int preferredBlockSize, int headerSize, boolean handlesEventAttributes) {
		this.stringHandler = stringHandler;
		this.headerSize = headerSize;
		this.handlesEventAttributes = handlesEventAttributes;

		usedBytesPerObject = buildAttributeIndices(attributes, 4 * headerSize, key2index, type2count, lit2index);

		// We need at least byteSize bytes for each object.
		// round this up to the nearest power of 2. This allows for
		// significantly faster lookup times.
		shift = 0;
		int s = 1;
		while (s < usedBytesPerObject) {
			s <<= 1;
			shift++;
		}
		bytesReservedPerObject = s;
		// for each object the number of bytes is a power of 2 (2^shift)

		int objectsPerBlock = 1;
		while (bytesReservedPerObject * objectsPerBlock < preferredBlockSize) {
			objectsPerBlock <<= 1;
			shift++;
		}

		// For each block, the number of objects is a power of two

		blockSize = objectsPerBlock * bytesReservedPerObject;
		mask = blockSize - 1;

		// shift can be used to retrieve the block for a given index and
		// mask to get the index in the block
		assert 1 << shift == blockSize;

		blocks = 1;

		store = new byte[1][blockSize];
		// we start writing at index 'pointer'
		numberOfObjects = 0;

	}

	public boolean getAttributeValueBoolean(int object, Attribute key) {
		return getAttributeValueBoolean(object, key2index.get(key));
	}

	public boolean getAttributeValueBoolean(int object, int attrIndex) {
		int objectIndex = object * bytesReservedPerObject;
		return readBoolean(store[objectIndex >>> shift], (objectIndex & mask) + attrIndex);
	}

	public long getAttributeValueDiscrete(int object, Attribute key) {
		return getAttributeValueDiscrete(object, key2index.get(key));
	}

	public long getAttributeValueDiscrete(int object, int attrIndex) {
		int objectIndex = object * bytesReservedPerObject;
		return readLong(store[objectIndex >>> shift], (objectIndex & mask) + attrIndex);
	}

	public double getAttributeValueContinuous(int object, Attribute key) {
		return getAttributeValueContinuous(object, key2index.get(key));
	}

	public double getAttributeValueContinuous(int object, int attrIndex) {
		int objectIndex = object * bytesReservedPerObject;
		return readDouble(store[objectIndex >>> shift], (objectIndex & mask) + attrIndex);
	}

	public long getAttributeValueTimestamp(int object, Attribute key) {
		return getAttributeValueTimestamp(object, key2index.get(key));
	}

	public long getAttributeValueTimestamp(int object, int attrIndex) {
		int objectIndex = object * bytesReservedPerObject;
		return readLong(store[objectIndex >>> shift], (objectIndex & mask) + attrIndex);
	}

	public String getAttributeValueLiteral(int object, Attribute key) {
		return getAttributeValueLiteral(object, key2index.get(key));
	}

	public String getAttributeValueLiteral(int object, int attrIndex) {
		int objectIndex = object * bytesReservedPerObject;
		int litIndex = readInt(store[objectIndex >>> shift], (objectIndex & mask) + attrIndex);
		return stringHandler.getLiteral(litIndex);
	}

	public int getAttributeValueLiteralIndex(int object, Attribute key) {
		return getAttributeValueLiteralIndex(object, key2index.get(key));
	}

	public int getAttributeValueLiteralIndex(int object, int attrIndex) {
		int objectIndex = object * bytesReservedPerObject;
		return readInt(store[objectIndex >>> shift], (objectIndex & mask) + attrIndex);
	}

	protected int writeInt(byte[] array, int index, int value) {
		array[index] = (byte) (value >>> 24);
		array[++index] = (byte) (value >>> 16);
		array[++index] = (byte) (value >>> 8);
		array[++index] = (byte) (value);
		return ++index;
	}

	protected int readInt(byte[] array, int index) {
		return ((array[index] & 0xFF) << 24) + ((array[index + 1] & 0xFF) << 16) + ((array[index + 2] & 0xFF) << 8)
				+ ((array[index + 3] & 0xFF));
	}

	protected int writeShort(byte[] array, int index, short value) {
		array[index] = (byte) (value >>> 8);
		array[++index] = (byte) (value);
		return ++index;
	}

	protected short readShort(byte[] array, int index) {
		return (short) ((array[index] << 8) + ((array[index + 1] & 0xFF)));
	}

	protected int writeBoolean(byte[] array, int index, boolean value) {
		array[index] = (byte) (value ? 1 : 0);
		return ++index;
	}

	protected boolean readBoolean(byte[] array, int index) {
		return array[index] == 1;
	}

	protected int writeLong(byte[] array, int index, long value) {
		index = writeInt(array, index, (int) (value >>> 32));
		index = writeInt(array, index, (int) (value));
		return index;
	}

	protected long readLong(byte[] array, int index) {
		long r1 = readInt(array, index) & 0xFFFFFFFFL;
		r1 = r1 << 32;
		int r2 = readInt(array, index + 4);
		return r1 + (r2 & 0xFFFFFFFFL);
	}

	protected int writeDouble(byte[] array, int index, double value) {
		writeLong(array, index, Double.doubleToLongBits(value));
		return index + 8;
	}

	protected double readDouble(byte[] array, int index) {
		return Double.longBitsToDouble(readLong(array, index));
	}

	protected int buildAttributeIndices(Collection<? extends XAttribute> attributes, int index,
			TObjectIntMap<Attribute> key2index, int[] key2count, TObjectIntMap<String> lit2index) {

		for (XAttribute globalAttribute : attributes) {
			int size = 0;
			if (globalAttribute instanceof XAttributeID) {
				continue;
			}

			long value;
			if (globalAttribute instanceof XAttributeBoolean) {
				// use 1 byte to store the boolean
				size = 1;
				value = ((XAttributeBoolean) globalAttribute).getValue() ? 1 : 0;
			} else if (globalAttribute instanceof XAttributeContinuous) {
				// use 8 bytes to store a double
				size = 8;
				value = Double.doubleToLongBits(((XAttributeContinuous) globalAttribute).getValue());
			} else if (globalAttribute instanceof XAttributeDiscrete) {
				// use 8 bytes to store a long
				size = 8;
				value = ((XAttributeDiscrete) globalAttribute).getValue();
			} else if (globalAttribute instanceof XAttributeTimestamp) {
				// use 8 bytes to store a long
				size = 8;
				value = ((XAttributeTimestamp) globalAttribute).getValueMillis();
			} else if (globalAttribute instanceof XAttributeLiteral) {
				// use 4 bytes to store an integer pointer into the array of
				// literals
				size = 4;
				String val = ((XAttributeLiteral) globalAttribute).getValue();
				value = stringHandler.addLiteral(val, lit2index);
			} else {
				throw new RuntimeException("Unknown Attribute Type Encountered");
			}

			// reserve 'size' bytes to store the information of attribute 'key'
			// in the storage
			Attribute attribute = new Attribute(globalAttribute, value, handlesEventAttributes);

			key2index.put(attribute, index);
			key2count[attribute.getType()]++;
			//			keyIndex2hash.put(index, MAXHASH);
			index += size;

		}
		this.keys = key2index.keys(new Attribute[key2index.size()]);
		Arrays.sort(keys);
		return index;
	}

	//	private long newHash(long oldHash, long newHash) {
	//		long h1 = oldHash & MASK1;
	//		if ((newHash & MASK1) < h1) {
	//			h1 = newHash & MASK1;
	//		}
	//		long h2 = oldHash & MASK2;
	//		// no need for bitshift as we remain in the positive range.
	//		if (((newHash & MASK2) >>> 21) < h2 >>> 21) {
	//			h2 = newHash & MASK2;
	//		}
	//		long h3 = oldHash & MASK3;
	//		if ((newHash >>> 42) < h3 >>> 42) {
	//			h3 = newHash & MASK3;
	//		}
	//		long hash = h3 | h2 | h1;
	//
	//		return hash;
	//	}

	//	public final static int SEED1 = 0x9747b28c;
	//	public final static int SEED2 = 0x117a1465;

	//	private long hash(char[] data) {
	//		return hash64(data);
	//	}
	//
	//	private long hash(long value) {
	//
	//		return hash(new char[] { (char) ((value >>> 24) & 0xff), (char) ((value >>> 16) & 0xff),
	//				(char) ((value >>> 8) & 0xff), (char) ((value) & 0xff) });
	//
	//	}

	//	/**
	//	 * Generates 64 bit hash from byte array of the given length and seed. Uses
	//	 * straight up Murmur3 hash function
	//	 * 
	//	 * @param data
	//	 *            byte array to hash
	//	 * @param length
	//	 *            length of the array to hash
	//	 * @param seed
	//	 *            initial seed value
	//	 * @return 64 bit hash of the given array
	//	 */
	//	public long hash64(final char[] data) {
	//		int length = data.length;
	//		int seed = 0xe17a1465;
	//
	//		final long m = 0xc6a4a7935bd1e995L;
	//		final int r = 47;
	//
	//		long h = (seed & 0xffffffffl) ^ (length * m);
	//
	//		int length8 = length / 8;
	//
	//		for (int i = 0; i < length8; i++) {
	//			final int i8 = i * 8;
	//			long k = ((long) data[i8 + 0] & 0xff) + (((long) data[i8 + 1] & 0xff) << 8)
	//					+ (((long) data[i8 + 2] & 0xff) << 16) + (((long) data[i8 + 3] & 0xff) << 24)
	//					+ (((long) data[i8 + 4] & 0xff) << 32) + (((long) data[i8 + 5] & 0xff) << 40)
	//					+ (((long) data[i8 + 6] & 0xff) << 48) + (((long) data[i8 + 7] & 0xff) << 56);
	//
	//			k *= m;
	//			k ^= k >>> r;
	//			k *= m;
	//
	//			h ^= k;
	//			h *= m;
	//		}
	//
	//		switch (length & 7) {
	//			case 7 :
	//				h ^= (long) (data[(length & ~7) + 6] & 0xff) << 48;
	//			case 6 :
	//				h ^= (long) (data[(length & ~7) + 5] & 0xff) << 40;
	//			case 5 :
	//				h ^= (long) (data[(length & ~7) + 4] & 0xff) << 32;
	//			case 4 :
	//				h ^= (long) (data[(length & ~7) + 3] & 0xff) << 24;
	//			case 3 :
	//				h ^= (long) (data[(length & ~7) + 2] & 0xff) << 16;
	//			case 2 :
	//				h ^= (long) (data[(length & ~7) + 1] & 0xff) << 8;
	//			case 1 :
	//				h ^= (long) (data[length & ~7] & 0xff);
	//				h *= m;
	//		}
	//
	//		h ^= h >>> r;
	//		h *= m;
	//		h ^= h >>> r;
	//
	//		// to avoid a h value of 0 is used three times as 0, we mask the result
	//		h ^= XORMASK;
	//
	//		return h;
	//	}

	protected void checkSizeForInsertion() {
		// size * byteSize is the current number of bytes in use
		// blockSize * blocks is the current available bytes.
		if (numberOfObjects * bytesReservedPerObject == blockSize * blocks) {
			if (blocks == store.length) {
				if (blocks > Integer.MAX_VALUE / 2) {
					throw new RuntimeException("Storage Full");
				}
				// double the storage.
				byte[][] oldStore = store;
				store = new byte[2 * blocks][];
				System.arraycopy(oldStore, 0, store, 0, blocks);

			}
			// we need to add a block
			store[blocks] = new byte[blockSize];
			blocks++;
		}
	}

	// writes the contents of one object. The assumed headerdata is a number
	// integers, as specified in the constructor.
	public void addObject(T object, TObjectIntMap<String> lit2index, Set<String> errorKeys, int... headerData) {
		checkSizeForInsertion();

		int block = (numberOfObjects * bytesReservedPerObject) >>> shift;
		int ptr = (numberOfObjects * bytesReservedPerObject) & mask;

		byte[] array = store[block];

		assert headerData.length == headerSize;
		for (int i = 0; i < headerData.length; i++) {
			// write the header data
			writeInt(array, ptr + i * 4, headerData[i]);
		}

		for (int i = 0; i < keys.length;) {
			i = writeAttributeContents(object.getAttributes(), array, ptr, i, key2index, lit2index, errorKeys);
		}
		numberOfObjects++;
	}

	public void updateHeader(int number, int... headerData) {
		int block = (number * bytesReservedPerObject) >>> shift;
		int ptr = (number * bytesReservedPerObject) & mask;

		byte[] array = store[block];

		assert headerData.length == headerSize;
		for (int i = 0; i < headerData.length; i++) {
			// write the header data
			writeInt(array, ptr + i * 4, headerData[i]);
		}

	}

	public T readObject(int number) {
		return readObjectFromBlock(store[(number * bytesReservedPerObject) / blockSize],
				(number * bytesReservedPerObject) & mask, false);

	}

	protected abstract T readObjectFromBlock(byte[] block, int number, boolean forAddingAttributes);

	protected int writeAttributeContents(XAttributeMap attributes, byte[] storage, int ptr, int keyIndex,
			TObjectIntMap<Attribute> key2Index, TObjectIntMap<String> lit2index, Set<String> errorKeys) {

		Attribute key = keys[keyIndex];
		String skey = key.getKey();
		int val = key2Index.get(key);
		byte type = key.getType();

		int index = ptr + val;

		//		long hash;
		XAttribute attribute = attributes.get(skey);
		if (attribute == null) {
			errorKeys.add(skey);
			// globally defined attribute not present. Use default value!
			switch (type) {
				case Attribute.BOOL :
					writeBoolean(storage, index, key.getDefaultValue() == 1);
					// hash irrelevant
					//					hash = MAXHASH;
					break;
				case Attribute.DISC :
				case Attribute.TIME_REL :
				case Attribute.TIME_ABS :
				case Attribute.CONT :
					//					hash = hash(key.getDefaultValue());
					writeLong(storage, index, key.getDefaultValue());
					break;
				case Attribute.LITR :
					//					hash = hash(key.getDefaultValue());
					writeInt(storage, index, stringHandler.addLiteral(skey, lit2index));
					break;
				default :
					throw new RuntimeException("Unexpected Attribute Type present in the log");
			}
			//			key2hash.put(keyIndex, newHash(key2hash.get(keyIndex), hash));
		} else {
			try {
				writeAttribute(attribute, storage, index, val, type, lit2index);
			} catch (ClassCastException e) {
				// There is a problem with one of the attributes. There's nothing I can do here other than 
				// rethrow the exception. The data is simply inconsistent.
				throw new RuntimeException("Attribute " + skey + " is declared as a global of type "
						+ Attribute.toString(type) + " but used using a wrong type in the log.", e);
			}

		}
		keyIndex++;
		return keyIndex;
	}

	private void writeAttribute(XAttribute attribute, byte[] storage, int index, int keyIndex, byte type,
			TObjectIntMap<String> lit2index) {
		//		long hash;
		long val;
		int litIndex;
		switch (type) {
			case Attribute.BOOL :
				writeBoolean(storage, index, ((XAttributeBoolean) attribute).getValue());
				//				hash = 0;
				break;
			case Attribute.DISC :
				val = ((XAttributeDiscrete) attribute).getValue();
				writeLong(storage, index, val);
				//				hash = hash(val);
				break;
			case Attribute.TIME_ABS :
			case Attribute.TIME_REL :
				val = ((XAttributeTimestamp) attribute).getValueMillis();
				writeLong(storage, index, val);
				//				hash = hash(val);
				break;
			case Attribute.CONT :
				double v = ((XAttributeContinuous) attribute).getValue();
				writeDouble(storage, index, v);
				//				hash = hash(Double.doubleToLongBits(v));
				break;
			case Attribute.LITR :
				if (attribute instanceof XLightLiteralAttribute.WithIndex) {
					litIndex = ((XLightLiteralAttribute.WithIndex) attribute).getIndex();
				} else {
					String value = ((XAttributeLiteral) attribute).getValue();
					litIndex = stringHandler.addLiteral(value, lit2index);
				}
				writeInt(storage, index, litIndex);
				//				hash = hash(stringHandler.getLiteralChars(litIndex));
				break;
			default :
				throw new RuntimeException("Unexpected Attribute Type");
		}
		//		key2hash.put(keyIndex, newHash(key2hash.get(keyIndex), hash));

	}

	/**
	 * returns a sorted list of attribute keys. The list is sorted based on the
	 * index in the array used to store event data.
	 * 
	 * @return
	 */
	public Attribute[] getKeysSortedByIndex() {
		Attribute[] result = Arrays.copyOf(keys, keys.length);// key2index.keys(new
		// Attribute[key2index.size()]);
		Arrays.sort(result, new Comparator<Attribute>() {

			@Override
			public int compare(Attribute a1, Attribute a2) {
				return key2index.get(a1) - key2index.get(a2);
			}
		});
		return result;
	}

	/**
	 * returns a sorted list of attribute keys. The list is sorted based on the
	 * index in the array used to store event data.
	 * 
	 * @return
	 */
	public Attribute[] getKeysSortedAlphabetically() {
		return keys;
	}

	/**
	 * returns a list of global keys. The list is sorted.
	 * 
	 * @return
	 */
	public Attribute[] getKeys() {
		return keys;// key2index.keys(new Attribute[key2index.size()]);
	}

	/**
	 * returns a sorted list of attribute keys of type type. The list is sorted
	 * based on the index in the array used to store data.
	 * 
	 * @return
	 */
	public Attribute[] getKeySortedByIndex(byte type) {
		Attribute[] result = new Attribute[type2count[type]];

		int j = 0;
		Attribute[] keys = getKeysSortedByIndex();
		for (int i = 0; i < keys.length; i++) {
			if (keys[i].getType() == type) {
				result[j++] = keys[i];
			}
		}
		return result;
	}

	/**
	 * returns a sorted list of attribute keys of type type. The list is sorted
	 * based on the index in the array used to store data.
	 * 
	 * @return
	 */
	public Attribute[] getKeysSortedAlphabetically(byte type) {
		Attribute[] result = new Attribute[type2count[type]];

		int j = 0;
		Attribute[] keys = getKeysSortedAlphabetically();
		for (int i = 0; i < keys.length; i++) {
			if (keys[i].getType() == type) {
				result[j++] = keys[i];
			}
		}
		return result;
	}

	public int[] getKeyIndices() {
		int[] result = key2index.values();
		return result;
	}

	public int[] getKeyIndices(byte type) {
		int[] indices = new int[type2count[type]];
		int j = 0;
		Attribute[] keys = getKeys();
		for (int i = 0; i < keys.length; i++) {
			if (keys[i].getType() == type) {
				indices[j++] = key2index.get(keys[i]);
			}
		}
		return indices;
	}

	public XAttribute getAttribute(int number, Attribute attribute, Map<String, XExtension> key2extension) {
		int block = (number * bytesReservedPerObject) >>> shift;
		int ptr = (number * bytesReservedPerObject) & mask;

		byte[] array = store[block];

		return readAttribute(array, attribute.getKey(), attribute.getType(), ptr + key2index.get(attribute),
				key2extension, true);
	}

	protected void readAttributeIntoScope(Attribute[] globalKeys, int ptr, byte[] storage, int keyIndex,
			XAttributeMap attributes, Map<String, XExtension> key2extension, boolean forAddingAttributes) {
		Attribute key = globalKeys[keyIndex];

		String skey = key.getKey();
		byte type = key.getType();

		int val = key2index.get(key);
		int index = ptr + val;

		XAttribute attr = readAttribute(storage, skey, type, index, key2extension, forAddingAttributes);
		if (attr != null) {
			attributes.put(skey, attr);
		}
	}

	public int getHeader(int number, int headerIndex) {
		int block = (number * bytesReservedPerObject) >>> shift;
		int ptr = (number * bytesReservedPerObject) & mask;

		byte[] array = store[block];
		return readInt(array, ptr + 4 * headerIndex);
	}

	private XAttribute readAttribute(byte[] storage, String skey, byte type, int index,
			Map<String, XExtension> key2extension, boolean forAddingAttributes) {

		XAttribute attr;
		switch (type) {
			case Attribute.BOOL :
				attr = FACTORY.createAttributeBoolean(skey, readBoolean(storage, index), key2extension.get(skey));
				break;
			case Attribute.DISC :
				attr = FACTORY.createAttributeDiscrete(skey, readLong(storage, index), key2extension.get(skey));
				break;
			case Attribute.TIME_ABS :
			case Attribute.TIME_REL :
				if (forAddingAttributes) {
					attr = new XLightTimestampAttribute(skey, readLong(storage, index), key2extension.get(skey));
				} else {
					attr = FACTORY.createAttributeTimestamp(skey, readLong(storage, index), key2extension.get(skey));
				}
				break;
			case Attribute.CONT :
				attr = FACTORY.createAttributeContinuous(skey, readDouble(storage, index), key2extension.get(skey));
				break;
			case Attribute.LITR :
				if (forAddingAttributes) {
					attr = new XLightLiteralAttribute(skey, stringHandler.getLiteral(readInt(storage, index)),
							key2extension.get(skey));
				} else {
					attr = FACTORY.createAttributeLiteral(skey, stringHandler.getLiteral(readInt(storage, index)),
							key2extension.get(skey));
				}
				break;
			default :
				throw new RuntimeException("Unexpected Attribute Type");
		}
		return attr;
	}

	public boolean isType(Attribute key, byte type) {
		return key.getType() == type;
	}

	public Statistics.Long getStatisticsDiscrete(Attribute attribute) {
		Statistics.Long stat = new Statistics.Long();

		int d = 0;
		for (int b = 0; b < blocks; b++) {
			for (int e = key2index.get(attribute); d < numberOfObjects && e < blockSize; e += bytesReservedPerObject) {
				stat.addObservation(readLong(store[b], e));
				d++;
			}
		}
		return stat;
	}

	public Statistics.Double getStatisticsContinuous(Attribute attribute) {
		Statistics.Double stat = new Statistics.Double();

		int d = 0;
		for (int b = 0; b < blocks; b++) {
			for (int e = key2index.get(attribute); d < numberOfObjects && e < blockSize; e += bytesReservedPerObject) {
				stat.addObservation(readDouble(store[b], e));
				d++;
			}
		}
		return stat;
	}

	public Statistics.Long getStatisticsTimestamp(Attribute attribute) {
		return getStatisticsDiscrete(attribute);
	}

	public Histogram.Long getHistogramTimestamp(Attribute attribute, Statistics.Long s, int buckets) {
		return getHistogramDiscrete(attribute, s, buckets);
	}

	public Histogram.Long getHistogramDiscrete(Attribute attribute, Statistics.Long s, int buckets) {
		Histogram.Long hist = new Histogram.Long(s, buckets);

		int d = 0;
		for (int b = 0; b < blocks; b++) {
			for (int e = key2index.get(attribute); d < numberOfObjects && e < blockSize; e += bytesReservedPerObject) {
				hist.addObservation(readLong(store[b], e));
				d++;
			}
		}
		return hist;
	}

	public Histogram.Double getHistogramContinuous(Attribute attribute, Statistics.Double s, int buckets) {
		Histogram.Double hist = new Histogram.Double(s, buckets);

		int d = 0;
		for (int b = 0; b < blocks; b++) {
			for (int e = key2index.get(attribute); d < numberOfObjects && e < blockSize; e += bytesReservedPerObject) {
				hist.addObservation(readDouble(store[b], e));
				d++;
			}
		}
		return hist;
	}

	public int[] getAttributeValueIndicesLiteral(Attribute attribute, boolean sort) {
		// first get all the values
		TIntSet values = new TIntHashSet(64);
		//				getEstimatedAttributeValueCount(attribute));
		TIntList listValues = new TIntArrayList(64);
		//				getEstimatedAttributeValueCount(attribute));
		int keyIndex = key2index.get(attribute);
		for (int o = 0; o < numberOfObjects; o++) {
			int val = getAttributeValueLiteralIndex(o, keyIndex);
			if (values.add(val)) {
				listValues.add(val);
			}
		}
		int[] result = listValues.toArray();
		if (sort && result.length > 0) {
			// then sort on String value
			new AlphanumLiteralIndexSorter().sort(stringHandler, result);
		}
		return result;
	}

	public String[] getAttributeValuesLiteral(Attribute attribute, int[] indices) {
		String[] result = new String[indices.length];
		for (int i = 0; i < indices.length; i++) {
			result[i] = stringHandler.getLiteral(indices[i]);
		}
		return result;
	}

	public String[] getAttributeValuesLiteral(Attribute attribute, boolean sort) {
		return getAttributeValuesLiteral(attribute, getAttributeValueIndicesLiteral(attribute, sort));
	}

	public long[] getAttributeValuesDiscrete(Attribute attribute, boolean sort) {
		// first get all the values
		TLongSet values = new TLongHashSet(64);
		//				getEstimatedAttributeValueCount(attribute));
		TLongList listValues = new TLongArrayList(64);
		//				getEstimatedAttributeValueCount(attribute));
		int keyIndex = key2index.get(attribute);
		for (int o = 0; o < numberOfObjects; o++) {
			long val = getAttributeValueDiscrete(o, keyIndex);
			if (values.add(val)) {
				listValues.add(val);
			}
		}
		long[] result = listValues.toArray();
		if (sort) {
			Arrays.sort(result);
		}
		return result;
	}

	public long[] getAttributeValuesTimestamp(Attribute attribute, boolean sort) {
		// first get all the values
		TLongSet values = new TLongHashSet(64);
		//				getEstimatedAttributeValueCount(attribute));
		TLongList listValues = new TLongArrayList(64);
		//				getEstimatedAttributeValueCount(attribute));
		int keyIndex = key2index.get(attribute);
		for (int o = 0; o < numberOfObjects; o++) {
			long val = getAttributeValueTimestamp(o, keyIndex);
			if (values.add(val)) {
				listValues.add(val);
			}
		}
		long[] result = listValues.toArray();
		if (sort) {
			Arrays.sort(result);
		}
		return result;
	}

	public double[] getAttributeValuesContinuous(Attribute attribute, boolean sort) {
		// first get all the values
		TDoubleSet values = new TDoubleHashSet(64);
		//				getEstimatedAttributeValueCount(attribute));
		TDoubleList listValues = new TDoubleArrayList(64);
		//				getEstimatedAttributeValueCount(attribute));
		int keyIndex = key2index.get(attribute);
		for (int o = 0; o < numberOfObjects; o++) {
			double val = getAttributeValueContinuous(o, keyIndex);
			if (values.add(val)) {
				listValues.add(val);
			}
		}
		double[] result = listValues.toArray();
		if (sort) {
			Arrays.sort(result);
		}
		return result;
	}

	public boolean[] getAttributeValuesBoolean(Attribute attribute, boolean sort) {
		boolean t = false, f = false;
		int keyIndex = key2index.get(attribute);
		for (int o = 0; (!t || !f) && o < numberOfObjects; o++) {
			boolean b = getAttributeValueBoolean(o, keyIndex);
			t |= b;
			f |= !b;
		}
		if (!t) {
			return new boolean[] { false };
		}
		if (!f) {
			return new boolean[] { true };
		}
		return new boolean[] { false, true };
	}

	/**
	 * adds an attribute to each object in the store. The calculator computes
	 * the value of the attribute for each object.
	 * 
	 * @param calculator
	 *            Objects will be given to the calculator in the order in which
	 *            they appeared in the original log
	 * @param newAttribute
	 *            the new attributes to be added to each object (a default value
	 *            is not provided since the calculator returns a result for each
	 *            event)
	 */
	public synchronized void addAttributes(AttributeValueCalculator<T> calculator, Attribute... newAttribute) {

		// We cannot avoid duplication of Strings completely, as the original
		// mapping has been removed from
		// memory for efficiency reasons. However, for the new attribute, no two
		// strings should be duplicated.
		TObjectIntMap<String> lit2index = new TObjectIntHashMap<String>(100, 0.5f, -1);

		int newUsedBytesPerObject = usedBytesPerObject;
		for (int i = 0; i < newAttribute.length; i++) {
			//			keyIndex2hash.put(newUsedBytesPerObject, MAXHASH);
			newUsedBytesPerObject += Attribute.sizeForType(newAttribute[i].type);
		}

		int objectsPerBlock = blockSize / bytesReservedPerObject;
		int newBlockSize = blockSize;
		int newShift = shift;
		int newByteSize = bytesReservedPerObject;
		while (newUsedBytesPerObject > newByteSize) {
			newBlockSize <<= 1;
			newByteSize <<= 1;
			newShift++;
		}

		// iterate through the blocks
		for (int b = 0; b < blocks; b++) {
			byte[] oldBlock = store[b];
			byte[] newBlock = oldBlock;
			if (newByteSize > bytesReservedPerObject) {
				newBlock = new byte[newBlockSize];
			}

			for (int e = 0; e < objectsPerBlock && b * objectsPerBlock + e < numberOfObjects; e++) {
				T object = readObjectFromBlock(oldBlock, (e * bytesReservedPerObject) & mask, true);

				int[] header = new int[headerSize];
				for (int i = 0; i < header.length; i++) {
					header[i] = readInt(oldBlock, ((e * bytesReservedPerObject) & mask) + i * 4);
				}

				// compute the attribute for one object.
				XAttribute[] newAttributeValues = calculator.computeAttributes(object, b * objectsPerBlock + e, header);

				// Copy the original data from oldBlock to newBlock
				if (newBlock != oldBlock) {
					System.arraycopy(oldBlock, e * bytesReservedPerObject, newBlock, e * newByteSize,
							bytesReservedPerObject);
				}
				// append the new attribute
				int newIndex = usedBytesPerObject;
				for (int i = 0; i < newAttributeValues.length; i++) {
					writeAttribute(newAttributeValues[i], newBlock, e * newByteSize + newIndex, bytesReservedPerObject,
							newAttribute[i].type, lit2index);
					newIndex += Attribute.sizeForType(newAttribute[i].type);
				}

			}
			store[b] = newBlock;
		}
		int newIndex = usedBytesPerObject;
		for (int i = 0; i < newAttribute.length; i++) {
			key2index.put(newAttribute[i], newIndex);
			type2count[newAttribute[i].getType()]++;
			newIndex += Attribute.sizeForType(newAttribute[i].type);
		}
		this.keys = key2index.keys(new Attribute[key2index.size()]);
		Arrays.sort(keys);
		this.bytesReservedPerObject = newByteSize;
		this.blockSize = newBlockSize;
		this.usedBytesPerObject = newUsedBytesPerObject;
		this.shift = newShift;
		this.mask = this.blockSize - 1;
		assert 1 << shift == blockSize;
	}

	//	/**
	//	 * returns an estimate for the number of different values an attribute in
	//	 * this log has. This is an estimate based on two simple probabilistic lossy
	//	 * counting measures and the actual value may differ.
	//	 * 
	//	 * @param key
	//	 * @return
	//	 */
	//	public int getEstimatedAttributeValueCount(Attribute key) {
	//		if (key.type == Attribute.BOOL) {
	//			return 2;
	//		}
	//		int index = key2index.get(key);
	//		long hashes = keyIndex2hash.get(index);
	//
	//		double m1 = 1 << 21;
	//		long h1 = hashes & MASK1;
	//		double e1 = h1 == 0 ? m1 : m1 / h1 - 1.0;
	//
	//		long h2 = (hashes & MASK2) >>> 21;
	//		double e2 = h2 == 0 ? m1 : m1 / h2 - 1.0;
	//
	//		double m2 = 1 << 22;
	//		long h3 = (hashes & MASK3) >>> 42;
	//		double e3 = h3 == 0 ? m2 : m2 / h3 - 1.0;
	//
	//		return (int) (((e1 + e2 + e3) / 3.0));
	//	}

	public boolean containsKey(Attribute key) {
		return key2index.containsKey(key);
	}

	public int getObjectCount() {
		return numberOfObjects;
	}

	public int getKeyIndex(Attribute attribute) {
		return key2index.get(attribute);
	}

	public void swap(int index1, int index2) {
		if (index1 == index2) {
			return;
		}

		byte[] block1 = store[index1 * bytesReservedPerObject >>> shift];
		int start1 = index1 * bytesReservedPerObject & mask;
		byte[] block2 = store[index2 * bytesReservedPerObject >>> shift];
		int start2 = index2 * bytesReservedPerObject & mask;

		byte[] temp = new byte[bytesReservedPerObject];
		System.arraycopy(block1, start1, temp, 0, bytesReservedPerObject);
		System.arraycopy(block2, start2, block1, start1, bytesReservedPerObject);
		System.arraycopy(temp, 0, block2, start2, bytesReservedPerObject);
	}

	public StringHandler getStringHandler() {
		return stringHandler;
	}

	public String toString(int number) {
		int block = (number * bytesReservedPerObject) >>> shift;
		int base = (number * bytesReservedPerObject) & mask;
		byte[] array = store[block];

		StringBuilder b = new StringBuilder();
		b.append("[");
		for (int i = 0; i < keys.length; i++) {
			if (i > 0) {
				b.append(",");
			}
			b.append(keys[i].getKey());
			b.append("=");
			int ptr = base + key2index.get(keys[i]);
			switch (keys[i].getType()) {
				case Attribute.BOOL :
					b.append(readBoolean(array, ptr));
					break;
				case Attribute.DISC :
					b.append(readLong(array, ptr));
					break;
				case Attribute.TIME_ABS :
					b.append(DF_ABS.format(new Date(readLong(array, ptr))));
					break;
				case Attribute.TIME_REL :
					b.append(DF_REL.format(new Date(readLong(array, ptr))));
					break;
				case Attribute.CONT :
					b.append(readDouble(array, ptr));
					break;
				case Attribute.LITR :
					b.append(stringHandler.getLiteral(readInt(array, ptr)));
					break;
			}
		}
		b.append("]");
		return b.toString();
	}

	public Attribute getAttributeForKey(String attribute) {
		int idx = Arrays.binarySearch(keys, new Attribute(attribute, Attribute.BOOL, 0, handlesEventAttributes));
		if (idx >= 0) {
			return keys[idx];
		} else {
			return null;
		}

	}

	public int getObjectByteSize() {
		return bytesReservedPerObject;
	}

	public byte[] getBytes(Attribute attribute, int number) {
		return getBytes(key2index.get(attribute), Attribute.sizeForType(attribute.getType()), number);
	}

	public byte[] getBytes(int index, int length, int number) {
		int block = (number * bytesReservedPerObject) >>> shift;
		int ptr = ((number * bytesReservedPerObject) & mask) + index;
		return Arrays.copyOfRange(store[block], ptr, ptr + length);
	}

	public int countValuesLiteral(int attributeIndex, String s) {
		int r = 0;
		for (int block = 0; block < blocks; block++) {
			for (int i = attributeIndex; i < blockSize; i += bytesReservedPerObject) {
				if (s.equals(stringHandler.getLiteral(readInt(store[block], i)))) {
					r++;
				}
			}
		}
		return r;
	}

}
