package org.processmining.ocbc.performance.logprojection.abstraction;

public class HeurisicMinerCausalDepencency implements CausalDependencyMatrix {

	protected final double[][] dependencies;
	protected final DirectSuccessionCounter counter;

	public HeurisicMinerCausalDepencency(DirectSuccessionCounter counter) {
		this.counter = counter;
		this.dependencies = new double[counter.getNumEventClasses()][counter
				.getNumEventClasses()];
		fillCausalMatrix();
		// transitivelyClose();
		scale(getMinimum());
	}

	protected double getMinimum() {
		double min = 2.0;
		for (int i = 0; i < dependencies.length; i++) {
			for (int j = 0; j < dependencies.length; j++) {
				if (dependencies[i][j] > 0 && dependencies[i][j] < min) {
					min = dependencies[i][j];
				}
			}
		}
		if (min > 1) {
			return 0;
		}
		return min;
	}

	protected void scale(double minVal) {
		double length = (1 - minVal) * 2;
		for (int i = 0; i < dependencies.length; i++) {
			for (int j = 0; j < dependencies.length; j++) {
				if (dependencies[i][j] > 0) {
					dependencies[i][j] = 0.5 + (dependencies[i][j] - minVal)
							/ length;
				}
			}
		}
	}

	protected void transitivelyClose() {
		int i = 0, j = 0;

		while (i < dependencies.length) {
			int minI = i;
			j = 0;
			while (j < dependencies.length) {
				int minJ = j;
				// find the first element that is not 0;
				if (dependencies[i][j] > 0) {

					for (int k = 0; k < dependencies.length; k++) {
						// fill row i and column j with products
						double trans = dependencies[i][j] * dependencies[j][k];
						if (dependencies[i][k] < trans) {
							dependencies[i][k] = trans;
							if (i < minI) {
								minI = i;
							}
						}
						trans = dependencies[k][i] * dependencies[i][j];
						if (dependencies[k][j] < trans) {
							dependencies[k][j] = trans;
							if (j < minJ) {
								minJ = j;
							}
						}
					}

				}
				if (minJ < j) {
					// restart at earlier column, since we updated a value on
					// that column
					j = minJ;
				} else {
					// continue at next column.
					j++;
				}
			}
			if (minI < i) {
				// restart at earlier row, since we updated a value on that row
				i = minI;
			} else {
				// continue at next row.
				i++;
			}
		}
	}

	protected void fillCausalMatrix() {
		for (int i = 0; i < dependencies.length; i++) {
			for (int j = 0; j < dependencies.length; j++) {

				if (i == j) {
					double iToI = counter.getSucCount(i, j);
					double causal = (iToI) / (iToI + 1);
					dependencies[i][j] = causal;
				} else {
					double iToJ = counter.getSucCount(i, j);
					double jToI = counter.getSucCount(j, i);
					double causal = (iToJ - jToI) / (iToJ + jToI + 1);

					iToJ = counter.getABACount(i, j);
					jToI = counter.getABACount(j, i);
					double causal2 = (iToJ + jToI) / (iToJ + jToI + 1);

					dependencies[i][j] = Math.max(0, Math.max(causal, causal2));
					// dependencies[i][j] = Math.max(0, causal);
				}
			}
		}
	}

	@Override
	public double getCausalDependency(int node1, int node2) {
		return dependencies[node1][node2];
	}

	@Override
	public int getNumberOfNodes() {
		return dependencies.length;
	}

	@Override
	public int getOccurenceCount(int i) {
		return counter.getOccurenceCount(i);
	}

	@Override
	public String getEventClass(int i) {
		return counter.getEventClass(i);
	}

	public String toString() {
		return "Heuristic miner";
	}

	@Override
	public int getCount(int node1, int node2) {
		return counter.getCount(node1, node2);
	}

}
