package org.processmining.ocbc.performance.logprojection.abstraction;

import org.processmining.ocbc.performance.logprojection.lightxlog.LightLog;

import gnu.trove.map.TIntByteMap;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TIntByteHashMap;
import gnu.trove.procedure.TIntByteProcedure;
import gnu.trove.procedure.TIntProcedure;

public class MultiPhaseContext {

	private final int event;

	private final boolean useFrequencies;

	private final TObjectIntMap<String> name2eventIndex;

	private final TIntByteMap preSet = new TIntByteHashMap(3, 0.7f,
			Integer.MIN_VALUE, (byte) -1) {
		/** {@inheritDoc} */
		@Override
		public boolean equals(Object other) {
			if (!(other instanceof TIntByteMap)) {
				return false;
			}
			TIntByteMap that = (TIntByteMap) other;
			if (that.size() != this.size()) {
				return false;
			}
			byte[] values = _values;
			byte[] states = _states;
			byte this_no_entry_value = getNoEntryValue();
			byte that_no_entry_value = that.getNoEntryValue();
			for (int i = values.length; i-- > 0;) {
				if (states[i] == FULL) {
					int key = _set[i];
					byte that_value = that.get(key);
					byte this_value = values[i];
					if ((this_value != that_value)
							&& (this_value != this_no_entry_value || that_value != that_no_entry_value)) {
						return false;
					}
				}
			}
			return true;
		}
	};

	private final TIntByteMap postSet = new TIntByteHashMap(3, 0.7f, -1,
			(byte) -1) {
		/** {@inheritDoc} */
		@Override
		public boolean equals(Object other) {
			if (!(other instanceof TIntByteMap)) {
				return false;
			}
			TIntByteMap that = (TIntByteMap) other;
			if (that.size() != this.size()) {
				return false;
			}
			byte[] values = _values;
			byte[] states = _states;
			byte this_no_entry_value = getNoEntryValue();
			byte that_no_entry_value = that.getNoEntryValue();
			for (int i = values.length; i-- > 0;) {
				if (states[i] == FULL) {
					int key = _set[i];
					byte that_value = that.get(key);
					byte this_value = values[i];
					if ((this_value != that_value)
							&& (this_value != this_no_entry_value || that_value != that_no_entry_value)) {
						return false;
					}
				}
			}
			return true;
		}
	};

	public MultiPhaseContext(int event, boolean useFrequencies,
			TObjectIntMap<String> name2eventIndex) {
		this.event = event;
		this.useFrequencies = useFrequencies;
		this.name2eventIndex = name2eventIndex;
	}

	public void addToPre(String name) {
		int eventIndex = name2eventIndex.get(name);
		if (eventIndex == LightLog.NONAME) {
			// event not recognized
			return;
		}
		if (useFrequencies) {
			getPreSet().adjustOrPutValue(eventIndex, (byte) 1, (byte) 1);
		} else {
			getPreSet().adjustOrPutValue(eventIndex, (byte) 0, (byte) 1);
		}
	}

	public void addToPost(String name) {
		int eventIndex = name2eventIndex.get(name);
		if (eventIndex == LightLog.NONAME) {
			// event not recognized
			return;
		}
		if (useFrequencies) {
			getPostSet().adjustOrPutValue(eventIndex, (byte) 1, (byte) 1);
		} else {
			getPostSet().adjustOrPutValue(eventIndex, (byte) 0, (byte) 1);
		}
	}

	@Override
	public boolean equals(Object o) {
		if (o instanceof MultiPhaseContext) {
			MultiPhaseContext c = (MultiPhaseContext) o;
			boolean b = c.getEvent() == getEvent()
					&& c.getPreSet().equals(getPreSet())
					&& c.getPostSet().equals(getPostSet());
			assert !b || hashCode() == c.hashCode();

			return b;
		}
		return false;
	}

	public int hashCode() {
		return getEvent() * 31 * 31 + getPreSet().hashCode() * 31
				+ getPostSet().hashCode();
	}

	public String toString() {
		return getPreSet().toString() + " (" + getEvent() + ") "
				+ getPostSet().toString();
	}

	public double getSimilarity(MultiPhaseContext c) {
		//
		// if (isDisjointPreSet(c)) {
		// return 0;
		// } else {
		// return 1;
		// }
		//
		double pre = getWeightedJaccardSimilarity(preSet, c.preSet);
		double post = getWeightedJaccardSimilarity(postSet, c.postSet);

		return Math.sqrt(pre) * Math.sqrt(post);
		// return (pre + post) / 2;

		// * getJaccardSimilarity(postSet, c.postSet);
		// return (getJaccardSimilarity(preSet, c.preSet) +
		// getJaccardSimilarity(
		// postSet, c.postSet)) / 2.0;
	}

	public boolean isDisjoint(MultiPhaseContext c) {
		return isDisjointPreSet(c) && isDisjointPostSet(c);
	}

	public boolean isDisjointPreSet(final MultiPhaseContext c) {
		return getPreSet().forEachKey(new TIntProcedure() {

			@Override
			public boolean execute(int key) {
				return !c.getPreSet().containsKey(key);
			}

		});
	}

	public boolean isDisjointPostSet(final MultiPhaseContext c) {
		return getPostSet().forEachKey(new TIntProcedure() {

			@Override
			public boolean execute(int key) {
				return !c.getPostSet().containsKey(key);
			}

		});
	}

	public int getEvent() {
		return event;
	}

	public TIntByteMap getPreSet() {
		return preSet;
	}

	public TIntByteMap getPostSet() {
		return postSet;
	}

	protected double getJaccardSimilarity(TIntByteMap map1, TIntByteMap map2) {
		// compute size of intersection over size of union;

		MyIntersectionProcedure procedure = new MyIntersectionProcedure(map2);
		map1.forEachEntry(procedure);
		MySummationProcedure procedure2 = new MySummationProcedure();
		map2.forEachEntry(procedure2);

		int intersectionSize = procedure.intersectionSize;
		int unionSize = procedure.sumValues + procedure2.sumValues;

		if (unionSize == 0) {
			// both sets empty, i.e. identical
			return 1;
		}
		return intersectionSize / (double) unionSize;
	}

	private static class MyIntersectionProcedure implements TIntByteProcedure {

		public int intersectionSize = 0;
		public int sumValues = 0;
		private TIntByteMap other;

		public MyIntersectionProcedure(TIntByteMap other) {
			this.other = other;
		}

		@Override
		public boolean execute(int a, byte b) {
			int otherB = other.get(a);
			if (otherB > 0) {
				intersectionSize += Math.min(b, otherB);
			}
			sumValues += b;
			return true;
		}
	}

	private static class MySummationProcedure implements TIntByteProcedure {

		public int sumValues = 0;

		@Override
		public boolean execute(int a, byte b) {
			sumValues += b;
			return true;
		}
	}

	protected double getWeightedJaccardSimilarity(TIntByteMap map1,
			TIntByteMap map2) {
		// compute size of intersection over size of union;

		MyMinMaxProcedure procedure = new MyMinMaxProcedure(map2);
		map1.forEachEntry(procedure);
		MySummationProcedure procedure2 = new MySummationProcedure();
		map2.forEachEntry(procedure2);

		int sumMin = procedure.sumMin;
		int sumMax = procedure.sumMax + procedure2.sumValues;

		if (sumMax == 0) {
			// both sets empty, i.e. identical
			return 1;
		}
		return sumMin / (double) sumMax;
	}

	private static class MyMinMaxProcedure implements TIntByteProcedure {

		public int sumMax = 0;
		public int sumMin = 0;
		private TIntByteMap other;

		public MyMinMaxProcedure(TIntByteMap other) {
			this.other = other;
		}

		@Override
		public boolean execute(int a, byte b) {
			int otherB = other.get(a);
			if (otherB > 0) {
				sumMin += Math.min(b, otherB);
			}
			// later we add otherB. If b was the maximum, then
			// b this adds to b. If otherB was maximum this adds to otherB.
			if (b > otherB) {
				sumMax += b - otherB;
			}
			return true;
		}
	}

	public void addToPre(TIntByteMap toAdd) {
		toAdd.forEachEntry(new TIntByteProcedure() {

			@Override
			public boolean execute(int a, byte b) {
				if (useFrequencies) {
					preSet.adjustOrPutValue(a, b, b);
				} else {
					preSet.adjustOrPutValue(a, (byte) 0, b);

				}
				return true;
			}
		});

	}

	public void addToPost(TIntByteMap toAdd) {
		toAdd.forEachEntry(new TIntByteProcedure() {

			@Override
			public boolean execute(int a, byte b) {
				if (useFrequencies) {
					postSet.adjustOrPutValue(a, b, b);
				} else {
					postSet.adjustOrPutValue(a, (byte) 0, b);

				}
				return true;
			}
		});

	}

}