package org.processmining.ocbc.performance.logprojection.plugins.dottedchart;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;

import javax.swing.JPanel;

import org.deckfour.xes.classification.XEventAttributeClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.LegendItem;
import org.jfree.chart.LegendItemCollection;
import org.jfree.chart.LegendItemSource;
import org.jfree.chart.axis.AxisSpace;
import org.jfree.chart.axis.AxisState;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.Timeline;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.axis.ValueTick;
import org.jfree.chart.event.PlotChangeEvent;
import org.jfree.chart.plot.CrosshairState;
import org.jfree.chart.plot.FastScatterPlot;
import org.jfree.chart.plot.Pannable;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.PlotRenderingInfo;
import org.jfree.chart.plot.PlotState;
import org.jfree.chart.plot.ValueAxisPlot;
import org.jfree.chart.plot.Zoomable;
import org.jfree.chart.util.ParamChecks;
import org.jfree.chart.util.RelativeDateFormat;
import org.jfree.data.Range;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.RectangleInsets;
import org.jfree.util.ObjectUtilities;
import org.jfree.util.PaintUtilities;
import org.processmining.ocbc.performance.logprojection.Attribute;
import org.processmining.ocbc.performance.logprojection.DataManager;
import org.processmining.ocbc.performance.logprojection.LogView;
import org.processmining.ocbc.performance.logprojection.Statistics;
import org.processmining.ocbc.performance.logprojection.StatisticsWithSpan;
import org.processmining.ocbc.performance.logprojection.calculators.ClassifierAttribute;
import org.processmining.ocbc.performance.logprojection.filters.EventFilter;
import org.processmining.ocbc.performance.logprojection.lightxlog.LightEvent;
import org.processmining.ocbc.performance.logprojection.lightxlog.LightLog;
import org.processmining.ocbc.performance.logprojection.lightxlog.LightTrace;

import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.map.TDoubleObjectMap;
import gnu.trove.map.TIntObjectMap;
import gnu.trove.map.hash.TDoubleObjectHashMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.procedure.TDoubleObjectProcedure;
import javafx.util.Pair;

/**
 * A fast scatter plot. The code is taken from FastScatterPlot in JFreeChart,
 * but this class did not allow for sufficient access to private fields through
 * the appropriate getters and setters.
 * 
 * Access to various methods of this class are synchronized as they can be very
 * time consuming on larger datasets. The callers should ensure that these
 * methods are not called from the EventDispatchThread is not absolutely
 * necessary.
 */
public class DottedChart extends Plot implements ValueAxisPlot, Pannable, Zoomable, LegendItemSource {

	private static class Random {

		long last;

		private final long a = 1664525L;
		private final long c = 1013904223L;

		public Random() {
			reset();
		}

		public float next() {
			last = (last * a + c) & 0x00000000FFFFFFFFL;
			float f = (float) last / (float) 0x00000000FFFFFFFFL;
			return f;
		}

		public void reset() {
			last = 3252535;
		}

	}

	private static class CategoryInfo implements Comparable<CategoryInfo> {
		public int categoryNumber;

		public double minValue = Double.MAX_VALUE;

		public CategoryInfo(int categoryNumber) {
			this.categoryNumber = categoryNumber;
		}

		public String toString() {
			return categoryNumber + " (" + minValue + ")";
		}

		public boolean equals(Object o) {
			return (o instanceof CategoryInfo && ((CategoryInfo) o).categoryNumber == categoryNumber);
		}

		public int hashCode() {
			return categoryNumber;
		}

		@Override
		public int compareTo(CategoryInfo o) {
			if (minValue > o.minValue) {
				return 1;
			} else if (minValue < o.minValue) {
				return -1;
			} else {
				return categoryNumber - o.categoryNumber;

			}
		}
	}

	public static enum AttributeType {
		XAXIS, YAXIS, SHAPE, COLOR
	};

	public static final class DottedChartException extends Exception {
		private static final long serialVersionUID = -5371513609048188840L;

		public DottedChartException(String message) {
			super(message);
		}
	}

	private final Random random = new Random();

	private final ReentrantReadWriteLock rwLock = new ReentrantReadWriteLock();
	private final ReadLock read = rwLock.readLock();
	private final WriteLock write = rwLock.writeLock();

	protected static final DateFormat DATEFORMATABS = new SimpleDateFormat();
	protected static final DateFormat DATEFORMATREL = new RelativeDateFormat();

	/** For serialization. */
	private static final long serialVersionUID = 7871545897358563521L;

	/** The default grid line stroke. */
	public static final Stroke DEFAULT_GRIDLINE_STROKE = new BasicStroke(0.5f, BasicStroke.CAP_BUTT,
			BasicStroke.JOIN_BEVEL, 0.0f, new float[] { 2.0f, 2.0f }, 0.0f);

	/** The default outline stroke. */
	public static final Stroke DEFAULT_LINE_STROKE = new BasicStroke(1.0f, BasicStroke.CAP_ROUND,
			BasicStroke.JOIN_ROUND);

	/** The default grid line paint. */
	public static final Paint DEFAULT_GRIDLINE_PAINT = Color.lightGray;

	private float maxDotCount = 5000;

	private boolean first = true;

	/** The x data range. */
	private Range xDataRange;

	/** The y data range. */
	private Range yDataRange;

	/** The domain axis (used for the x-values). */
	private ValueAxis domainAxis;

	/** The range axis (used for the y-values). */
	private ValueAxis rangeAxis;

	/** A flag that controls whether the domain grid-lines are visible. */
	private boolean domainGridlinesVisible;

	/** The stroke used to draw the domain grid-lines. */
	private transient Stroke domainGridlineStroke;

	/** The paint used to draw the domain grid-lines. */
	private transient Paint domainGridlinePaint;

	/** A flag that controls whether the range grid-lines are visible. */
	private boolean rangeGridlinesVisible;

	/** The stroke used to draw the range grid-lines. */
	private transient Stroke rangeGridlineStroke;

	/** The paint used to draw the range grid-lines. */
	private transient Paint rangeGridlinePaint;

	private Attribute xAttribute;

	private Attribute yAttribute;

	// stores the required color for each event. 
	// The last 10 bits Should be interpreted as a unsigned , range 0 - 1023
	// The first first bit is a flag indicating whether or not to actually draw the event
	// Bits 2,3,4,5 are used to store the Shape (range 0-15)
	private char[] colorAndShape;

	private static final int VISIBLEMASK = 0x8000;
	private static final int SHAPEMASK = 0x7C00;
	private static final int COLORMASK = 0x03FF;

	//	// stores the shape and the filter for each event. Should be interpreted
	//	// as index into the shapeMap if positive. If negative, the corresponding
	//	// event should not be drawn. The shape it originally had is kept (i.e. the
	//	// value is -index-1;
	//	private byte[] shapes;

	private static final int MAXCOLORS = 1024;
	private Color[] colorMap;
	private String[] colorLabels;

	private static final int NUMSHAPES = 8;
	private byte[] shapeMap;
	private String[] shapeLabels;

	private TIntObjectMap<CategoryInfo> xCategory2LitIndex;

	private LogView log;

	private TIntObjectMap<CategoryInfo> yCategory2LitIndex;

	private double xDataLength;

	private double yDataLength;

	private Attribute shapeAttribute;

	private Attribute colorAttribute;

	private Statistics xStats;

	private Statistics yStats;

	private boolean lines = false;

	private boolean connectSameEventFlag = false;//indicate if connect the same event; true:yes
	
	private DateFormat dateFormat = new RelativeDateFormat();

	private double minimumSize = 6.0;

	private double dotSize = minimumSize;

	private boolean xIsDiscrete;

	private boolean yIsDiscrete;

	private int xAttributeIndex;

	private int yAttributeIndex;

	private final Set<EventFilter> eventFilterSet;

	private int[][] xBucketCount;

	private JPanel intensityPanel;

	private IntensityChart intensityChart;

	private int numVisisbleEvents;

	private double lastRenderTime;

	private double dotScale = 1.0;

	private static final int[] TIMES = new int[] { 1000, 60, 5, 3, 4 };

	private static final long MAXBUCKETS = 128;
	
	//record the x and y coordinates for the reference event in each instance (represent by index, 0..(size-1))
	private Map<Integer, Pair> instancePositionX = new HashMap<Integer, Pair>();
	
	//record the x and y coordinates for the reference event in each instance (represent by index, 0..(size-1))
	private Map<Integer, Pair> eventPositionX = new HashMap<Integer, Pair>();
	
	private Timeline asboluteTimeLine;
	
	/**
	 * @throws DottedChartException
	 * 
	 */
	public DottedChart(LogView log, Attribute xAttribute, Attribute yAttribute, Attribute colorAttribute,
			Attribute shapeAttribute) throws DottedChartException {
		this.log = log;
		try {
			setColorAttribute(colorAttribute, false);
		} catch (DottedChartException e) {
			setColorAttribute(null, false);
		}
		try {
			setShapeAttribute(shapeAttribute, false);
		} catch (DottedChartException e) {
			setShapeAttribute(null, false);
		}

		setXAxisAttribute(xAttribute, false);
		setYAxisAttribute(yAttribute, false);

		eventFilterSet = new HashSet<EventFilter>(5);

	}

	private void createXAxisBuckets(int size, int colors) {
		xBucketCount = new int[Math.max(size, 3)][colors];
	}

	public void setXAxisAttribute(Attribute attribute, boolean signalChange) throws DottedChartException {
		write.lock();
		try {
			DataManager<?> manager;
			String label;
			if (attribute.isEventAttribute()) {
				manager = log.getEventDataManager();
				label = "Event: " + attribute.getKey();
			} else {
				manager = log.getTraceDataManager();
				label = "Trace: " + attribute.getKey();
			}
			this.xCategory2LitIndex = null;
			this.xAttributeIndex = manager.getKeyIndex(attribute);
			if (attribute.getType() == Attribute.TIME_ABS) {
				String attributeName = attribute.getKey();
				if(attributeName.equals("time:timestamp")){
					this.domainAxis = new DateAxis(label);				
//					asboluteTimeLine = ((DateAxis) this.domainAxis).getTimeline();			
					xStats = log.getStatisticsTimestamp(attribute);
					xIsDiscrete = false;
				}
				else if(attributeName.equals("relativetime:timestamp")){
					this.domainAxis = new DateAxis(label);
					((DateAxis) this.domainAxis).setDateFormatOverride(dateFormat);
					xStats = log.getStatisticsTimestamp(attribute);
					xIsDiscrete = false;
				}			
			} else if (attribute.getType() == Attribute.TIME_REL) {
				this.domainAxis = new DateAxis(label);
				((DateAxis) this.domainAxis).setDateFormatOverride(dateFormat);
				xStats = log.getStatisticsTimestamp(attribute);
				xIsDiscrete = false;
			} else if (attribute.getType() == Attribute.CONT) {
				this.domainAxis = new NumberAxis(label);
				xStats = log.getStatisticsContinuous(attribute);
				xIsDiscrete = false;
				// choose a bucket size
				double s = 1.1 * (xStats.getMaximumDouble() - xStats.getMinimumDouble());
				while (s > MAXBUCKETS) {
					s /= 2;
				}
				createXAxisBuckets((int) s, colorMap.length);
			} else if (attribute.getType() == Attribute.DISC) {
				this.domainAxis = new NumberAxis(label);
				xStats = log.getStatisticsDiscrete(attribute);
				xIsDiscrete = true;
				// choose a bucket size
				long s = ((Statistics.Long) xStats).getMaximum() - ((Statistics.Long) xStats).getMinimum() + 1;
				while (s > MAXBUCKETS) {
					s >>= 1;
				}
				createXAxisBuckets((int) s, colorMap.length);
			} else if (attribute.getType() == Attribute.BOOL) {
				this.domainAxis = new NumberAxis(label);
				xStats = new StatisticsWithSpan.Long(0, 1, 0, 0, 0, 0, 1);
				xIsDiscrete = true;
				createXAxisBuckets(2, colorMap.length);
			} else if (attribute.getType() == Attribute.LITR) {
				// obtain a list of literals in use. The list is sorted based on
				// the
				// alphanumeric value of the underlying literal if the attribute
				// is
				// an event attribute. Otherwise, the sorting of traces in the
				// log
				// is respected
				int[] literals = manager.getAttributeValueIndicesLiteral(attribute, attribute.isEventAttribute());
				// reverse the map for lookup by renderer
				xCategory2LitIndex = new TIntObjectHashMap<CategoryInfo>((literals.length * 4) / 3, .75f, -1);
				for (int i = 0; i < literals.length; i++) {
					xCategory2LitIndex.put(literals[i], new CategoryInfo(i));
				}
				xStats = new StatisticsWithSpan.Long(0, literals.length - 1, 0, 0, 0, 0, literals.length);
				xIsDiscrete = true;
				this.domainAxis = new LiteralAxis(label, manager, literals);
				int size = literals.length;
				while (size > MAXBUCKETS) {
					size >>= 1;
				}

				createXAxisBuckets(size, colorMap.length);
			} else {
				throw new DottedChartException("Unknown attribute type");
			}

			if (xIsDiscrete) {
				this.xDataRange = new Range(xStats.getMinimumDouble() - 0.5, xStats.getMaximumDouble() + 0.5);
			} else {
				this.xDataRange = new Range(xStats.getMinimumDouble(), xStats.getMaximumDouble());
			}

			if ((attribute.getType() != Attribute.TIME_ABS) && (attribute.getType() != Attribute.TIME_REL)) {
				this.domainAxis.setLowerMargin(.05);
				this.domainAxis.setUpperMargin(.05);
				((NumberAxis) this.domainAxis).setAutoRangeStickyZero(false);
				((NumberAxis) this.domainAxis).setAutoRangeIncludesZero(false);
			} else {
				long span = (long) (1.1 * (((Statistics.Long) xStats).getMaximum() - ((Statistics.Long) xStats)
						.getMinimum()));
				if (span <= 0) {
					span = 1;
				}
				int i = 0;
				while (span > MAXBUCKETS && i < TIMES.length) {
					span /= TIMES[i++];
				}
				while (span > MAXBUCKETS) {
					span >>= 1;
				}
				createXAxisBuckets((int) span, colorMap.length);
			}

			this.xDataLength = xDataRange.getLength();
			this.xAttribute = attribute;
			this.domainAxis.setPlot(this);
			this.domainAxis.addChangeListener(this);
			this.domainGridlinesVisible = true;
			this.domainGridlinePaint = FastScatterPlot.DEFAULT_GRIDLINE_PAINT;
			this.domainGridlineStroke = FastScatterPlot.DEFAULT_GRIDLINE_STROKE;
			this.domainAxis.setAutoRange(true);
			this.setDomainPannable(true);

			if (signalChange) {
				fireChangeEvent();
			}
		} finally {
			write.unlock();
		}
	}

	public void setYAxisAttribute(Attribute attribute, boolean signalChange) throws DottedChartException {
		write.lock();

		try {
			DataManager<?> manager;
			String label;
			if (attribute.isEventAttribute()) {
				manager = log.getEventDataManager();
				label = "Event: " + attribute.getKey();
			} else {
				manager = log.getTraceDataManager();
				label = "Trace: " + attribute.getKey();
			}
			this.yCategory2LitIndex = null;
			this.rangeAxis = new NumberAxis(label);
			if (attribute.getType() == Attribute.TIME_ABS) {
				this.rangeAxis = new DateAxis(label);
				yIsDiscrete = false;
				yStats = log.getStatisticsTimestamp(attribute);
			} else if (attribute.getType() == Attribute.TIME_REL) {
				this.rangeAxis = new DateAxis(label);
				((DateAxis) this.rangeAxis).setDateFormatOverride(dateFormat);
				yIsDiscrete = false;
				yStats = log.getStatisticsTimestamp(attribute);
			} else if (attribute.getType() == Attribute.CONT) {
				yIsDiscrete = false;
				yStats = log.getStatisticsContinuous(attribute);
			} else if (attribute.getType() == Attribute.DISC) {
				yStats = log.getStatisticsDiscrete(attribute);
				yIsDiscrete = true;
			} else if (attribute.getType() == Attribute.BOOL) {
				yStats = new StatisticsWithSpan.Long(0, 1, 0, 0, 0, 0, 1);
				yIsDiscrete = true;
			} else if (attribute.getType() == Attribute.LITR) {
				// obtain a list of literals in use. The list is sorted based on
				// the
				// alphanumeric value of the underlying literal if the attribute
				// is
				// an event attribute. Otherwise, the sorting of traces in the
				// log
				// is respected
				int[] literals = manager.getAttributeValueIndicesLiteral(attribute, attribute.isEventAttribute());
				// reverse the map for lookup by renderer
				yCategory2LitIndex = new TIntObjectHashMap<CategoryInfo>((literals.length * 4) / 3, .75f, -1);
				for (int i = 0; i < literals.length; i++) {
					yCategory2LitIndex.put(literals[i], new CategoryInfo(i));
				}
				this.rangeAxis = new LiteralAxis(label, manager, literals);
				yIsDiscrete = true;
				yStats = new StatisticsWithSpan.Long(0, literals.length - 1, 0, 0, 0, 0, literals.length);
			} else {
				throw new DottedChartException("Unknown attribute type");
			}

			if (yIsDiscrete) {
				this.yDataRange = new Range(yStats.getMinimumDouble() - 0.5, yStats.getMaximumDouble() + 0.5);
			} else {
				this.yDataRange = new Range(yStats.getMinimumDouble(), yStats.getMaximumDouble());
			}

			this.rangeAxis.setInverted(true);
			if ((attribute.getType() != Attribute.TIME_ABS) && (attribute.getType() != Attribute.TIME_REL)) {
				this.rangeAxis.setLowerMargin(.05);
				this.rangeAxis.setUpperMargin(.05);
				((NumberAxis) this.rangeAxis).setAutoRangeStickyZero(false);
				((NumberAxis) this.rangeAxis).setAutoRangeIncludesZero(false);
			}

			this.rangeAxis.setInverted(true);
			this.yDataLength = yDataRange.getLength();
			this.yAttribute = attribute;
			this.yAttributeIndex = manager.getKeyIndex(yAttribute);
			this.rangeAxis.setPlot(this);
			this.rangeAxis.addChangeListener(this);
			this.rangeGridlinesVisible = true;
			this.rangeGridlinePaint = FastScatterPlot.DEFAULT_GRIDLINE_PAINT;
			this.rangeGridlineStroke = FastScatterPlot.DEFAULT_GRIDLINE_STROKE;
			this.rangeAxis.setAutoRange(true);
			this.setRangePannable(true);

			if (signalChange) {
				fireChangeEvent();
			}
		} finally {
			write.unlock();
		}

	}

	private int map(int event, boolean forEventAttribute) {
		if (forEventAttribute) {
			return event;
		} else {
			return log.getEventDataManager().getTraceForEvent(event);
		}
	}

	public void addEventFilter(EventFilter eventFilter, boolean signalChange) {
		synchronized (eventFilterSet) {
			if (!eventFilterSet.add(eventFilter)) {
				eventFilterSet.remove(eventFilter);
				eventFilterSet.add(eventFilter);
			}
		}
		applyEventFilters(signalChange);
	}

	public void removeEventFilter(EventFilter eventFilter, boolean signalChange) {
		synchronized (eventFilterSet) {
			eventFilterSet.remove(eventFilter);
		}
		applyEventFilters(signalChange);
	}

	public void applyEventFilters(boolean signalChange) {
		read.lock();
		try {
			for (int t = 0; t < log.numTraces(); t++) {
				int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
				for (int e = log.getFirstEventNumber(t); e < max; e++) {
					// clear the filter flag for event e
					//					shapes[e] &= 0x7F;
					colorAndShape[e] |= VISIBLEMASK;
					for (EventFilter f : eventFilterSet) {
						if (!f.keepEvent(log, t, e)) {
							//							shapes[e] |= 0x80;
							colorAndShape[e] &= ~VISIBLEMASK;
							break;
						}
					}
				}
			}
			if (signalChange) {
				fireChangeEvent();
			}
		} finally {
			read.unlock();
		}
	}

	public void setColorAttribute(Attribute colorAttribute, boolean signalChange) throws DottedChartException {

		write.lock();

		try {
			if (colorAndShape == null) {
				colorAndShape = new char[log.numEvents()];
				//				if (colors == null) {
				//					colors = new short[log.numEvents()];
				Arrays.fill(colorAndShape, (char) VISIBLEMASK);
			}
			if (colorAttribute == null) {
				colorMap = getUniqueColors(1);
				colorLabels = new String[] { "all" };
				this.colorAttribute = colorAttribute;
				if (xBucketCount != null) {
					createXAxisBuckets(xBucketCount.length, 1);
				}
				if (signalChange) {
					fireChangeEvent();
				}
				return;
			}

			DataManager<?> manager;
			if (colorAttribute.isEventAttribute()) {
				manager = log.getEventDataManager();
			} else {
				manager = log.getTraceDataManager();
			}
			System.out.println("colorAttribute.getType()" + colorAttribute.getType());
			// let's get the the colors and shapes right
			switch (colorAttribute.getType()) {
				case Attribute.BOOL :
					colorMap = getUniqueColors(2);
					colorLabels = new String[] { "true", "false"};
					for (int e = 0; e < log.numEvents(); e++) {
						boolean val = manager.getAttributeValueBoolean(map(e, colorAttribute.isEventAttribute()),
								colorAttribute);
						//						colors[e] = val ? (byte) 1 : 0;
						colorAndShape[e] &= ~COLORMASK;
//						colorAndShape[e] |= val ? 1 : 0;
						colorAndShape[e] |= val ? 0 : 1;

					}
					break;
				case Attribute.CONT : {
					double[] map = manager.getAttributeValuesContinuous(colorAttribute, true);
					if (map.length > MAXCOLORS) {
						throw new DottedChartException("<html>At most " + MAXCOLORS
								+ " colors are supported.<br>The chosen attribute has " + map.length + "</html>");
					}
					// map provides an indexing structure from the possible values
					// to
					// integers.
					colorMap = getUniqueColors(map.length);
					colorLabels = new String[map.length];
					for (int i = 0; i < map.length; i++) {
						colorLabels[i] = Double.toString(map[i]);
					}

					for (int e = 0; e < log.numEvents(); e++) {
						double val = manager.getAttributeValueContinuous(map(e, colorAttribute.isEventAttribute()),
								colorAttribute);
						//						colors[e] = (short) Arrays.binarySearch(map, val);
						colorAndShape[e] &= ~COLORMASK;
						colorAndShape[e] |= Arrays.binarySearch(map, val);

					}
					break;
				}
				case Attribute.LITR : {
					int[] map = manager.getAttributeValueIndicesLiteral(colorAttribute, true);
					if (map.length > MAXCOLORS) {
						throw new DottedChartException("<html>At most " + MAXCOLORS
								+ " colors are supported.<br>The chosen attribute has " + map.length + "</html>");
					}
					// map provides an indexing structure from the possible values
					// to
					// integers.
					colorMap = getUniqueColors(map.length);
					colorLabels = new String[map.length];
					for (int i = 0; i < map.length; i++) {
						colorLabels[i] = manager.getStringHandler().getLiteral(map[i]);
					}
					for (int e = 0; e < log.numEvents(); e++) {
						int val = manager.getAttributeValueLiteralIndex(map(e, colorAttribute.isEventAttribute()),
								colorAttribute);
						//						colors[e] = (short) linearSearch(map, val);
						colorAndShape[e] &= ~COLORMASK;
						colorAndShape[e] |= linearSearch(map, val);
						
						System.out.println("Attribute name in LITR, val: " + e + ", " + val);
						String referenceTarget = log.getEventDataManager().readObject(e).getAttributes().get("pattern:name").toString();
						String attributeName = colorAttribute.getKey();
						System.out.println("referenceTarget name in LITR: " + e + ", " + referenceTarget);
						if(attributeName.equals("pattern:name")){
							if(referenceTarget.equals("reference")){
								colorAndShape[e] &= ~COLORMASK;
								//set the color of reference events as 0, i.e., red
								colorAndShape[e] |= 0;
							}	
							else{
								colorAndShape[e] &= ~COLORMASK;
								//set the color of reference events as 1, i.e., green
								colorAndShape[e] |= 1;
							}
						}
						
					}
					break;
				}
				case Attribute.DISC : {
					long[] map = manager.getAttributeValuesDiscrete(colorAttribute, true);
					if (map.length > MAXCOLORS) {
						throw new DottedChartException("<html>At most " + MAXCOLORS
								+ " colors are supported.<br>The chosen attribute has " + map.length + "</html>");
					}
					// map provides an indexing structure from the possible values
					// to
					// integers.
					colorMap = getUniqueColors(map.length);
					colorLabels = new String[map.length];
					for (int i = 0; i < map.length; i++) {
						colorLabels[i] = Long.toString(map[i]);
					}
					for (int e = 0; e < log.numEvents(); e++) {
						long val = manager.getAttributeValueDiscrete(map(e, colorAttribute.isEventAttribute()),
								colorAttribute);
						//						colors[e] = (short) Arrays.binarySearch(map, val);
						colorAndShape[e] &= ~COLORMASK;
						colorAndShape[e] |= Arrays.binarySearch(map, val);

					}
					break;
				}
				case Attribute.TIME_ABS :
				case Attribute.TIME_REL : {
					long[] map = manager.getAttributeValuesTimestamp(colorAttribute, true);
					if (map.length > MAXCOLORS) {
						throw new DottedChartException("<html>At most " + MAXCOLORS
								+ " colors are supported.<br>The chosen attribute has " + map.length + "</html>");
					}
					// map provides an indexing structure from the possible values
					// to
					// integers.
					colorMap = getUniqueColors(map.length);
					colorLabels = new String[map.length];
					if (colorAttribute.getType() == Attribute.TIME_ABS) {
						for (int i = 0; i < map.length; i++) {
							colorLabels[i] = DATEFORMATABS.format(map[i]);
						}
					} else {
						for (int i = 0; i < map.length; i++) {
							colorLabels[i] = DATEFORMATREL.format(map[i]);
						}
					}
					for (int e = 0; e < log.numEvents(); e++) {
						long val = manager.getAttributeValueTimestamp(map(e, colorAttribute.isEventAttribute()),
								colorAttribute);
						//						colors[e] = (short) (Arrays.binarySearch(map, val) & 0xFFFF);
						colorAndShape[e] &= ~COLORMASK;
						colorAndShape[e] |= Arrays.binarySearch(map, val);
					}
					break;
				}
			}
			this.colorAttribute = colorAttribute;
			createXAxisBuckets(xBucketCount.length, colorMap.length);
			if (signalChange) {
				fireChangeEvent();
			}
		} finally {
			write.unlock();
		}
	}

	private Color[] getUniqueColors(int amount) {
		//		Color[] colors = new Color[amount];
		//
		//		if (amount == 1) {
		//			colors[0] = Color.RED;
		//			return colors;
		//		}

		// use whole circle: 2*PI
		//		double frequency = 6.28 / amount;
		//		for (int i = 0; i < amount; i++) {
		//			float g = (float) (Math.sin(frequency * i) * 0.5f + 0.5f);
		//			float r = (float) (Math.sin(frequency * i + 2.1) * 0.5f + 0.5f);
		//			float b = (float) (Math.sin(frequency * i + 4.2) * 0.5f + 0.5f);
		//			colors[i] = new Color(r, g, b, 0.4f);
		//		}

		//		float dx = 1.0f / (amount - 1);
		//		for (int i = 0; i < amount; i++) {
		//			colors[i] = (getColor(i * dx));
		//		}

		return ColorUtils.generateVisuallyDistinctColors(amount, .8f, .3f);

	}

	public static Color getColor(float x) {
		float r = 0.0f;
		float g = 0.0f;
		float b = 1.0f;
		if (x >= 0.0f && x < 0.2f) {
			x = x / 0.2f;
			r = 0.0f;
			g = x;
			b = 1.0f;
		} else if (x >= 0.2f && x < 0.4f) {
			x = (x - 0.2f) / 0.2f;
			r = 0.0f;
			g = 1.0f;
			b = 1.0f - x;
		} else if (x >= 0.4f && x < 0.6f) {
			x = (x - 0.4f) / 0.2f;
			r = x;
			g = 1.0f;
			b = 0.0f;
		} else if (x >= 0.6f && x < 0.8f) {
			x = (x - 0.6f) / 0.2f;
			r = 1.0f;
			g = 1.0f - x;
			b = 0.0f;
		} else if (x >= 0.8f && x <= 1.0f) {
			x = (x - 0.8f) / 0.2f;
			r = 1.0f;
			g = 0.0f;
			b = x;
		}
		return new Color(r, g, b);
	}

	public void setShapeAttribute(Attribute shapeAttribute, boolean signalChange) throws DottedChartException {

		write.lock();
		try {
			//			if (shapes == null) {
			//				shapes = new byte[log.numEvents()];
			//			}
			if (colorAndShape == null) {
				colorAndShape = new char[log.numEvents()];
				Arrays.fill(colorAndShape, (char) VISIBLEMASK);
			}
			if (shapeAttribute == null) {
				shapeMap = new byte[1];
				shapeLabels = new String[] { "all" };
				this.shapeAttribute = shapeAttribute;
				if (signalChange) {
					fireChangeEvent();
				}
				return;
			}

			DataManager<?> manager;
			if (shapeAttribute.isEventAttribute()) {
				manager = log.getEventDataManager();
			} else {
				manager = log.getTraceDataManager();
			}

			shapeMap = new byte[NUMSHAPES];

			for (int i = 0; i < NUMSHAPES; i++) {
				shapeMap[i] = (byte) (i);
			}

			// let's get the the colors and shapes right
			switch (shapeAttribute.getType()) {
				case Attribute.BOOL :
					shapeLabels = new String[] { "false", "true" };
					for (int e = 0; e < log.numEvents(); e++) {
						boolean val = manager.getAttributeValueBoolean(map(e, shapeAttribute.isEventAttribute()),
								shapeAttribute);
						//						shapes[e] &= 0x80;
						//						shapes[e] |= (byte) (val ? 5 : 0);
						// clear shape and visibility
						colorAndShape[e] &= ~SHAPEMASK;
						colorAndShape[e] |= (val ? 5 : 0) << 10;

					}
					break;
				case Attribute.CONT : {
					double[] map = manager.getAttributeValuesContinuous(shapeAttribute, true);
					if (map.length > shapeMap.length) {
						throw new DottedChartException("<html>At most " + shapeMap.length
								+ " shapes are supported.<br>The chosen attribute has " + map.length + "</html>");
					}
					// map provides an indexing structure from the possible values
					// to
					// integers.
					shapeLabels = new String[map.length];
					for (int i = 0; i < map.length; i++) {
						shapeLabels[i] = Double.toString(map[i]);
					}
					for (int e = 0; e < log.numEvents(); e++) {
						double val = manager.getAttributeValueContinuous(map(e, shapeAttribute.isEventAttribute()),
								shapeAttribute);
						//						shapes[e] &= 0x80;
						//						shapes[e] |= (byte) Arrays.binarySearch(map, val);
						// clear shape and visibility
						colorAndShape[e] &= ~SHAPEMASK;
						colorAndShape[e] |= Arrays.binarySearch(map, val) << 10;
					}
					break;
				}
				case Attribute.LITR : {
					int[] map = manager.getAttributeValueIndicesLiteral(shapeAttribute, true);
					if (map.length > shapeMap.length) {
						throw new DottedChartException("<html>At most " + shapeMap.length
								+ " shapes are supported.<br>The chosen attribute has " + map.length + "</html>");
					}
					// map provides an indexing structure from the possible values
					// to
					// integers.
					shapeLabels = new String[map.length];
					for (int i = 0; i < map.length; i++) {
						shapeLabels[i] = manager.getStringHandler().getLiteral(map[i]);
					}
					for (int e = 0; e < log.numEvents(); e++) {
						int val = manager.getAttributeValueLiteralIndex(map(e, shapeAttribute.isEventAttribute()),
								shapeAttribute);
						colorAndShape[e] &= ~SHAPEMASK;
						colorAndShape[e] |= linearSearch(map, val) << 10;
						
						System.out.println("Attribute name in LITR, val: " + e + ", " + val);
						String referenceTarget = log.getEventDataManager().readObject(e).getAttributes().get("pattern:name").toString();
						String attributeName = shapeAttribute.getKey();
						System.out.println("referenceTarget name in LITR: " + e + ", " + referenceTarget);
						if(attributeName.equals("pattern:name")){
							if(referenceTarget.equals("reference")){
								colorAndShape[e] &= ~SHAPEMASK;
								//set the shape of reference events as 1, i.e., square
								colorAndShape[e] |= 1 << 10; 
							}	
							else{
								colorAndShape[e] &= ~SHAPEMASK;
								//set the shape of target events as 0, i.e., circle
								colorAndShape[e] |= 0 << 10;
							}
						}
						//						shapes[e] &= 0x80;
						//						shapes[e] |= (byte) linearSearch(map, val);						
					}
					break;
				}
				case Attribute.DISC : {
					long[] map = manager.getAttributeValuesDiscrete(shapeAttribute, true);
					if (map.length > shapeMap.length) {
						throw new DottedChartException("<html>At most " + shapeMap.length
								+ " shapes are supported.<br>The chosen attribute has " + map.length + "</html>");
					}
					// map provides an indexing structure from the possible values
					// to
					// integers.
					shapeLabels = new String[map.length];
					for (int i = 0; i < map.length; i++) {
						shapeLabels[i] = Long.toString(map[i]);
					}
					for (int e = 0; e < log.numEvents(); e++) {
						long val = manager.getAttributeValueDiscrete(map(e, shapeAttribute.isEventAttribute()),
								shapeAttribute);
						//						shapes[e] &= 0x80;
						//						shapes[e] |= (byte) Arrays.binarySearch(map, val);
						colorAndShape[e] &= ~SHAPEMASK;
						colorAndShape[e] |= Arrays.binarySearch(map, val) << 10;
					}
					break;
				}
				case Attribute.TIME_ABS :
				case Attribute.TIME_REL : {
					long[] map = manager.getAttributeValuesTimestamp(shapeAttribute, true);
					if (map.length > shapeMap.length) {
						throw new DottedChartException("<html>At most " + shapeMap.length
								+ " shapes are supported.<br>The chosen attribute has " + map.length + "</html>");
					}
					// map provides an indexing structure from the possible values
					// to
					// integers.
					shapeLabels = new String[map.length];
					if (shapeAttribute.getType() == Attribute.TIME_ABS) {
						for (int i = 0; i < map.length; i++) {
							shapeLabels[i] = DATEFORMATABS.format(map[i]);
						}
					} else {
						for (int i = 0; i < map.length; i++) {
							shapeLabels[i] = DATEFORMATREL.format(map[i]);
						}
					}
					for (int e = 0; e < log.numEvents(); e++) {
						long val = manager.getAttributeValueTimestamp(map(e, shapeAttribute.isEventAttribute()),
								shapeAttribute);
						//						shapes[e] &= 0x80;
						//						shapes[e] |= (byte) Arrays.binarySearch(map, val);
						colorAndShape[e] &= ~SHAPEMASK;
						colorAndShape[e] |= Arrays.binarySearch(map, val) << 10;
					}
					break;
				}
			}
			this.shapeAttribute = shapeAttribute;
			// this.shapeAttributeIndex = manager.getKeyIndex(shapeAttribute);
			if (signalChange) {
				fireChangeEvent();
			}
		} finally {
			write.unlock();
		}
	}

	private static int linearSearch(int[] map, int val) {
		for (int i = 0; i < map.length; i++) {
			if (map[i] == val) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Returns a short string describing the plot type.
	 * 
	 * @return A short string describing the plot type.
	 */
	@Override
	public String getPlotType() {
		return "Dotted Chart";
	}

	/**
	 * Returns the orientation of the plot.
	 * 
	 * @return The orientation (always {@link PlotOrientation#VERTICAL}).
	 */
	@Override
	public PlotOrientation getOrientation() {
		return PlotOrientation.VERTICAL;
	}

	/**
	 * Returns the domain axis for the plot.
	 * 
	 * @return The domain axis (never <code>null</code>).
	 * 
	 * @see #setDomainAxis(ValueAxis)
	 */
	public ValueAxis getDomainAxis() {
		return this.domainAxis;
	}

	/**
	 * Sets the domain axis and sends a {@link PlotChangeEvent} to all
	 * registered listeners.
	 * 
	 * @param axis
	 *            the axis (<code>null</code> not permitted).
	 * 
	 * @since 1.0.3
	 * 
	 * @see #getDomainAxis()
	 */
	public void setDomainAxis(ValueAxis axis) {
		ParamChecks.nullNotPermitted(axis, "axis");
		this.domainAxis = axis;
		fireChangeEvent();
	}

	/**
	 * Returns the range axis for the plot.
	 * 
	 * @return The range axis (never <code>null</code>).
	 * 
	 * @see #setRangeAxis(ValueAxis)
	 */
	public ValueAxis getRangeAxis() {
		return this.rangeAxis;
	}

	/**
	 * Sets the range axis and sends a {@link PlotChangeEvent} to all registered
	 * listeners.
	 * 
	 * @param axis
	 *            the axis (<code>null</code> not permitted).
	 * 
	 * @since 1.0.3
	 * 
	 * @see #getRangeAxis()
	 */
	public void setRangeAxis(ValueAxis axis) {
		ParamChecks.nullNotPermitted(axis, "axis");
		this.rangeAxis = axis;
		fireChangeEvent();
	}

	/**
	 * Returns <code>true</code> if the domain gridlines are visible, and
	 * <code>false</code> otherwise.
	 * 
	 * @return <code>true</code> or <code>false</code>.
	 * 
	 * @see #setDomainGridlinesVisible(boolean)
	 * @see #setDomainGridlinePaint(Paint)
	 */
	public boolean isDomainGridlinesVisible() {
		return this.domainGridlinesVisible;
	}

	/**
	 * Sets the flag that controls whether or not the domain grid-lines are
	 * visible. If the flag value is changed, a {@link PlotChangeEvent} is sent
	 * to all registered listeners.
	 * 
	 * @param visible
	 *            the new value of the flag.
	 * 
	 * @see #getDomainGridlinePaint()
	 */
	public void setDomainGridlinesVisible(boolean visible) {
		if (this.domainGridlinesVisible != visible) {
			this.domainGridlinesVisible = visible;
			fireChangeEvent();
		}
	}

	/**
	 * Returns the stroke for the grid-lines (if any) plotted against the domain
	 * axis.
	 * 
	 * @return The stroke (never <code>null</code>).
	 * 
	 * @see #setDomainGridlineStroke(Stroke)
	 */
	public Stroke getDomainGridlineStroke() {
		return this.domainGridlineStroke;
	}

	/**
	 * Sets the stroke for the grid lines plotted against the domain axis and
	 * sends a {@link PlotChangeEvent} to all registered listeners.
	 * 
	 * @param stroke
	 *            the stroke (<code>null</code> not permitted).
	 * 
	 * @see #getDomainGridlineStroke()
	 */
	public void setDomainGridlineStroke(Stroke stroke) {
		ParamChecks.nullNotPermitted(stroke, "stroke");
		this.domainGridlineStroke = stroke;
		fireChangeEvent();
	}

	/**
	 * Returns the paint for the grid lines (if any) plotted against the domain
	 * axis.
	 * 
	 * @return The paint (never <code>null</code>).
	 * 
	 * @see #setDomainGridlinePaint(Paint)
	 */
	public Paint getDomainGridlinePaint() {
		return this.domainGridlinePaint;
	}

	/**
	 * Sets the paint for the grid lines plotted against the domain axis and
	 * sends a {@link PlotChangeEvent} to all registered listeners.
	 * 
	 * @param paint
	 *            the paint (<code>null</code> not permitted).
	 * 
	 * @see #getDomainGridlinePaint()
	 */
	public void setDomainGridlinePaint(Paint paint) {
		ParamChecks.nullNotPermitted(paint, "paint");
		this.domainGridlinePaint = paint;
		fireChangeEvent();
	}

	/**
	 * Returns <code>true</code> if the range axis grid is visible, and
	 * <code>false</code> otherwise.
	 * 
	 * @return <code>true</code> or <code>false</code>.
	 * 
	 * @see #setRangeGridlinesVisible(boolean)
	 */
	public boolean isRangeGridlinesVisible() {
		return this.rangeGridlinesVisible;
	}

	/**
	 * Sets the flag that controls whether or not the range axis grid lines are
	 * visible. If the flag value is changed, a {@link PlotChangeEvent} is sent
	 * to all registered listeners.
	 * 
	 * @param visible
	 *            the new value of the flag.
	 * 
	 * @see #isRangeGridlinesVisible()
	 */
	public void setRangeGridlinesVisible(boolean visible) {
		if (this.rangeGridlinesVisible != visible) {
			this.rangeGridlinesVisible = visible;
			fireChangeEvent();
		}
	}

	/**
	 * Returns the stroke for the grid lines (if any) plotted against the range
	 * axis.
	 * 
	 * @return The stroke (never <code>null</code>).
	 * 
	 * @see #setRangeGridlineStroke(Stroke)
	 */
	public Stroke getRangeGridlineStroke() {
		return this.rangeGridlineStroke;
	}

	/**
	 * Sets the stroke for the grid lines plotted against the range axis and
	 * sends a {@link PlotChangeEvent} to all registered listeners.
	 * 
	 * @param stroke
	 *            the stroke (<code>null</code> permitted).
	 * 
	 * @see #getRangeGridlineStroke()
	 */
	public void setRangeGridlineStroke(Stroke stroke) {
		ParamChecks.nullNotPermitted(stroke, "stroke");
		this.rangeGridlineStroke = stroke;
		fireChangeEvent();
	}

	/**
	 * Returns the paint for the grid lines (if any) plotted against the range
	 * axis.
	 * 
	 * @return The paint (never <code>null</code>).
	 * 
	 * @see #setRangeGridlinePaint(Paint)
	 */
	public Paint getRangeGridlinePaint() {
		return this.rangeGridlinePaint;
	}

	/**
	 * Sets the paint for the grid lines plotted against the range axis and
	 * sends a {@link PlotChangeEvent} to all registered listeners.
	 * 
	 * @param paint
	 *            the paint (<code>null</code> not permitted).
	 * 
	 * @see #getRangeGridlinePaint()
	 */
	public void setRangeGridlinePaint(Paint paint) {
		ParamChecks.nullNotPermitted(paint, "paint");
		this.rangeGridlinePaint = paint;
		fireChangeEvent();
	}

	/**
	 * Draws the fast scatter plot on a Java 2D graphics device (such as the
	 * screen or a printer).
	 * 
	 * @param g2
	 *            the graphics device.
	 * @param area
	 *            the area within which the plot (including axis labels) should
	 *            be drawn.
	 * @param anchor
	 *            the anchor point (<code>null</code> permitted).
	 * @param parentState
	 *            the state from the parent plot (ignored).
	 * @param info
	 *            collects chart drawing information (<code>null</code>
	 *            permitted).
	 */
	@Override
	public void draw(Graphics2D g2, Rectangle2D area, Point2D anchor, PlotState parentState, PlotRenderingInfo info) {

		// set up info collection...
		if (info != null) {
			info.setPlotArea(area);
		}

		// adjust the drawing area for plot insets (if any)...
		RectangleInsets insets = getInsets();
		insets.trim(area);

		AxisSpace space = calculateAxisSpace(g2, area);

		Rectangle2D dataArea = space.shrink(area, null);

		if (info != null) {
			info.setDataArea(dataArea);
		}

		// draw the plot background and axes...
		drawBackground(g2, dataArea);

		AxisState domainAxisState = this.domainAxis.draw(g2, dataArea.getMaxY(), area, dataArea, RectangleEdge.BOTTOM,
				info);
		AxisState rangeAxisState = this.rangeAxis
				.draw(g2, dataArea.getMinX(), area, dataArea, RectangleEdge.LEFT, info);
		drawDomainGridlines(g2, dataArea, domainAxisState.getTicks());
		drawRangeGridlines(g2, dataArea, rangeAxisState.getTicks());

		Shape originalClip = g2.getClip();
		Composite originalComposite = g2.getComposite();

		g2.clip(dataArea);
		g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getForegroundAlpha()));

		render(g2, dataArea, info, null);

		g2.setClip(originalClip);
		g2.setComposite(originalComposite);
		drawOutline(g2, dataArea);

	}

	protected AxisSpace calculateAxisSpace(Graphics2D g2, Rectangle2D plotArea) {
		AxisSpace space = new AxisSpace();
		space = calculateRangeAxisSpace(g2, plotArea, space);
		space = calculateDomainAxisSpace(g2, plotArea, space);
		return space;
	}

	/**
	 * Calculates the space required for the domain axis/axes.
	 * 
	 * @param g2
	 *            the graphics device.
	 * @param plotArea
	 *            the plot area.
	 * @param space
	 *            a carrier for the result (<code>null</code> permitted).
	 * 
	 * @return The required space.
	 */
	protected AxisSpace calculateDomainAxisSpace(Graphics2D g2, Rectangle2D plotArea, AxisSpace space) {
		if (space == null) {
			space = new AxisSpace();
		}

		AxisSpace tmpSpace = new AxisSpace();
		this.domainAxis.reserveSpace(g2, this, plotArea, RectangleEdge.BOTTOM, tmpSpace);
		space.ensureAtLeast(Math.min(tmpSpace.getBottom(), plotArea.getHeight() * .2), RectangleEdge.BOTTOM);

		return space;

	}

	/**
	 * Calculates the space required for the range axis/axes.
	 * 
	 * @param g2
	 *            the graphics device.
	 * @param plotArea
	 *            the plot area.
	 * @param space
	 *            a carrier for the result (<code>null</code> permitted).
	 * 
	 * @return The required space.
	 */
	protected AxisSpace calculateRangeAxisSpace(Graphics2D g2, Rectangle2D plotArea, AxisSpace space) {

		if (space == null) {
			space = new AxisSpace();
		}

		AxisSpace tmpSpace = new AxisSpace();
		this.rangeAxis.reserveSpace(g2, this, plotArea, RectangleEdge.LEFT, tmpSpace);
		space.ensureAtLeast(Math.min(tmpSpace.getLeft(), plotArea.getWidth() * .2), RectangleEdge.LEFT);

		return space;

	}

	/**
	 * Draws a representation of the data within the dataArea region. The
	 * <code>info</code> and <code>crosshairState</code> arguments may be
	 * <code>null</code>.
	 * 
	 * @param g2
	 *            the graphics device.
	 * @param dataArea
	 *            the region in which the data is to be drawn.
	 * @param info
	 *            an optional object for collection dimension information.
	 * @param crosshairState
	 *            collects crosshair information (<code>null</code> permitted).
	 */
	public void render(Graphics2D g2, Rectangle2D dataArea, PlotRenderingInfo info, CrosshairState crosshairState) {

		read.lock();
		try {
			random.reset();

			// domainAxis.getRange().getLength()/ xDataLength;
			// rangeAxis.getRange().getLength()/yDataLength;

			long start = System.currentTimeMillis();

			for (int i = 0; i < xBucketCount.length; i++) {
				Arrays.fill(xBucketCount[i], 0);
			}

			double zx = xDataLength / domainAxis.getRange().getLength();
			if (xIsDiscrete) {
				// compute pixels per value
				zx *= dataArea.getWidth() / domainAxis.getRange().getLength();
			} else {
				// Assume a preference for a width of 1200 pixels
				zx *= dataArea.getWidth() / 1200.0;
			}
			double zy = yDataLength / rangeAxis.getRange().getLength();
			if (yIsDiscrete) {
				// compute pixels per value
				zy *= dataArea.getHeight() / Math.round(rangeAxis.getRange().getLength());
			} else {
				// Assume a preference for a height of 800 pixels.
				zy *= dataArea.getHeight() / 800.0;
			}
			double halfDotSize = dotScale * minimumSize * Math.max(1.0, Math.sqrt(Math.min(625, Math.min(zx, zy))));
			this.dotSize = halfDotSize;
			halfDotSize /= 2;

			double bucketFactor = dataArea.getWidth() / xBucketCount.length;

			// First count the events we want to render before actually
			// rendering them. This method is fast compared to the
			// rendering itself.
			int count;
//			if(log.getTimeType().equals("absolute")){
//				count = renderEvents(g2, dataArea, halfDotSize, bucketFactor, true);
//			}
//			else{
//				count = renderEventsRelativeTime(g2, dataArea, halfDotSize, bucketFactor, true);
//			}
			
			count = renderEvents(g2, dataArea, halfDotSize, bucketFactor, true);

			
			this.numVisisbleEvents = count;
			// long dStart = System.currentTimeMillis();

			// Now that we know how many events are visible, render the events
			int max;
//			if(log.getTimeType().equals("absolute")){
//				max = renderEvents(g2, dataArea, halfDotSize, bucketFactor, false);
//			}
//			else{
//				max = renderEventsRelativeTime(g2, dataArea, halfDotSize, bucketFactor, false);
//			}
			
			max = renderEvents(g2, dataArea, halfDotSize, bucketFactor, false);
//			max = relateEvents(g2, dataArea, halfDotSize, bucketFactor, true);
			//connect same events by Guangming on June 19th, 2018
			renderSameEvents(g2, dataArea, halfDotSize, bucketFactor, false);
			
			long iStart = System.currentTimeMillis();
			this.lastRenderTime = (iStart - start) / 1000.0;

			if (first) {
				// rendered the first time, with limited dotCount;
				maxDotCount = 500 * Math.min(maxDotCount, count) / (iStart - start);
				first = false;
				render(g2, dataArea, info, crosshairState);
			}

			if (intensityPanel != null) {
				intensityChart.setData(xBucketCount, max, colorMap, dataArea.getMinX());
				intensityPanel.repaint();
			}

			// long end = System.currentTimeMillis();

			// XXX: DEBUG CODE
			// System.out.println("Time to render all:    " + (end - start));
			// System.out.println("Time to count:         " + (dStart - start));
			// System.out.println("Time to draw:          " + (iStart -
			// dStart));
			// System.out.println("Time to render bottom: " + (end - iStart));
			// System.out.println("Fraction: " + renderFraction);
		} finally {
			read.unlock();
		}
	}

	public void setLinesOn(boolean lines, boolean signalChange) {
		this.lines = lines;
		if (signalChange) {
			fireChangeEvent();
		}
	}

	public void connectSameEvents(boolean connectSameEventFlag, boolean signalChange) {
		this.connectSameEventFlag = connectSameEventFlag;
		if (signalChange) {
			fireChangeEvent();
		}
	}
	
	private int renderEvents(Graphics2D g2, Rectangle2D dataArea, double halfDotSize, double bucketFactor,
			boolean countOnly) {

		g2.setStroke(DEFAULT_LINE_STROKE);

		double x = 0, y = 0;
		int lastX, lastY;
		int m = 0;

//		if(this.domainAxis instanceof DateAxis){
//        	((DateAxis) this.domainAxis).setDateFormatOverride(new SimpleDateFormat("yyyy-MM-dd"));
//        }
		
		String referenceActivity = log.getReferenceActivity();
		String targetActivity = log.getTargetActivity();
		
		for (int t = 0; t < log.numTraces(); t++) {
			lastX = -1;
			lastY = -1;


			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			int eReference = -1;
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				String activity = log.getEventDataManager().readObject(e).getAttributes().get("concept:name").toString();
				if(activity.equals(referenceActivity)){
					eReference = e;
					System.out.println("eReference: " + eReference);
					break;
				}
			}
			
			if (!xAttribute.isEventAttribute()) {
				x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			}
			if (!yAttribute.isEventAttribute()) {
				y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
			}
			
			if (xAttribute.isEventAttribute()) {
				x = getValue(log.getEventDataManager(), eReference, xAttributeIndex, xAttribute.getType(),
						xCategory2LitIndex);
			}
			if (yAttribute.isEventAttribute()) {
				y = getValue(log.getEventDataManager(), eReference, yAttributeIndex, yAttribute.getType(),
						yCategory2LitIndex);
			}
			
			lastX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
			lastY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));
			
			
			
			if (!xAttribute.isEventAttribute()) {
				x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			}
			if (!yAttribute.isEventAttribute()) {
				y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
			}
//			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			for (int e = log.getFirstEventNumber(t); e < max; e++) {

				if (xAttribute.isEventAttribute()) {
					x = getValue(log.getEventDataManager(), e, xAttributeIndex, xAttribute.getType(),
							xCategory2LitIndex);
				}
				if (yAttribute.isEventAttribute()) {
					y = getValue(log.getEventDataManager(), e, yAttributeIndex, yAttribute.getType(),
							yCategory2LitIndex);
				}

				if (countOnly) {
					m++;
				} else {
					g2.setPaint(colorMap[colorAndShape[e] & COLORMASK]);
					//					g2.setPaint(colorMap[colors[e] & 0xFF]);

					int transX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
					int transY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));

					// to fill the counters we translate the x value to a value
					// from
					// 0-255

					if (numVisisbleEvents < maxDotCount || random.next() < maxDotCount / numVisisbleEvents) {

						if (lines && lastY >= 0) {
//						if (lines) {
							g2.setColor(Color.BLACK);
							g2.drawLine(lastX, lastY, transX, transY);
						}
//						lastX = transX;
//						lastY = transY;

//						lastX -= halfDotSize;
//						lastY -= halfDotSize;
						
						transX -= halfDotSize;
						transY -= halfDotSize;
						
						String referenceTarget = log.getEventDataManager().readObject(e).getAttributes().get("pattern:name").toString();
						
//						if(referenceTarget.equals("reference")){
//							g2.setColor(Color.pink);
//						}
//						else if(referenceTarget.equals("target")){
//							g2.setColor(Color.yellow);
//						}
						g2.setPaint(colorMap[colorAndShape[e] & COLORMASK]);
						//						switch (shapeMap[shapes[e]]) {
						byte shape = shapeMap[(colorAndShape[e] & SHAPEMASK) >> 10];

						renderShape(g2, halfDotSize, transX, transY, shape);
					} else {
						transX -= halfDotSize;
						transY -= halfDotSize;
						
//						lastX -= halfDotSize;
//						lastY -= halfDotSize;
					}
				}
			}
		}
		return m;

	}

	private int renderEventsRelativeTime(Graphics2D g2, Rectangle2D dataArea, double halfDotSize, double bucketFactor,
			boolean countOnly) {

		g2.setStroke(DEFAULT_LINE_STROKE);

		double x = 0, y = 0;
		int lastX, lastY;
		int m = 0;
		eventPositionX.clear();
		instancePositionX.clear();
		String referenceActivity = log.getReferenceActivity();
		String targetActivity = log.getTargetActivity();
		
		int firstReferenceEventNumber = 0;
		for (int t = 0; t < log.numTraces(); t++) {
			lastX = -1;
			lastY = -1;


			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			int eReference = -1;
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				String activity = log.getEventDataManager().readObject(e).getAttributes().get("concept:name").toString();
				if(activity.equals(referenceActivity)){
					eReference = e;
//					System.out.println("eReference: " + eReference);
					break;
				}
			}
			
			if(t == 0){
				firstReferenceEventNumber = eReference;
			}
			if (!xAttribute.isEventAttribute()) {
				x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			}
			if (!yAttribute.isEventAttribute()) {
				y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
			}
			
			if (xAttribute.isEventAttribute()) {
				x = getValue(log.getEventDataManager(), eReference, xAttributeIndex, xAttribute.getType(),
						xCategory2LitIndex);
			}
			if (yAttribute.isEventAttribute()) {
				y = getValue(log.getEventDataManager(), eReference, yAttributeIndex, yAttribute.getType(),
						yCategory2LitIndex);
			}
			
			lastX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
			lastY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));
			
			
			
			if (!xAttribute.isEventAttribute()) {
				x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
//				if (x < domainAxis.getRange().getLowerBound() || x > domainAxis.getRange().getUpperBound()) {
//					continue;
//				}
			}
			if (!yAttribute.isEventAttribute()) {
				y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
//				if (y < rangeAxis.getRange().getLowerBound() || y > rangeAxis.getRange().getUpperBound()) {
//					continue;
//				}
			}
//			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			for (int e = log.getFirstEventNumber(t); e < max; e++) {

				if (xAttribute.isEventAttribute()) {
					x = getValue(log.getEventDataManager(), e, xAttributeIndex, xAttribute.getType(),
							xCategory2LitIndex);
//					if (x < domainAxis.getRange().getLowerBound() || x > domainAxis.getRange().getUpperBound()) {
//						continue;
//					}
				}
				if (yAttribute.isEventAttribute()) {
					y = getValue(log.getEventDataManager(), e, yAttributeIndex, yAttribute.getType(),
							yCategory2LitIndex);
//					if (y < rangeAxis.getRange().getLowerBound() || y > rangeAxis.getRange().getUpperBound()) {
//						continue;
//					}
				}

				if (countOnly) {
					m++;
				} else {
					g2.setPaint(colorMap[colorAndShape[e] & COLORMASK]);
					//					g2.setPaint(colorMap[colors[e] & 0xFF]);

					
					int transX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
					int transY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));

					// to fill the counters we translate the x value to a value
					// from
					// 0-255

					if (numVisisbleEvents < maxDotCount || random.next() < maxDotCount / numVisisbleEvents) {
						
						transX -= halfDotSize;
						transY -= halfDotSize;

						//						switch (shapeMap[shapes[e]]) {
//						byte shape = shapeMap[(colorAndShape[e] & SHAPEMASK) >> 10];
//
//						renderShape(g2, halfDotSize, transX, transY, shape);
						
//						System.out.println("e after renderShape: " + e);
						Pair<Integer, Integer> position = new Pair<Integer, Integer>(transX, transY);
						eventPositionX.put(e, position);
						String activity = log.getEventDataManager().readObject(e).getAttributes().get("concept:name").toString();						
						if(referenceActivity.equals(activity)){							
							instancePositionX.put(t, position);
						}
						
					} else {
						transX -= halfDotSize;
						transY -= halfDotSize;						
					}
					
					if (x < domainAxis.getRange().getLowerBound() || x > domainAxis.getRange().getUpperBound()) {
						continue;
					}
					if (y < rangeAxis.getRange().getLowerBound() || y > rangeAxis.getRange().getUpperBound()) {
						continue;
					}					
				}
			}
		}
		
		
		//set the coordinates as relative coordinates (the first reference event is the base)
		DateAxis dateAxis = new DateAxis();

		XAttributeTimestamp firstReferenceEventTime = (XAttributeTimestamp) log.getEventDataManager().
		readObject(log.getFirstEventNumber(firstReferenceEventNumber)).
		getAttributes().get("time:timestamp");
		
		
		System.out.println("timeStamp.getValue().getTime();: " + firstReferenceEventTime.getValue().getTime());
		

        RelativeDateFormat relativeDateFormat = new RelativeDateFormat();
        relativeDateFormat.setBaseMillis(firstReferenceEventTime.getValue().getTime());
        dateAxis.setDateFormatOverride(relativeDateFormat);
        relativeDateFormat.setShowZeroHours(true);
        if(this.domainAxis instanceof DateAxis){
        	((DateAxis) this.domainAxis).setDateFormatOverride(relativeDateFormat);
        }
        
		//render events based on relative coordinates
		int baseX = 0;
		int baseY = 0;
		if(instancePositionX.containsKey(0)){
			baseX = (int) instancePositionX.get(0).getKey();
			baseY = (int) instancePositionX.get(0).getValue();
		}
		
		for (int t = 0; t < log.numTraces(); t++) {
			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			if(instancePositionX.containsKey(t)){
				int referenceX = (int) instancePositionX.get(t).getKey();
				int referenceY = (int) instancePositionX.get(t).getValue();
				for (int e = log.getFirstEventNumber(t); e < max; e++) {
					byte shape = shapeMap[(colorAndShape[e] & SHAPEMASK) >> 10];
					g2.setPaint(colorMap[colorAndShape[e] & COLORMASK]);
					if(eventPositionX.containsKey(e)){
						int transX = (int) eventPositionX.get(e).getKey();
						int transY = (int) eventPositionX.get(e).getValue();
						renderShape(g2, halfDotSize, transX - (referenceX - baseX), transY, shape);						
						
						if (numVisisbleEvents < maxDotCount || random.next() < maxDotCount / numVisisbleEvents) {

							if (lines && referenceY >= 0) {
//							if (lines) {
								g2.drawLine(referenceX - (referenceX - baseX) + (int) halfDotSize, 
										referenceY + (int) halfDotSize, transX - (referenceX - baseX) + (int) halfDotSize, transY + (int) halfDotSize);
							}
						}
						
						x = domainAxis.java2DToValue(transX - (referenceX - baseX), dataArea, RectangleEdge.BOTTOM);
						y = rangeAxis.java2DToValue(transY, dataArea, RectangleEdge.LEFT);
						if (x < domainAxis.getRange().getLowerBound() || x  > domainAxis.getRange().getUpperBound()) {
							continue;
						}
						if (y < rangeAxis.getRange().getLowerBound() || y > rangeAxis.getRange().getUpperBound()) {
							continue;
						}
						//- (referenceX - baseX)
						int count = (int) ((transX - (referenceX - baseX) + halfDotSize - dataArea.getMinX()) / bucketFactor);
//						System.out.println("(int) ((transX - (referenceX - baseX) + halfDotSize - dataArea.getMinX()) / bucketFactor): " + (int) ((transX - (referenceX - baseX) + halfDotSize - dataArea.getMinX()) / bucketFactor));
						if(0 <= count && count < xBucketCount.length){
							int me = ++xBucketCount[count][(colorAndShape[e] & COLORMASK)];
							if (me > m) {
								m = me;
							}
						}					
					}					
				}			
			}						
		}
		return m;

	}
	
	/**
	 * connect each reference event to its all corresponding target events 
	 * @param g2
	 * @param dataArea
	 * @param halfDotSize
	 * @param bucketFactor
	 * @param countOnly
	 * @return
	 */
	private int relateEvents(Graphics2D g2, Rectangle2D dataArea, double halfDotSize, double bucketFactor,
			boolean countOnly) {

		g2.setStroke(DEFAULT_LINE_STROKE);

		double x = 0, y = 0;
		int lastX, lastY;
		int m = 0;


		String referenceActivity = log.getReferenceActivity();
		String targetActivity = log.getTargetActivity();
		
		for (int t = 0; t < log.numTraces(); t++) {
			lastX = -1;
			lastY = -1;

			//get the reference event
			
//			int eReference = -1;
//			for (int e = log.getFirstEventNumber(t); e < max; e++) {
//				String activity = log.getEventDataManager().readObject(e).getAttributes().get("concept:name").toString();
//				if(activity.equals(referenceActivity)){
//					eReference = e;
//					System.out.println("eReference: " + eReference);
//					break;
//				}
//			}
//			
//			if (!xAttribute.isEventAttribute()) {
//				x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
//			}
//			if (!yAttribute.isEventAttribute()) {
//				y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
//			}
//			
//			if (xAttribute.isEventAttribute()) {
//				x = getValue(log.getEventDataManager(), eReference, xAttributeIndex, xAttribute.getType(),
//						xCategory2LitIndex);
//			}
//			if (yAttribute.isEventAttribute()) {
//				y = getValue(log.getEventDataManager(), eReference, yAttributeIndex, yAttribute.getType(),
//						yCategory2LitIndex);
//			}
//			
//			lastX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
//			lastY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));
//			
//			lastX -= halfDotSize;
//			lastY -= halfDotSize;
			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			if (!xAttribute.isEventAttribute()) {
				x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			}
			if (!yAttribute.isEventAttribute()) {
				y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
			}

			for (int e = log.getFirstEventNumber(t); e < max; e++) {

				if (xAttribute.isEventAttribute()) {
					x = getValue(log.getEventDataManager(), e, xAttributeIndex, xAttribute.getType(),
							xCategory2LitIndex);
				}
				if (yAttribute.isEventAttribute()) {
					y = getValue(log.getEventDataManager(), e, yAttributeIndex, yAttribute.getType(),
							yCategory2LitIndex);
				}

				if (countOnly) {
					m++;
				} else {
					g2.setPaint(colorMap[colorAndShape[e] & COLORMASK]);
					//					g2.setPaint(colorMap[colors[e] & 0xFF]);

					int transX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
					int transY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));

					// to fill the counters we translate the x value to a value
					// from
					// 0-255

					if (numVisisbleEvents < maxDotCount || random.next() < maxDotCount / numVisisbleEvents) {

						if (lines && lastY >= 0) {
//						if (lines) {
							g2.drawLine(lastX, lastY, transX, transY);
						}
						lastX = transX;
						lastY = transY;

						transX -= halfDotSize;
						transY -= halfDotSize;

						//						switch (shapeMap[shapes[e]]) {
						byte shape = shapeMap[(colorAndShape[e] & SHAPEMASK) >> 10];

						renderShape(g2, halfDotSize, transX, transY, shape);
					} else {
						transX -= halfDotSize;
						transY -= halfDotSize;
					}
				}
			}
		}
		return m;

	}
	
	private int relateEventsOld(Graphics2D g2, Rectangle2D dataArea, double halfDotSize, double bucketFactor,
			boolean countOnly) {
		System.out.println("renderEvents: ");
		g2.setStroke(DEFAULT_LINE_STROKE);

		double x = 0, y = 0;
		int lastX, lastY;
		int m = 0;

		String referenceActivity = log.getReferenceActivity();
		String targetActivity = log.getTargetActivity();
		
		System.out.println("referenceActivity: " + referenceActivity);
		System.out.println("targetActivity: " + targetActivity);
		for (int t = 0; t < log.numTraces(); t++) {
			lastX = -1;
			lastY = -1;
			
			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			int eReference = -1;
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				String activity = log.getEventDataManager().readObject(e).getAttributes().get("concept:name").toString();
				if(activity.equals(referenceActivity)){
					eReference = e;
					System.out.println("eReference: " + eReference);
					break;
				}
			}
			
			if (!xAttribute.isEventAttribute()) {
				x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			}
			if (!yAttribute.isEventAttribute()) {
				y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
			}
			
			if (xAttribute.isEventAttribute()) {
				x = getValue(log.getEventDataManager(), eReference, xAttributeIndex, xAttribute.getType(),
						xCategory2LitIndex);
			}
			if (yAttribute.isEventAttribute()) {
				y = getValue(log.getEventDataManager(), eReference, yAttributeIndex, yAttribute.getType(),
						yCategory2LitIndex);
			}
			lastX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
			lastY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));
			
			lastX -= halfDotSize;
			lastY -= halfDotSize;
			
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				if (xAttribute.isEventAttribute()) {
					x = getValue(log.getEventDataManager(), e, xAttributeIndex, xAttribute.getType(),
							xCategory2LitIndex);
				}
				if (yAttribute.isEventAttribute()) {
					y = getValue(log.getEventDataManager(), e, yAttributeIndex, yAttribute.getType(),
							yCategory2LitIndex);
				}
		
				int transX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
				int transY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));
				
				transX -= halfDotSize;
				transY -= halfDotSize;
				
				if (lines && lastY >= 0) {
//					if (lines) {
//						g2.drawLine(lastX, lastY, transX, transY);
					g2.drawLine(lastX, lastY, transX, transY);
				}
				
				transX -= halfDotSize;
				transY -= halfDotSize;
				
				g2.setPaint(colorMap[colorAndShape[e] & COLORMASK]);
				byte shape = shapeMap[(colorAndShape[e] & SHAPEMASK) >> 10];
				renderShape(g2, halfDotSize, transX, transY, shape);
				
				
			}
			
		}
		return m;

	}
	
	/**
	 * draw line between the same events
	 * @param g2
	 * @param dataArea
	 * @param halfDotSize
	 * @param bucketFactor
	 * @param countOnly
	 * @return
	 */
	private int renderSameEvents(Graphics2D g2, Rectangle2D dataArea, double halfDotSize, double bucketFactor,
			boolean countOnly) {
//		System.out.println("enter renderSameEvents");
		g2.setStroke(DEFAULT_LINE_STROKE);

		double x = 0, y = 0;
		int m = 0;

		for (int t = 0; t < log.numTraces()-1; t++) {
//			System.out.println("t: " + t);
			if (!xAttribute.isEventAttribute()) {
				x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			}
			if (!yAttribute.isEventAttribute()) {
				y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
			}
			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
//			System.out.println("log.getFirstEventNumber(t): " + log.getFirstEventNumber(t));
//			System.out.println("log.getTraceLength(t): " + log.getTraceLength(t));
//			System.out.println("max: " + max);
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				String eventId = log.getEventDataManager().readObject(e).getAttributes().get("id:name").toString();
//				System.out.println("e: " + e);
//				System.out.println("renderSameEvents eventId: " + eventId);
				if (xAttribute.isEventAttribute()) {
					x = getValue(log.getEventDataManager(), e, xAttributeIndex, xAttribute.getType(),
							xCategory2LitIndex);
				}
				if (yAttribute.isEventAttribute()) {
					y = getValue(log.getEventDataManager(), e, yAttributeIndex, yAttribute.getType(),
							yCategory2LitIndex);
				}

				int transX = (int) (domainAxis.valueToJava2D(x, dataArea, RectangleEdge.BOTTOM));
				int transY = (int) (rangeAxis.valueToJava2D(y, dataArea, RectangleEdge.LEFT));
	
				//check the Next (t+1) case
				double xN = 0, yN = 0;
				if (!xAttribute.isEventAttribute()) {
					xN = getValue(log.getTraceDataManager(), t+1, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
				}
				if (!yAttribute.isEventAttribute()) {
					yN = getValue(log.getTraceDataManager(), t+1, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
				}
				int maxN = log.getFirstEventNumber(t+1) + log.getTraceLength(t+1);
				for (int eN = log.getFirstEventNumber(t+1); eN < maxN; eN++) {
					String eventIdN = log.getEventDataManager().readObject(eN).getAttributes().get("id:name").toString();
					
					if (xAttribute.isEventAttribute()) {
						xN = getValue(log.getEventDataManager(), eN, xAttributeIndex, xAttribute.getType(),
								xCategory2LitIndex);
					}
					if (yAttribute.isEventAttribute()) {
						yN = getValue(log.getEventDataManager(), eN, yAttributeIndex, yAttribute.getType(),
								yCategory2LitIndex);
					}
					
					int transXN = (int) (domainAxis.valueToJava2D(xN, dataArea, RectangleEdge.BOTTOM));
					int transYN = (int) (rangeAxis.valueToJava2D(yN, dataArea, RectangleEdge.LEFT));
					
					if (connectSameEventFlag && eventIdN.equals(eventId)) {
						g2.drawLine(transX, transY, transXN, transYN);
					}						
				}					
			}			
		}
		return m;
	}
	
	protected void renderShape(Graphics2D g2, double halfDotSize, int transX, int transY, byte shape) {
		switch (shape) {
			case 0 :
				// circle
				g2.fillOval(transX + (int) (halfDotSize * 0.1), transY + (int) (halfDotSize * 0.1),
						(int) (halfDotSize * 1.8), (int) (halfDotSize * 1.8));
				break;

			case 1 :
				// square
				g2.fillRect(transX + (int) (halfDotSize * 0.15), transY + (int) (halfDotSize * 0.15),
						(int) (halfDotSize * 1.7), (int) (halfDotSize * 1.7));

				break;

			case 2 :
				// triangle up
				g2.fillPolygon(new int[] { transX, transX + (int) halfDotSize, transX + (int) (2 * halfDotSize) },
						new int[] { transY + (int) (1.7 * halfDotSize), transY, transY + (int) (1.7 * halfDotSize) }, 3);
				break;

			case 3 :
				// triangle down
				g2.fillPolygon(new int[] { transX, transX + (int) halfDotSize, transX + (int) (2 * halfDotSize) },
						new int[] { transY + (int) (0.3 * halfDotSize), transY + (int) (2 * halfDotSize),
								transY + (int) (0.3 * halfDotSize) }, 3);
				break;
			case 4 :
				// diamond
				g2.fillPolygon(new int[] { transX, transX + (int) halfDotSize, transX + (int) (2 * halfDotSize),
						transX + (int) halfDotSize }, new int[] { transY + (int) (halfDotSize), transY,
						transY + (int) (halfDotSize), transY + (int) (2 * halfDotSize) }, 4);
				break;
			case 5 :
				// pentagon
				g2.fillPolygon(new int[] { transX, transX + (int) halfDotSize, transX + (int) (2 * halfDotSize),
						transX + (int) (1.58 * halfDotSize), transX + (int) (0.42 * halfDotSize) }, new int[] {
						transY + (int) (0.95 * halfDotSize), transY, transY + (int) (0.95 * halfDotSize),
						transY + (int) (2 * halfDotSize), transY + (int) (2 * halfDotSize) }, 5);
				break;
			case 6 :
				// star
				g2.fillPolygon(new int[] { transX, transX + (int) (2 * halfDotSize),
						transX + (int) (0.42 * halfDotSize), transX + (int) halfDotSize,
						transX + (int) (1.58 * halfDotSize) }, new int[] { transY + (int) (0.95 * halfDotSize),
						transY + (int) (0.95 * halfDotSize), transY + (int) (2 * halfDotSize), transY,
						transY + (int) (2 * halfDotSize) }, 5);
				break;
			case 7 :
				// bowtie
				g2.fillPolygon(new int[] { transX, transX + (int) (2 * halfDotSize), transX + (int) halfDotSize,
						transX + (int) halfDotSize }, new int[] { transY + (int) (halfDotSize),
						transY + (int) (halfDotSize), transY, transY + (int) (2 * halfDotSize) }, 4);
				break;

		}
	}

	public double getDotScale() {
		return this.dotScale;
	}

	public void setDotScale(double scale, boolean signalChange) {
		this.dotScale = scale;
		if (signalChange) {
			fireChangeEvent();
		}
	}

	public void setDefaultSize(double size) {
		this.minimumSize = size;
	}

	public double getDefaultSize() {
		return this.minimumSize;
	}

	private double getValue(DataManager<?> manager, int o, int attributeIndex, byte type,
			TIntObjectMap<CategoryInfo> categoryMap) {
		switch (type) {
			case Attribute.BOOL :
				return manager.getAttributeValueBoolean(o, attributeIndex) ? 1 : 0;
			case Attribute.DISC :
				return manager.getAttributeValueDiscrete(o, attributeIndex);
			case Attribute.CONT :
				return manager.getAttributeValueContinuous(o, attributeIndex);
			case Attribute.TIME_ABS :
			case Attribute.TIME_REL :
				return manager.getAttributeValueTimestamp(o, attributeIndex);
			case Attribute.LITR :
				// return manager.getAttributeValueLiteralIndex(o, attribute);
				return categoryMap.get(manager.getAttributeValueLiteralIndex(o, attributeIndex)).categoryNumber;
			default :
				return 0;
		}
	}

	/**
	 * Draws the gridlines for the plot, if they are visible.
	 * 
	 * @param g2
	 *            the graphics device.
	 * @param dataArea
	 *            the data area.
	 * @param ticks
	 *            the ticks.
	 */
	@SuppressWarnings("rawtypes")
	protected void drawDomainGridlines(Graphics2D g2, Rectangle2D dataArea, List ticks) {
		if (!isDomainGridlinesVisible()) {
			return;
		}
		Object saved = g2.getRenderingHint(RenderingHints.KEY_STROKE_CONTROL);
		g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);
		Iterator iterator = ticks.iterator();
		while (iterator.hasNext()) {
			ValueTick tick = (ValueTick) iterator.next();
			double v = this.domainAxis.valueToJava2D(tick.getValue(), dataArea, RectangleEdge.BOTTOM);
			Line2D line = new Line2D.Double(v, dataArea.getMinY(), v, dataArea.getMaxY());
			g2.setPaint(getDomainGridlinePaint());
			g2.setStroke(getDomainGridlineStroke());
			g2.draw(line);
		}
		g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, saved);
	}

	/**
	 * Draws the gridlines for the plot, if they are visible.
	 * 
	 * @param g2
	 *            the graphics device.
	 * @param dataArea
	 *            the data area.
	 * @param ticks
	 *            the ticks.
	 */
	@SuppressWarnings("rawtypes")
	protected void drawRangeGridlines(Graphics2D g2, Rectangle2D dataArea, List ticks) {

		if (!isRangeGridlinesVisible()) {
			return;
		}
		Object saved = g2.getRenderingHint(RenderingHints.KEY_STROKE_CONTROL);
		g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);

		Iterator iterator = ticks.iterator();
		while (iterator.hasNext()) {
			ValueTick tick = (ValueTick) iterator.next();
			double v = this.rangeAxis.valueToJava2D(tick.getValue(), dataArea, RectangleEdge.LEFT);
			Line2D line = new Line2D.Double(dataArea.getMinX(), v, dataArea.getMaxX(), v);
			g2.setPaint(getRangeGridlinePaint());
			g2.setStroke(getRangeGridlineStroke());
			g2.draw(line);
		}
		g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, saved);
	}

	/**
	 * Returns the range of data values to be plotted along the axis, or
	 * <code>null</code> if the specified axis isn't the domain axis or the
	 * range axis for the plot.
	 * 
	 * @param axis
	 *            the axis (<code>null</code> permitted).
	 * 
	 * @return The range (possibly <code>null</code>).
	 */
	@Override
	public Range getDataRange(ValueAxis axis) {
		Range result = null;
		if (axis == this.domainAxis) {
			result = this.xDataRange;
		} else if (axis == this.rangeAxis) {
			result = this.yDataRange;
		}
		return result;
	}

	/**
	 * /** Multiplies the range on the domain axis by the specified factor.
	 * 
	 * @param factor
	 *            the zoom factor.
	 * @param info
	 *            the plot rendering info.
	 * @param source
	 *            the source point.
	 */
	@Override
	public void zoomDomainAxes(double factor, PlotRenderingInfo info, Point2D source) {
		this.domainAxis.resizeRange(factor);
	}

	/**
	 * Multiplies the range on the domain axis by the specified factor.
	 * 
	 * @param factor
	 *            the zoom factor.
	 * @param info
	 *            the plot rendering info.
	 * @param source
	 *            the source point (in Java2D space).
	 * @param useAnchor
	 *            use source point as zoom anchor?
	 * 
	 * @see #zoomRangeAxes(double, PlotRenderingInfo, Point2D, boolean)
	 * 
	 * @since 1.0.7
	 */
	@Override
	public void zoomDomainAxes(double factor, PlotRenderingInfo info, Point2D source, boolean useAnchor) {

		if (useAnchor) {
			// get the source coordinate - this plot has always a VERTICAL
			// orientation
			double sourceX = source.getX();
			double anchorX = this.domainAxis.java2DToValue(sourceX, info.getDataArea(), RectangleEdge.BOTTOM);
			this.domainAxis.resizeRange2(factor, anchorX);
		} else {
			this.domainAxis.resizeRange(factor);
		}

	}

	/**
	 * Zooms in on the domain axes.
	 * 
	 * @param lowerPercent
	 *            the new lower bound as a percentage of the current range.
	 * @param upperPercent
	 *            the new upper bound as a percentage of the current range.
	 * @param info
	 *            the plot rendering info.
	 * @param source
	 *            the source point.
	 */
	@Override
	public void zoomDomainAxes(double lowerPercent, double upperPercent, PlotRenderingInfo info, Point2D source) {
		this.domainAxis.zoomRange(lowerPercent, upperPercent);
	}

	/**
	 * Multiplies the range on the range axis/axes by the specified factor.
	 * 
	 * @param factor
	 *            the zoom factor.
	 * @param info
	 *            the plot rendering info.
	 * @param source
	 *            the source point.
	 */
	@Override
	public void zoomRangeAxes(double factor, PlotRenderingInfo info, Point2D source) {
		this.rangeAxis.resizeRange(factor);
	}

	/**
	 * Multiplies the range on the range axis by the specified factor.
	 * 
	 * @param factor
	 *            the zoom factor.
	 * @param info
	 *            the plot rendering info.
	 * @param source
	 *            the source point (in Java2D space).
	 * @param useAnchor
	 *            use source point as zoom anchor?
	 * 
	 * @see #zoomDomainAxes(double, PlotRenderingInfo, Point2D, boolean)
	 * 
	 * @since 1.0.7
	 */
	@Override
	public void zoomRangeAxes(double factor, PlotRenderingInfo info, Point2D source, boolean useAnchor) {

		if (useAnchor) {
			// get the source coordinate - this plot has always a VERTICAL
			// orientation
			double sourceY = source.getY();
			double anchorY = this.rangeAxis.java2DToValue(sourceY, info.getDataArea(), RectangleEdge.LEFT);
			this.rangeAxis.resizeRange2(factor, anchorY);
		} else {
			this.rangeAxis.resizeRange(factor);
		}

	}

	/**
	 * Zooms in on the range axes.
	 * 
	 * @param lowerPercent
	 *            the new lower bound as a percentage of the current range.
	 * @param upperPercent
	 *            the new upper bound as a percentage of the current range.
	 * @param info
	 *            the plot rendering info.
	 * @param source
	 *            the source point.
	 */
	@Override
	public void zoomRangeAxes(double lowerPercent, double upperPercent, PlotRenderingInfo info, Point2D source) {
		this.rangeAxis.zoomRange(lowerPercent, upperPercent);
	}

	/**
	 * Returns <code>true</code>.
	 * 
	 * @return A boolean.
	 */
	@Override
	public boolean isDomainZoomable() {
		return true;
	}

	/**
	 * Returns <code>true</code>.
	 * 
	 * @return A boolean.
	 */
	@Override
	public boolean isRangeZoomable() {
		return true;
	}

	/**
	 * Returns <code>true</code> if panning is enabled for the domain axes, and
	 * <code>false</code> otherwise.
	 * 
	 * @return A boolean.
	 * 
	 * @since 1.0.13
	 */
	@Override
	public boolean isDomainPannable() {
		return true;
	}

	/**
	 * Sets the flag that enables or disables panning of the plot along the
	 * domain axes.
	 * 
	 * @param pannable
	 *            the new flag value.
	 * 
	 * @since 1.0.13
	 */
	public void setDomainPannable(boolean pannable) {
	}

	/**
	 * Returns <code>true</code> if panning is enabled for the range axes, and
	 * <code>false</code> otherwise.
	 * 
	 * @return A boolean.
	 * 
	 * @since 1.0.13
	 */
	@Override
	public boolean isRangePannable() {
		return true;
	}

	/**
	 * Sets the flag that enables or disables panning of the plot along the
	 * range axes.
	 * 
	 * @param pannable
	 *            the new flag value.
	 * 
	 * @since 1.0.13
	 */
	public void setRangePannable(boolean pannable) {
	}

	/**
	 * Pans the domain axes by the specified percentage.
	 * 
	 * @param percent
	 *            the distance to pan (as a percentage of the axis length).
	 * @param info
	 *            the plot info
	 * @param source
	 *            the source point where the pan action started.
	 * 
	 * @since 1.0.13
	 */
	@Override
	public void panDomainAxes(double percent, PlotRenderingInfo info, Point2D source) {
		if (!isDomainPannable() || this.domainAxis == null) {
			return;
		}
		double length = this.domainAxis.getRange().getLength();
		double adj = percent * length;
		if (this.domainAxis.isInverted()) {
			adj = -adj;
		}
		this.domainAxis.setRange(this.domainAxis.getLowerBound() + adj, this.domainAxis.getUpperBound() + adj);
	}

	/**
	 * Pans the range axes by the specified percentage.
	 * 
	 * @param percent
	 *            the distance to pan (as a percentage of the axis length).
	 * @param info
	 *            the plot info
	 * @param source
	 *            the source point where the pan action started.
	 * 
	 * @since 1.0.13
	 */
	@Override
	public void panRangeAxes(double percent, PlotRenderingInfo info, Point2D source) {
		if (!isRangePannable() || this.rangeAxis == null) {
			return;
		}
		double length = this.rangeAxis.getRange().getLength();
		double adj = percent * length;
		if (this.rangeAxis.isInverted()) {
			adj = -adj;
		}
		this.rangeAxis.setRange(this.rangeAxis.getLowerBound() + adj, this.rangeAxis.getUpperBound() + adj);
	}

	/**
	 * Tests an arbitrary object for equality with this plot. Note that
	 * <code>FastScatterPlot</code> carries its data around with it (rather than
	 * referencing a dataset), and the data is included in the equality test.
	 * 
	 * @param obj
	 *            the object (<code>null</code> permitted).
	 * 
	 * @return A boolean.
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == this) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (!(obj instanceof DottedChart)) {
			return false;
		}
		DottedChart that = (DottedChart) obj;
		if (!ObjectUtilities.equal(this.domainAxis, that.domainAxis)) {
			return false;
		}
		if (!ObjectUtilities.equal(this.rangeAxis, that.rangeAxis)) {
			return false;
		}
		if (this.domainGridlinesVisible != that.domainGridlinesVisible) {
			return false;
		}
		if (!PaintUtilities.equal(this.domainGridlinePaint, that.domainGridlinePaint)) {
			return false;
		}
		if (!ObjectUtilities.equal(this.domainGridlineStroke, that.domainGridlineStroke)) {
			return false;
		}
		if (!this.rangeGridlinesVisible == that.rangeGridlinesVisible) {
			return false;
		}
		if (!PaintUtilities.equal(this.rangeGridlinePaint, that.rangeGridlinePaint)) {
			return false;
		}
		if (!ObjectUtilities.equal(this.rangeGridlineStroke, that.rangeGridlineStroke)) {
			return false;
		}
		return true;
	}

	public Attribute getXAxisAttribute() {
		return xAttribute;
	}

	public Attribute getYAxisAttribute() {
		return yAttribute;
	}

	public Attribute getColorAttribute() {
		return colorAttribute;
	}

	public Attribute getShapeAttribute() {
		return shapeAttribute;
	}

	//
	// public boolean isYAxisRelative() {
	// return yAxisIsRelative;
	// }
	//
	// public boolean isXAxisRelative() {
	// return xAxisIsRelative;
	// }

	public void setAttribute(AttributeType type, Attribute attribute, boolean signalChange) throws DottedChartException {
		switch (type) {
			case COLOR :
				setColorAttribute(attribute, signalChange);
				break;
			case SHAPE :
				setShapeAttribute(attribute, signalChange);
				break;
			case XAXIS :
				setXAxisAttribute(attribute, signalChange);
				break;
			case YAXIS :
				setYAxisAttribute(attribute, signalChange);
				break;
		}
	}

	public Attribute getAttribute(AttributeType type) {
		switch (type) {
			case COLOR :
				return getColorAttribute();
			case SHAPE :
				return getShapeAttribute();
			case XAXIS :
				return getXAxisAttribute();
			case YAXIS :
				return getYAxisAttribute();
			default :
				return null;
		}
	}

	public String getToolTipForScreenLocation(ChartPanel panel, MouseEvent event) {
		read.lock();
		try {
			StringBuilder builder = new StringBuilder();
			builder.append("<html>");
			// long start = System.currentTimeMillis();
			// There are two options. Either we go through the data to find the
			// event
			// whose X and Y attributes are as in the click, or we go through
			// the
			// events to determine whose attributes would be projected onto
			// X and Y on screen.
			Point2D p = panel.translateScreenToJava2D(event.getPoint());
			Rectangle2D plotArea = panel.getScreenDataArea();
			if (p.getX() < plotArea.getX() && p.getY() <= plotArea.getHeight()) {
				// in Range Axis
				//TODO: show tooltip with actual value?
			} else if (p.getX() >= plotArea.getX() && p.getY() > plotArea.getHeight()) {
				// in Domain Axis
				//TODO: show tooltip with actual value?
			} else {

				double s = dotSize * 0.8;
				double chartX = domainAxis.java2DToValue(p.getX() - s / 2, plotArea, RectangleEdge.BOTTOM);
				double chartY = rangeAxis.java2DToValue(p.getY() - s / 2, plotArea, RectangleEdge.LEFT);

				double chartW = s * domainAxis.getRange().getLength() / plotArea.getWidth();
				double chartH = s * rangeAxis.getRange().getLength() / plotArea.getHeight();
				// we can go both ways.
				// dottedChart.getDomainAxis().valueToJava2D(value, plotArea,
				// RectangleEdge.BOTTOM);

				Rectangle2D rect = new Rectangle2D.Double(chartX, chartY, chartW, chartH);

				if (xAttribute.isEventAttribute() && yAttribute.isEventAttribute()) {
					getEventsForLocationXYEvent(rect, builder);
				} else if (xAttribute.isEventAttribute()) {
					getEventsForLocationXEventYTrace(rect, builder);
				} else if (yAttribute.isEventAttribute()) {
					getEventsForLocationXTraceYEvent(rect, builder);
				} else {
					getEventsForLocationXYTrace(rect, builder);
				}

				// long end = System.currentTimeMillis();
				// System.out.println("Found events in " + (end - start)
				// + " milliseconds.");
			}
			builder.append("</html>");
			return builder.toString();
		} finally {
			read.unlock();
		}
	}

	private void getEventsForLocationXYEvent(Rectangle2D rect, StringBuilder builder) {
		int found = 0;
		double x = 0, y = 0;
		for (int t = 0; t < log.numTraces(); t++) {

			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				//				if (shapes[e] < 0) {
				if ((colorAndShape[e] & VISIBLEMASK) != VISIBLEMASK) {
					continue;
				}
				x = getValue(log.getEventDataManager(), e, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
				y = getValue(log.getEventDataManager(), e, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
				if (rect.contains(x, y)) {
					builder.append("(t:" + t + ",e:" + (e - log.getFirstEventNumber(t)) + "):");
					builder.append(log.getEventDataManager().toString(e));
					builder.append("<br>");
					found++;
					if (found > 10) {
						builder.append("...<br>");
						return;
					}
				}
			}
		}
	}

	private void getEventsForLocationXTraceYEvent(Rectangle2D rect, StringBuilder builder) {
		int found = 0;
		double x = 0, y = 0;
		for (int t = 0; t < log.numTraces(); t++) {

			x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);

			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				//				if (shapes[e] < 0) {
				if ((colorAndShape[e] & VISIBLEMASK) != VISIBLEMASK) {
					continue;
				}
				y = getValue(log.getEventDataManager(), e, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);

				if (rect.contains(x, y)) {
					builder.append("(t:" + t + ",e:" + (e - log.getFirstEventNumber(t)) + "):");
					builder.append(log.getEventDataManager().toString(e));
					builder.append("<br>");
					found++;
					if (found > 10) {
						builder.append("...<br>");
						return;
					}
				}
			}
		}
	}

	private void getEventsForLocationXEventYTrace(Rectangle2D rect, StringBuilder builder) {
		int found = 0;
		double x = 0, y = 0;
		for (int t = 0; t < log.numTraces(); t++) {

			y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);

			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				//				if (shapes[e] < 0) {
				if ((colorAndShape[e] & VISIBLEMASK) != VISIBLEMASK) {
					continue;
				}
				x = getValue(log.getEventDataManager(), e, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);

				if (rect.contains(x, y)) {
					builder.append("(t:" + t + ",e:" + (e - log.getFirstEventNumber(t)) + "):");
					builder.append(log.getEventDataManager().toString(e));
					builder.append("<br>");
					found++;
					if (found > 10) {
						builder.append("...<br>");
						return;
					}
				}
			}
		}
	}

	private void getEventsForLocationXYTrace(Rectangle2D rect, StringBuilder builder) {
		int found = 0;
		double x = 0, y = 0;
		// double xOff = 0, yOff = 0;
		for (int t = 0; t < log.numTraces(); t++) {

			x = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			y = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);

			int max = log.getFirstEventNumber(t) + log.getTraceLength(t);
			for (int e = log.getFirstEventNumber(t); e < max; e++) {
				//				if (shapes[e] < 0) {
				if ((colorAndShape[e] & VISIBLEMASK) != VISIBLEMASK) {
					continue;
				}
				if (rect.contains(x, y)) {
					builder.append("(t:" + t + ",e:" + (e - log.getFirstEventNumber(t)) + "):");
					builder.append(log.getEventDataManager().toString(e));
					builder.append("<br>");
					found++;
					if (found > 10) {
						builder.append("...<br>");
						return;
					}
				}
			}
		}
	}

	public boolean isYAxisContinuous() {
		return (yAttribute.getType() & Attribute.CONTINUOUS_FLAG) == Attribute.CONTINUOUS_FLAG;
	}

	public boolean isXAxisContinuous() {
		return (xAttribute.getType() & Attribute.CONTINUOUS_FLAG) == Attribute.CONTINUOUS_FLAG;
	}

	public XLog exportViewToXLog(boolean visibleOnly, String name) throws DottedChartException {
		read.lock();
		try {
			assert !isYAxisContinuous();

			// find all events in the view.
			// the range axis is discrete, so cast the lower and upper bound to
			// long
			long yMin = visibleOnly ? (long) rangeAxis.getRange().getLowerBound() : Long.MIN_VALUE;
			long yMax = visibleOnly ? (long) rangeAxis.getRange().getUpperBound() : Long.MAX_VALUE;

			final double xMin = visibleOnly ? domainAxis.getRange().getLowerBound() : Double.NEGATIVE_INFINITY;
			final double xMax = visibleOnly ? domainAxis.getRange().getUpperBound() : Double.POSITIVE_INFINITY;

			// First identify all traces by finding all different values on the
			// y axis an event refers to.
			// The map points to the new trace size,
			// i.e. the number of events in the new trace.
			final TDoubleObjectMap<TIntList> newTraces = new TDoubleObjectHashMap<TIntList>();

			getTraces(newTraces, yMin, yMax, xMin, xMax);

			final XLog newXLog = log.getLogDataManager().readObject(0);
			XConceptExtension.instance().assignName(newXLog, name);

			final XFactory factory = XFactoryRegistry.instance().currentDefault();

			newTraces.forEachEntry(new TDoubleObjectProcedure<TIntList>() {

				@Override
				public boolean execute(double newTraceId, TIntList eventList) {
					// Insert events sorted by the appropriate value of the
					// xAxis
					// attribute
					// export the trace to XTrace and go to the next.
					int i = 0;

					// First determine the trace number in the storage.
					int number = eventList.get(0);
					int t = log.getTraceForEvent(number);

					// newTrace contains the events for this trace, sorted by
					// the
					// xAxisAttribute value. Now export and add to log.
					XTrace xTrace;
					if (yAttribute.getKey().equals(XConceptExtension.KEY_NAME)) {
						// the trace identifier doesn't change, so keep original
						// trace attributes
						xTrace = log.getTraceDataManager().readObject(t);
					} else {
						xTrace = factory.createTrace();
					}

					// the new trace should be identified by the attribute value
					// on
					// the y Axis. This becomes the conceptname of the trace
					DataManager<?> manager;
					double yValue;
					if (yAttribute.isEventAttribute()) {
						manager = log.getEventDataManager();
						yValue = getValue(manager, number, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
					} else {
						assert t >= 0;
						manager = log.getTraceDataManager();
						number = t;
						yValue = getValue(manager, t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
					}
					if (yAttribute.getType() == Attribute.LITR
							&& !yAttribute.getKey().equals(XConceptExtension.KEY_NAME)) {
						// in case the attribute is Literal, we need a string
						XConceptExtension.instance().assignName(xTrace,
								log.getLiteral(manager.getAttributeValueLiteralIndex(number, yAttributeIndex)));
					} else if (yAttribute.getType() == Attribute.BOOL) {
						// in case of a boolean, convert to "true" or "false"
						XConceptExtension.instance().assignName(xTrace, yValue > 0.5 ? "true" : "false");
					} else if (yAttribute.getType() == Attribute.DISC) {
						// in case of a discrete, use the value
						XConceptExtension.instance().assignName(xTrace, Long.toString((long) yValue));
					}

					// now add the events to the trace.
					for (i = 0; i < eventList.size(); i++) {
						xTrace.add(log.getEventDataManager().readObject(eventList.get(i)));
					}

					newXLog.add(xTrace);

					return true;
				}

			});

			newXLog.getExtensions().addAll(log.getExtensions().values());

			// now add classifiers according to the color and shape attribute as
			// well as the original classifiers.
			newXLog.getClassifiers().addAll(log.getClassifiers());
			if (colorAttribute != null) {
				if (colorAttribute instanceof ClassifierAttribute) {
					newXLog.getClassifiers().add(
							new XEventAttributeClassifier("Color by DottedChart",
									((ClassifierAttribute) colorAttribute).getOriginalKeys()));
				} else {
					newXLog.getClassifiers().add(
							new XEventAttributeClassifier("Color by DottedChart", colorAttribute.getKey()));
				}
			}
			if (shapeAttribute != null) {
				if (shapeAttribute instanceof ClassifierAttribute) {
					newXLog.getClassifiers().add(
							new XEventAttributeClassifier("Shape by DottedChart",
									((ClassifierAttribute) shapeAttribute).getOriginalKeys()));
				} else {
					newXLog.getClassifiers().add(
							new XEventAttributeClassifier("Shape by DottedChart", shapeAttribute.getKey()));
				}
			}

			// add eventGlobals
			for (Attribute eventGlobal : log.getEventDataManager().getKeys()) {
				newXLog.getGlobalEventAttributes().add(eventGlobal.toXAttribute(factory, log.getExtensions(), log));
			}

			// if the trace notion did not change, add trace attributes as
			// globals
			if (yAttribute.getKey().equals(XConceptExtension.KEY_NAME)) {
				// add traceGlobals
				for (Attribute traceGlobal : log.getTraceDataManager().getKeys()) {
					newXLog.getGlobalTraceAttributes().add(traceGlobal.toXAttribute(factory, log.getExtensions(), log));
				}
			}
			return newXLog;
		} finally {
			read.unlock();
		}

	}

	private int binarySearch(DataManager<?> manager, TIntList newTrace, double value, int currentSize) {
		int low = 0;
		currentSize -= 1;

		while (low <= currentSize) {
			int mid = (low + currentSize) >>> 1;

			// get the value of the xAttribute for the event at index
			// mid
			double midVal;
			if (xAttribute.isEventAttribute()) {
				midVal = getValue(manager, newTrace.get(mid), xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			} else {
				int t = log.getTraceForEvent(newTrace.get(mid));
				midVal = getValue(manager, t, xAttributeIndex, xAttribute.getType(), xCategory2LitIndex);
			}

			if (midVal < value)
				low = mid + 1;
			else if (midVal > value)
				currentSize = mid - 1;
			else
				// insertion point is after or before mid
				// we coose after here, so that the ordering
				// of the original sequence of events is preserved
				return mid + 1;
		}
		// insertion point is before low
		return low;
	}

	public LightLog exportViewToLightLog(boolean visibleOnly, String name) throws DottedChartException {

		read.lock();

		try {
			assert !isYAxisContinuous();

			// find all events in the view.
			// the range axis is discrete, so cast the lower and upper bound to
			// long
			long yMin = visibleOnly ? (long) rangeAxis.getRange().getLowerBound() : Long.MIN_VALUE;
			long yMax = visibleOnly ? (long) rangeAxis.getRange().getUpperBound() : Long.MAX_VALUE;

			final double xMin = visibleOnly ? domainAxis.getRange().getLowerBound() : Double.NEGATIVE_INFINITY;
			final double xMax = visibleOnly ? domainAxis.getRange().getUpperBound() : Double.POSITIVE_INFINITY;

			// First identify all traces by finding all different values on the
			// y axis an event refers to.
			// The map points to the new trace size,
			// i.e. the number of events in the new trace.
			final TDoubleObjectMap<TIntList> newTraces = new TDoubleObjectHashMap<TIntList>();

			getTraces(newTraces, yMin, yMax, xMin, xMax);

			final LightLog newLog = new LightLog(log, name, colorMap.length);

			newTraces.forEachEntry(new TDoubleObjectProcedure<TIntList>() {

				@Override
				public boolean execute(double newTraceId, TIntList eventList) {
					// Insert events sorted by the appropriate value of the
					// xAxis
					// attribute
					// export the trace to XTrace and go to the next.

					int i = 0;

					// First determine the trace number in the storage.
					int number = eventList.get(0);
					int t = log.getTraceForEvent(number);

					// newTrace contains the events for this trace, sorted by
					// the
					// xAxisAttribute value. Now export and add to log.
					LightTrace trace;

					// the new trace should be identified by the attribute value
					// on
					// the y Axis. This becomes the conceptname of the trace
					DataManager<?> manager;
					double yValue;
					if (yAttribute.isEventAttribute()) {
						manager = log.getEventDataManager();
						yValue = getValue(manager, number, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
					} else {
						assert t >= 0;
						manager = log.getTraceDataManager();
						number = t;
						yValue = getValue(manager, t, yAttributeIndex, yAttribute.getType(), yCategory2LitIndex);
					}
					String name;
					if (yAttribute.getType() == Attribute.LITR
							&& !yAttribute.getKey().equals(XConceptExtension.KEY_NAME)) {
						// in case the attribute is Literal, we need a string
						name = log.getLiteral(manager.getAttributeValueLiteralIndex(number, yAttributeIndex));
					} else if (yAttribute.getType() == Attribute.BOOL) {
						// in case of a boolean, convert to "true" or "false"
						name = yValue > 0.5 ? "true" : "false";
					} else if (yAttribute.getType() == Attribute.DISC) {
						// in case of a discrete, use the value
						name = Long.toString((long) yValue);
					} else {
						name = "Trace " + newTraceId;
					}
					// keep trace attributes only if the original conceptname is
					// used for the yAxis
					trace = new LightTrace(log, name, yAttribute.getKey().equals(XConceptExtension.KEY_NAME), eventList
							.size());
					newLog.addTrace(trace);

					if (colorAttribute.isEventAttribute()) {
						manager = log.getEventDataManager();
					} else {
						manager = log.getTraceDataManager();
					}

					// now add the events to the trace.
					for (i = 0; i < eventList.size(); i++) {
						int e = eventList.get(i);
						name = getEventClass(manager, e);
						//						trace.addEvent(new LightEvent(log, e, name, colors[e]));
						trace.addEvent(new LightEvent(log, e, name, colorAndShape[e] & COLORMASK));
					}

					return true;
				}

			});
			return newLog;
		} finally {
			read.unlock();
		}

	}

	// private String showOptionPane(String message) {
	//
	// MyRunnable r = new MyRunnable(message);
	// try {
	// SwingUtilities.invokeAndWait(r);
	// } catch (InvocationTargetException | InterruptedException e) {
	// return null;
	// }
	// return r.result;
	// }
	//
	// private static final class MyRunnable implements Runnable {
	// String result;
	// private String message;
	//
	// public MyRunnable(String message) {
	// this.message = message;
	// }
	//
	// @Override
	// public void run() {
	// // TODO Auto-generated method stub
	// result = JOptionPane.showInputDialog(message);
	// }
	// }

	private void getTraces(TDoubleObjectMap<TIntList> newTraces, double yMin, double yMax, double xMin, double xMax)
			throws DottedChartException {
		try {
			// we need to look at all events as we build an quick indexing
			// structure
			for (int t = 0; t < log.numTraces(); t++) {
				double trace = 0;
				double xVal = 0;
				if (!yAttribute.isEventAttribute()) {
					trace = getValue(log.getTraceDataManager(), t, yAttributeIndex, yAttribute.getType(),
							yCategory2LitIndex);
				}
				if (!xAttribute.isEventAttribute()) {
					xVal = getValue(log.getTraceDataManager(), t, xAttributeIndex, xAttribute.getType(),
							xCategory2LitIndex);
				}
				int lastEvent = log.getFirstEventNumber(t) + log.getTraceLength(t);
				for (int e = log.getFirstEventNumber(t); e < lastEvent; e++) {

					if ((colorAndShape[e] & VISIBLEMASK) != VISIBLEMASK) {
						continue;
					}
					if (yAttribute.isEventAttribute()) {
						trace = getValue(log.getEventDataManager(), e, yAttributeIndex, yAttribute.getType(),
								yCategory2LitIndex);
					}
					if (trace < yMin || trace > yMax) {
						// ignore the event as it's outside of the range
						continue;
					}
					if (xAttribute.isEventAttribute()) {
						xVal = getValue(log.getEventDataManager(), e, xAttributeIndex, xAttribute.getType(),
								xCategory2LitIndex);
					}
					if (xVal < xMin || xVal > xMax) {
						// ignore the event as it's outside of the range
						continue;
					}

					// add a new trace if this is not already present
					TIntList list = newTraces.get(trace);
					if (list == null) {
						list = new TIntArrayList(10);
						newTraces.put(trace, list);
					}
					int offset = binarySearch(log.getEventDataManager(), list, xVal, list.size());
					list.insert(offset, e);
				}
			}
		} catch (IllegalStateException e) {
			// thrown by TLongIntMap if keyset full.
			throw new DottedChartException("Cannot export a log with more than " + Integer.MAX_VALUE + " traces.");
		}
	}

	public LogView getLogView() {
		return log;
	}

	public void setIntensityPanel(JPanel intensityPanel, IntensityChart intensityChart) {
		this.intensityPanel = intensityPanel;
		this.intensityChart = intensityChart;
	}

	private final static DecimalFormat FORMAT = new DecimalFormat("#.##");

	public String getPercentageRendered() {
		return numVisisbleEvents < maxDotCount ? "100" : FORMAT.format(100.0f * maxDotCount / numVisisbleEvents);
	}

	public String getRenderTime() {
		return FORMAT.format(lastRenderTime);
	}

	public Map<String, Color> getColorMap() {
		// create a new hashMap
		Map<String, Color> map = new HashMap<String, Color>(colorMap.length);
		DataManager<?> manager;
		// iterate over all events to find their string representation
		// and the color that belongs to it.
		if (colorAttribute.isEventAttribute()) {
			manager = log.getEventDataManager();
		} else {
			manager = log.getTraceDataManager();
		}

		boolean[] done = new boolean[colorMap.length];
		// now add the events to the trace.
		//		for (int e = 0; e < colors.length; e++) {
		for (int e = 0; e < colorAndShape.length; e++) {
			// a visible event in this log.
			//			int color = colors[e] & 0xFF;
			int color = (colorAndShape[e] & COLORMASK);
			if (!done[color]) {
				String name = getEventClass(manager, e);
				map.put(name, colorMap[color]);
				done[color] = true;
			}
		}

		return map;

	}

	public String getEventClass(DataManager<?> manager, int e) {
		String name;
		switch (colorAttribute.getType()) {
			case Attribute.BOOL : {
				boolean val = manager.getAttributeValueBoolean(map(e, colorAttribute.isEventAttribute()),
						colorAttribute);
				name = val ? "true" : "false";
				break;
			}
			case Attribute.CONT : {
				double val = manager.getAttributeValueContinuous(map(e, colorAttribute.isEventAttribute()),
						colorAttribute);
				name = Double.toString(val);
				break;
			}
			case Attribute.LITR : {
				name = manager.getAttributeValueLiteral(map(e, colorAttribute.isEventAttribute()), colorAttribute);
				break;
			}
			case Attribute.DISC : {
				long val = manager.getAttributeValueDiscrete(map(e, colorAttribute.isEventAttribute()), colorAttribute);
				name = Long.toString(val);
				break;
			}
			case Attribute.TIME_ABS : {
				long val = manager
						.getAttributeValueTimestamp(map(e, colorAttribute.isEventAttribute()), colorAttribute);
				name = DATEFORMATABS.format(new Date(val));
				break;
			}
			case Attribute.TIME_REL : {
				long val = manager
						.getAttributeValueTimestamp(map(e, colorAttribute.isEventAttribute()), colorAttribute);
				name = DATEFORMATREL.format(new Date(val));
				break;
			}
			default :
				name = "";
		}
		return name;
	}

	public LegendItemCollection getLegendItems() {
		LegendItemCollection itemCollection = new LegendItemCollection();
		//	     for (Comparable comparable : legendKeys) {
		//	        Paint paint = // get the paint you want

		for (int i = 0; i < colorMap.length; i++) {
			itemCollection.add(new LegendItem(colorLabels[i], colorLabels[i], colorLabels[i], "",
					new Rectangle2D.Double(0, 0, 10, 10), colorMap[i]));
		}
		for (int i = 0; i < shapeLabels.length; i++) {

			Shape s;

			switch (shapeMap[i]) {
				case 0 :
					// circle
					s = new Ellipse2D.Double(0, 0, 10, 10);
					break;
				case 1 :
					// square
					s = new Rectangle2D.Double(0, 0, 10, 10);
					break;
				case 2 :
					// triangle up
					s = new GeneralPath();
					((GeneralPath) s).moveTo(5, 0);
					((GeneralPath) s).lineTo(10, 10);
					((GeneralPath) s).lineTo(0, 10);
					((GeneralPath) s).closePath();
					break;
				case 3 :
					// triangle down
					s = new GeneralPath();
					((GeneralPath) s).moveTo(5, 10);
					((GeneralPath) s).lineTo(0, 0);
					((GeneralPath) s).lineTo(10, 0);
					((GeneralPath) s).closePath();
					break;
				case 4 :
					// diamond
					s = new GeneralPath();
					((GeneralPath) s).moveTo(0, 5);
					((GeneralPath) s).lineTo(5, 10);
					((GeneralPath) s).lineTo(10, 5);
					((GeneralPath) s).lineTo(5, 0);
					((GeneralPath) s).closePath();
					break;
				case 5 :
					// pentagon
					s = new GeneralPath();
					((GeneralPath) s).moveTo(5, 0);
					((GeneralPath) s).lineTo(10, 4);
					((GeneralPath) s).lineTo(8, 10);
					((GeneralPath) s).lineTo(2, 10);
					((GeneralPath) s).lineTo(0, 4);
					((GeneralPath) s).closePath();
					break;
				case 6 :
					// star
					s = new GeneralPath();
					((GeneralPath) s).moveTo(5, 0);
					((GeneralPath) s).lineTo(8, 10);
					((GeneralPath) s).lineTo(0, 4);
					((GeneralPath) s).lineTo(10, 4);
					((GeneralPath) s).lineTo(2, 10);
					((GeneralPath) s).closePath();
					break;
				case 7 :
					// bowtie
					s = new GeneralPath();
					((GeneralPath) s).moveTo(0, 0);
					((GeneralPath) s).lineTo(10, 10);
					((GeneralPath) s).lineTo(10, 0);
					((GeneralPath) s).lineTo(0, 10);
					((GeneralPath) s).closePath();
					break;
				default :
					assert false;
					s = new GeneralPath();

			}
			itemCollection.add(new LegendItem(shapeLabels[i], shapeLabels[i], shapeLabels[i], "", s, Color.BLACK));

		}
		//	     }
		return itemCollection;
	}

	/**
	 * fire change event
	 * added by Guangming on June 21, 2018
	 */
    public void fireChangeEvent() {
    	System.out.println("enter fireChangeEvent");
 //   	super.fireChangeEvent();
    	notifyListeners(new PlotChangeEvent(this));
    	System.out.println("exist fireChangeEvent");
    }
}
